/*****************************************************************************
 *
 * $RCSfile: AboutForm.c,v $
 *
 * About Form event loop for FetchMap
 *   - Display form when opened
 *   - Close form when OK button is pressed 
 *
 * created: 2003-01-04
 *
 * History:
 * 2003-01-04:	first version
 * 2003-02-09:  replaced labels by field to be able to display more text,
 *                replaced bitmap
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2003/10/24 15:56:42 $
 *
 * $Revision: 1.1.1.1 $
 *
 * $Log: AboutForm.c,v $
 * Revision 1.1.1.1  2003/10/24 15:56:42  mp
 * Imported FetchMap to CVS revision control.
 *
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include "ResourceDefines.h"
#include "ScrollUtils.h"

extern void SetFieldText(UInt16 fldID, Char *srcP, Boolean append, 
                           Boolean redraw);

Boolean AboutFormHandleEvent(EventPtr eventP)
{
  Boolean	handled = false;
  FormPtr	frmP;
  FieldPtr	fldP;
  MemHandle	msgH; 
  MemPtr	msgP;

  switch (eventP->eType) {
    case frmOpenEvent:
      frmP = FrmGetActiveForm();
      
      /* get message string from resource */
      msgH = DmGet1Resource(strRsc, AboutString);
      if (msgH != NULL) {
        /* string handle valid, lock it */
        msgP = MemHandleLock(msgH);
	
	/* display message string in field */
        SetFieldText(AboutField, msgP, false, true);
	UpdateScrollbar(FrmGetActiveForm(), AboutField, AboutScrollBar);
        
	/* unlock string handle */
	MemHandleUnlock(msgH);
	
	/* release string resource */
        DmReleaseResource(msgH);
      }
      FrmDrawForm(frmP);
      handled = true;
      break;

    case keyDownEvent:
     {
	WChar	c = eventP->data.keyDown.chr;

	// check for virtual char. On 2-byte system (like Japanese),
	// we could have a real character that maps to 
	// vchrPageUp or vchrPageDown
	if (TxtGlueCharIsVirtual(eventP->data.keyDown.modifiers, c)) {
          if (c == vchrPageUp) {
            PageScroll(FrmGetActiveForm(), AboutField,
	      AboutScrollBar, winUp);
	    handled = true;
	  } else if (c == vchrPageDown) {
	    PageScroll(FrmGetActiveForm(), AboutField,
	      AboutScrollBar, winDown);
	    handled = true;
          }
	}
      }
      break;
  
    case sclRepeatEvent:
      ScrollLines(FrmGetActiveForm(), AboutField, AboutScrollBar,
	eventP->data.sclRepeat.newValue - 
	eventP->data.sclRepeat.value, false);
      break;
      
    case ctlSelectEvent:
      switch (eventP->data.ctlSelect.controlID) {
        case AboutOKButton:
	  /* Deallocate Fields' Text Memory */
          SetFieldText(AboutField, NULL, false, true);
          FrmGotoForm(MainForm);
          handled = true;
          break;
      }
      
    default:
      break;
  }

  return (handled);
}
