/*****************************************************************************
 *
 * $RCSfile: DownloadMap.c,v $
 *
 * FetchMap
 *
 * Map Download
 *   - network connection set-up, HTTP transfer and file creation
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005/03/05 15:14:49 $
 *
 * $Revision: 1.6 $
 *
 * $Log: DownloadMap.c,v $
 * Revision 1.6  2005/03/05 15:14:49  mp
 * added VFSFileCreate()
 *
 * Revision 1.5  2005/03/05 11:55:11  mp
 * added VFS support
 *
 * Revision 1.4  2004/04/04 13:44:15  mp
 * added variable
 *
 * Revision 1.3  2004/03/23 20:28:34  mp
 * Cleaned up database name generation. Removed code for Mapblast URL
 * generation. Added URL generation with gen_expedia_url() (expedia.c).
 * Added handling of HTTP redirection (required for Expedia).
 *
 * Revision 1.2  2003/11/16 14:54:52  mp
 * replaced printDouble by format_number
 *
 * Revision 1.1  2003/10/27 21:20:21  mp
 * Moved downloadMap() from MainForm.c, replaced SetFieldText(StatusField, ...)
 * by Progress Manager functions.
 *
 *
 *
 ****************************************************************************/

#include <PalmOS.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include <sys_socket.h>
#include "ResourceDefines.h"
#include "common.h"
#include "Constants.h"
#include "DownloadMap.h"
#include "fp.h"
#include "expedia.h"
#include "http_lib.h"

#define MAX_REDIR	5	/* max. number of HTTP redirections */
#define NUM_DIGITS	7
//#define DEBUG_URL


/* GLOBALS */
Err	errno;		/* used by NetLib */

/* LOCALS */
//static Char InputErrMsg[] = "Invalid Input! ";
//static Char InputOkMsg[]  = "Input o.k.     ";
static Char ConnectMsg[]  = "Connecting...  ";
static Char ConnErrMsg[]  = "Conn. Error!   ";
static Char TransferMsg[] = "Transferring...";
static Char HttpErrMsg[]  = "HTTP Error!    ";
static Char FileErrMsg[]  = "File Error!    ";
//static Char DoneMsg[]     = "Done.          ";
//static Char ClrMsg[]      = "               ";

/*
 * downloadMap:
 *   Retrieve map via HTTP from Mapblast server. The required parameters 
 *     position (center of map), scale, and map size in pixels are read
 *     from preferences.
 *
 * Note:	Worst Case Length of Database Name:
 *		200M-52.1614999-110.3126601.gif
 *		200M+52.1614999+110.3126601.gif
 * Returns:
 * --------
 *	Error Code
 *	(see http_lib.h for http_lib error codes,
 *	 see FileStream.h for FileStream Manager error codes
 *	 (fileErrorClass Offset = 0x1600))
 */
Int16 downloadMap(MapParamType* map, ProgressPtr prgP, UInt16 vfs_vol,
                  Boolean use_vfs)
{
  Char		url[160];	/* URL string */
  Char		tmp[60];	/* temporary string */
  Int16		ret;		/* HTTP library return code */
  UInt16	redir_cnt = 0;	/* HTTP redirection counter */
  Char		*filename=NULL;	/* HTTP filename */
  MemPtr	pathName;	/* VFS pathname */
  long		lg;		/* content length */
  Char		typeBuf[32];	/* content type */
  Char		*data=NULL;	/* content buffer */
  FileHand	fd;		/* file handle */
  FileRef	fr;		/* file reference (VFS) */
  Char		dbname[32];	/* database (file) name */
  Err		err;		/* file error code */
  UInt16	ifErrs;		/* network IF errors */
  
  /* --- Create URL and Database Name --- */
  MemSet(dbname, 32, 0);
  MemSet(url, 160, 0);
      
  /* Add scale to Database Name */
  if (map->scale >= 1000000L) {
    /* 1,000,000 -> 1M */
    StrPrintF(tmp, "%luM", map->scale / 1000000L);
  } else if (map->scale >= /*1000*/10000L) {
    /* 1,000 -> 1k */
    StrPrintF(tmp, "%luk", map->scale / 1000L);
  } else {
    StrPrintF(tmp, "%lu", map->scale);
  }
  StrCopy(dbname, tmp);
    
  /* Add Latitude */
  format_number(map->lat, NUM_DIGITS, tmp);
  
  if (map->lat >= 0) {
    StrCat(dbname, "+");
  }
  StrCat(dbname, tmp);
  
  /* Add Longitude */
  format_number(map->lon, NUM_DIGITS, tmp);
  
  if (map->lon >= 0) {
    StrCat(dbname, "+");
  }
  StrCat(dbname, tmp);
  StrCat(dbname, ".gif");
  
#ifdef DEBUG_DBNAME
  WinDrawChars(dbname, StrLen(dbname), 0, 100);
#endif

  /* generate URL for expedia */
  /* TBD: add UI for region / server selection */
  gen_expedia_url(url, map->lat, map->lon, map->scale, map->width, map->height,
    EXPEDIA_MAP_DEF /* reg */, EXPEDIA_SRV_COM /* srv */);
  
#ifdef DEBUG_URL
  WinDrawChars(&url[0], 30, 0, 15);
  WinDrawChars(&url[30], 30, 0, 30);
  WinDrawChars(&url[60], 30, 0, 45);
  WinDrawChars(&url[90], 30, 0, 60);
  WinDrawChars(&url[120], StrLen(&url[120]), 0, 75);
#endif
  
  /* Note: see sys_socket.h for AppNetTimeout */
  //   AppNetTimeout = sysTicksPerSecond * 5;
  AppNetTimeout = sysTicksPerSecond * 500;

  PrgUpdateDialog(prgP, 0, 0, ConnectMsg, true);
  
  /* open network library, this will establish the dial-up connection! */
  err = NetLibOpen(AppNetRefnum, &ifErrs);  					  
  
  if (((err == 0) || (err == netErrAlreadyOpen)) && (ifErrs == 0)) {
    PrgUpdateDialog(prgP, 0, 0, TransferMsg, true);
  } else {
    NetLibClose(AppNetRefnum, true /* immediate */);
    PrgUpdateDialog(prgP, 0, 0, ConnErrMsg, true);
    return err;
  }
  
  /* Handle multiple levels of indirection */
  ret = http_parse_url(url, &filename);
  /*
  do {
    ret = http_parse_url(url, &filename);
    if (ret < 0) {
      httpErrAlert(ret);
      PrgUpdateDialog(prgP, 0, 0, HttpErrMsg, true);
      return ret;
    }
    
    ret = http_head(filename, &lg, typeBuf, url);
  } while (((ret == RED301) || (ret == RED302)) && (redir_cnt++ < MAX_REDIR));
      
  if (ret < 0) {
    httpErrAlert(ret);
    PrgUpdateDialog(prgP, 0, 0, HttpErrMsg, true);
    return ret;
  }
  */

#ifdef DEBUG
  {
    Char		tmpBuf[32];	/* content length (as string) */

    /* set Length-Field to content length */
    sprintf(tmpBuf, "%d", lg);
    
    WinDrawChars(tmpBuf, StrLen(tmpBuf), 0, 10);
    WinDrawChars(typeBuf, StrLen(typeBuf), 0, 20);
    WinDrawChars(filename, StrLen(filename), 0, 30);
  }  
#endif  

  if (!use_vfs) {
    /* using File Streaming */
    fd = FileOpen(  0, /* cardNo */
		    dbname, /* nameP */
		    kFileTypeId, /* type */
		    kFileCreatorId, /* creator */
		    fileModeReadWrite, /* openMode */
		    &err /* errP */);
  } else {
    /* using VFS */
    
    /* generate absolute file path */
    pathName = MemPtrNew(sizeof(FILEDIR) + StrLen(dbname) + 2);
    StrCopy(pathName, FILEDIR);
    StrCat(pathName, "/");
    StrCat(pathName, dbname);
    
    err = VFSFileCreate(vfs_vol, /* volRefNum */
                        pathName /* pathNameP */);
    
    if (err == errNone || err == vfsErrFileAlreadyExists) {
      err = VFSFileOpen(vfs_vol, /* volRefNum */
    			pathName, /* pathNameP */
			vfsModeWrite, /* openMode */
			&fr /* fileRefP*/ );
    }
    MemPtrFree(pathName);
  }
  
  if (err != 0) {
    fileErrAlert(err);
    PrgUpdateDialog(prgP, 0, 0, FileErrMsg, true);
    return err;
  }

  
  /* get HTTP contents (stored in buffer 'data') */
  ret = http_get2(filename, &data, &lg, typeBuf, fr);
  
  if (ret < 0) {
    httpErrAlert(ret);
    PrgUpdateDialog(prgP, 0, 0, HttpErrMsg, true);
    return ret;
  }



#if 0
  if (!use_vfs) {
    FileWrite(    fd, /* stream */
		  data, /* dataP */
		  sizeof(char), /* objSize */
		  lg, /* numObj */
		  &err /* errP */);
  } else {
    err = VFSFileWrite(	fr, /* fileRef */
    			lg, /* numBytes */
			data, /* dataP */
			NULL /* numBytesWrittenP */);
  }
#endif
  
  if (err != 0) {
    fileErrAlert(err);
    PrgUpdateDialog(prgP, 0, 0, FileErrMsg, true);
    return err;
  }

  if (!use_vfs) {
    err = FileClose(fd /* stream */);
  } else {
    err = VFSFileClose(fr);
  }
  
  return 0;
}

Boolean callbackFunction(PrgCallbackDataPtr cbP)
{
  StrCopy(cbP->textP, cbP->message);
  
  return true;
}
