/*****************************************************************************
 *
 * $RCSfile: ScrollUtils.c,v $
 *
 * Copyright (c) 2000-2001, Neil Rhodes and Julie McKeehan
 *    neil@pobox.com
 * All rights reserved.
 *
 * From the book "Palm Programming, the Developer's Guide, 2nd edition"
 * by O'Reilly.
 *
 * Permission granted to use this file however you see fit.
 *
 * $Author: mp $
 *
 * $Date: 2003/10/24 15:56:42 $
 *
 * $Revision: 1.1.1.1 $
 *
 * $Log: ScrollUtils.c,v $
 * Revision 1.1.1.1  2003/10/24 15:56:42  mp
 * Imported FetchMap to CVS revision control.
 *
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include "ScrollUtils.h"

void UpdateScrollbar(FormPtr form, UInt16 fieldID, UInt16 scrollbarID)
{
   ScrollBarPtr   scroll;
   FieldPtr       field;
   UInt16         currentPosition;
   UInt16         textHeight;
   UInt16         fieldHeight;
   UInt16         maxValue;
   
   field = (FieldPtr) FrmGetObjectPtr(form, 
     FrmGetObjectIndex(form, fieldID));
   FldGetScrollValues(field, &currentPosition, &textHeight, &fieldHeight);

   // if the field is 3 lines, and the text height is 4 lines
   // then we can scroll so that the first line is at the top 
   // (scroll position 0) or so the second line is at the top
   // (scroll postion 1). These two values are enough to see
   // the entire text.
   if (textHeight > fieldHeight)
      maxValue = textHeight - fieldHeight;
   else if (currentPosition)
      maxValue = currentPosition;
   else
      maxValue = 0;
      
   scroll = (ScrollBarPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, scrollbarID));
   
   // on a page scroll, want to overlap by one line (to provide context)
   SclSetScrollBar(scroll, (Int16) currentPosition, 0, 
     (Int16) maxValue, (Int16) fieldHeight - 1);
}

void ScrollLines(FormPtr form, UInt16 fieldID, UInt16 scrollbarID,
  Int16 numLinesToScroll, Boolean redraw)
{
   FormPtr        frm = FrmGetActiveForm();
   FieldPtr       field;
   
   field = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
   if (numLinesToScroll < 0)
      FldScrollField(field, (UInt16) -numLinesToScroll, winUp);
   else
      FldScrollField(field, (UInt16) numLinesToScroll, winDown);
      
   // if there are blank lines at the end and we scroll up, FldScrollField
   // makes the blank lines disappear. Therefore, we've got to update
   // the scrollbar
   if ((FldGetNumberOfBlankLines(field) && numLinesToScroll < 0) ||
      redraw)
      UpdateScrollbar(form, fieldID, scrollbarID);
}

void PageScroll(FormPtr form, UInt16 fieldID, UInt16 scrollbarID,
  WinDirectionType direction)
{
   FieldPtr       field;
   
   field = (FieldPtr) FrmGetObjectPtr(form, 
     FrmGetObjectIndex(form, fieldID));
   if (FldScrollable(field, direction)) {
      Int16 linesToScroll = (Int16) FldGetVisibleLines(field) - 1;
      
      if (direction == winUp)
         linesToScroll = -linesToScroll;
      ScrollLines(form, fieldID, scrollbarID, linesToScroll, true);
   }
}
