/*****************************************************************************
 *
 * $RCSfile: SetFieldText.c,v $
 *
 * FetchMap
 * Text Field Utility Function
 *
 * created: 2003-01-04
 *
 * History:
 * 2003-01-04:	first version
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004/04/04 14:01:40 $
 *
 * $Revision: 1.2 $
 *
 * $Log: SetFieldText.c,v $
 * Revision 1.2  2004/04/04 14:01:40  mp
 * added include directive
 *
 * Revision 1.1.1.1  2003/10/24 15:56:42  mp
 * Imported FetchMap to CVS revision control.
 *
 *
 *
 ****************************************************************************/
#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include "MainForm.h"


/***********************************************************************
 *
 * FUNCTION:	SetFieldText
 *
 * DESCRIPTION:	Set field object's text handle.  Will reuse an existing
 *						text handle, if any
 *
 * PARAMETERS:	fldID	-- field object id
 *		srcP	-- source text pointer
 *		append	-- if true, the new text will be appended
 *		redraw	-- if true, field will be redrawn
 *
 * RETURNED:    nothing.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
void SetFieldText(UInt16 fldID, Char *srcP, Boolean append, Boolean redraw)
{
  MemHandle 		stringH;
  UInt16		oldLength = 0;
  UInt16		newSize;
  Char 			*stringP;
  FieldPtr		fldP;
  FormPtr		frmP;
  
  frmP = FrmGetActiveForm();
  fldP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP,fldID));

  if (!srcP || !*srcP) 
	{				// handle clearing field as well
	  FldFreeMemory(fldP);
	  if ( redraw )	 
		{
		  FldEraseField( fldP );
		  FldDrawField( fldP );
		}
	  return;
	}
  
  
  newSize = StrLen(srcP) + 1;		// initialize new size
  stringH = FldGetTextHandle( fldP );	// get the current text handle
  FldSetTextHandle( fldP, 0 );		// release this handle from field

  // Resize the existing handle, if any
  if ( stringH ) 
	{
	  if ( append ) 
		{
		  stringP = MemHandleLock( (MemHandle)stringH );
		  oldLength = StrLen(stringP);
		  newSize += oldLength;
		  MemPtrUnlock( stringP );
		}
	  if ( MemHandleResize ( (MemHandle)stringH, newSize) )
		  goto Exit;
	} // Resize the existing handle, if any
  
  // Otherwise, allocate a new handle
  else 
	{
	  stringH = MemHandleNew( newSize );	// allocate a new chunk
	  if ( !stringH )	return;
	}

  // Append the new text
  stringP = MemHandleLock( (MemHandle)stringH );
  StrCopy( stringP + oldLength, srcP );			// copy the new text
  MemPtrUnlock( stringP );
  
Exit:
  FldSetTextHandle( fldP, stringH );		// set the new text handle
  if ( redraw )	 
	{
	  FldEraseField( fldP );
	  FldDrawField( fldP );
	}
}
