/*****************************************************************************
 *
 * $RCSfile: err_alert.c,v $
 *
 * FetchMap
 * Error Alert Functions
 *   - Translates return/error codes to appropriate messages
 *   - only most likely http_lib and File Streaming Errors are implemented
 *   - unknown codes are simply shown without translation 
 *
 * created: 2003-02-08
 *
 * History:
 * 2003-01-08:	first version
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005/03/05 11:52:35 $
 *
 * $Revision: 1.2 $
 *
 * $Log: err_alert.c,v $
 * Revision 1.2  2005/03/05 11:52:35  mp
 * added VFS support
 *
 * Revision 1.1.1.1  2003/10/24 15:56:42  mp
 * Imported FetchMap to CVS revision control.
 *
 *
 *
 ****************************************************************************/
 
#include <PalmOS.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include "ResourceDefines.h"
#include "http_lib.h"
#include "err_alert.h"

/* see http_lib.c for error codes */
void httpErrAlert(Int16	ret)
{
  Char errmsg[40];			/* error message */
  
  switch (ret) {
    /* client side errors */
    /* Note: some error codes are combined to save memory */
    case ERRHOST:
      StrCopy(errmsg, "No such host.");
      break;

    case ERRCONN:
      StrCopy(errmsg, "Can't connect to host.");
      break;          
    
    case ERRSOCK:
    case ERRWRHD:
    case ERRWRDT:
    case ERRRDHD:
      StrCopy(errmsg, "Can't create/read/write socket.");
      break;
    
    case ERRPAHD:
      StrCopy(errmsg, "ERRPAHD: Invalid data.");
      break;


    case ERRNULL:
      StrCopy(errmsg, "ERRNULL: Invalid data.");
      break;

    case ERRNOLG:
      StrCopy(errmsg, "ERRNOLG: Invalid data.");
      break;

    case ERRRDDT:
      StrCopy(errmsg, "ERRRDDT: Invalid data.");
      break;
      
    case ERRMEM:
      StrCopy(errmsg, "Can't allocate memory.");
      break;
    
    /* host side errors */
    case ERR400:
      StrCopy(errmsg, "Invalid query.");
      break;
    
    case ERR403:
      StrCopy(errmsg, "Forbidden access.");
      break;
    
    case 404:
      /* missing in http_lib, but not quite uncommon */
      StrCopy(errmsg, "File not found.");
      break;
   
    case ERR408:
      StrCopy(errmsg, "Request timeout.");
      break;
   
    case ERR500:
      StrCopy(errmsg, "Server error.");
      break;
   
    case ERR501:
      StrCopy(errmsg, "Not implemented.");
      break;
   
    case ERR503:
      StrCopy(errmsg, "Service overloaded.");
      break;
      
    
    default:
      StrIToA(errmsg, ret);
  } /* switch (ret) */
  
  FrmCustomAlert(HttpErrAlert, errmsg, NULL, NULL);
}

/* see FileStream.h for error codes */
void fileErrAlert(Err err)
{
   Char errmsg[40];			/* error message */
   
   switch (err) {
     case fileErrMemError:
     case vfsErrVolumeFull:
       StrCopy(errmsg, "Out of memory.");
       break;
     
     case fileErrInvalidParam:
       StrCopy(errmsg, "Invalid parameter.");
       break;

     case fileErrReplaceError:
       StrCopy(errmsg, "Couldn't replace file.");
       break;
     
     case fileErrCreateError:
       StrCopy(errmsg, "Couldn't create file.");
       break;
     
     case fileErrOpenError:
     case expErrNotOpen:
     case vfsErrBadName:
     case vfsErrFileNotFound:
     case vfsErrFilePermissionDenied:
       StrCopy(errmsg, "Couldn't open file.");
       break;
     
     case expErrCardReadOnly:
       StrCopy(errmsg, "Card is read-only.");
       break;
     
     case vfsErrVolumeBadRef:
       StrCopy(errmsg, "Volume not mounted.");
       break;     
     
     case fileErrReadOnly:
     case fileErrPermissionDenied:
     case vfsErrFileBadRef:
     case vfsErrIsADirectory:
     case vfsErrNoFileSystem:
       StrCopy(errmsg, "Couldn't write file.");
       break;
       
     default:
       StrIToH(errmsg, err);
   } /* switch (err) */
       
   FrmCustomAlert(FileErrAlert, errmsg, NULL, NULL);
}
