/*****************************************************************************
 *
 * $RCSfile: expedia.c,v $
 *
 * Generate URL for Map Download from Expedia
 *
 * Based on fetchmap.pl 1.07 by Kevin Stephens, Sven Fichtner et al.
 *
 * This program is Copyright (C) 03/2004 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005/03/05 11:40:01 $
 *
 * $Revision: 1.2 $
 *
 * $Log: expedia.c,v $
 * Revision 1.2  2005/03/05 11:40:01  mp
 * moved static function declaration from expedia.h
 *
 * Revision 1.1  2004/03/23 20:57:58  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "expedia.h"
#include "fp.h"

#define EXPEDIAFACT	3950	/* meaning? */

static const Char	*map_regions[] = {
  "WLD0809",	/* EXPEDIA_MAP_DEF */
  "EUR0809",	/* EXPEDIA_MAP_EU */
  "USA0809",	/* EXPEDIA_MAP_US */
  "WLD0809"	/* EXPEDIA_MAP_WORLD */
};

static const Char	*map_servers[] = {
  "com",	/* EXPEDIA_SRV_COM */
  "de"		/* EXPEDIA_SRV_DE */
};

static const UInt16 expedia_alts[] = {
  1, 3, 6, 12, 25, 50, 150, 800, 2000, 7000, 12000
};

/* Note: function must reside in the default code section! */
static UInt16 scaletoalt(UInt32 scale, UInt32 *act_scale);


/*****************************************************************************
 * FUNCTION:	scaletoalt
 *
 * DESCRIPTION:	Map scale to Expedia altitude (???).
 *
 * PARAMETERS:	scale		-- desired scale
 *		act_scale	-- ptr to actual scale
 *				  (or NULL if not required)
 *
 * RETURNED:	Expedia Altitude
 *****************************************************************************/
static
UInt16 scaletoalt(UInt32 scale, UInt32 *act_scale)
{
  double ns;
  double di = 999999.0;		/* difference desired scale vs. actual scale */ 
  UInt16 i;			/* loop index */
  UInt16 found = 0;		/* index to resulting altitude */
  UInt16 alti = 0;		/* expedia altitude */
  
  ns = scale/EXPEDIAFACT;

  for (i=0; i < (sizeof(expedia_alts)/sizeof(UInt16)); i++) {              
    if (abs(ns-expedia_alts[i]) < di) {
      di = ns-expedia_alts[i];
      found = i;
    }
  }

  alti = expedia_alts[found];
  
  if (act_scale) {
    *act_scale = alti * EXPEDIAFACT;
  }
  
#if VERBOSE
  {
    char tmp[160]

    sprintf(tmp, "Using expedia altitude %u for requested scale %lu:1\
actual scale %lu:1", alti, scale, alti * EXPEDIAFACT);
  }
#endif

  return alti;
}


/*****************************************************************************
 * FUNCTION:	gen_expedia_url
 *
 * DESCRIPTION:	Generate Expedia URL.
 *
 * PARAMETERS:	strP		-- ptr to string buffer
 *		lat		-- latitude (in degrees)
 *		lon		-- longitude (in degrees)
 *		scale		-- map scale
 *		width		-- width (in pixels)
 *		height		-- height (in pixels)
 *		reg		-- Expedia region
 *		srv		-- Expedia server 
 *
 * RETURNED:	%
 *****************************************************************************/
void gen_expedia_url(Char *strP, double lat, double lon, UInt32 scale,
  UInt16 width, UInt16 height, UInt16 reg, UInt16 srv)
{
  UInt16	alti;		/* expedia altitude */
  Char		reg_str[8];	/* map region string */
  Char		lat_str[20];	/* latitude string */
  Char		lon_str[20];	/* longitude string */

  /* get Expedia altitude */
  //  alti = scaletoalt(scale, NULL);
  
  /* copy Expedia map region */
  /*
  switch (reg) {  
  case EXPEDIA_MAP_US:
    case EXPEDIA_MAP_EU:
    case EXPEDIA_MAP_WORLD:
      StrCopy(reg_str, map_regions[reg]);
      break;
      
    case EXPEDIA_MAP_DEF:
      if (lon < -30.0) {
	StrCopy(reg_str, map_regions[EXPEDIA_MAP_US]);
      } else {
	StrCopy(reg_str, map_regions[EXPEDIA_MAP_EU]);
      }
      break;
  }
  */

  /* copy start of URL */
  //  StrCopy(strP, "http://www.expedia.");
  StrCopy(strP, "http://osmap.abmh.de");
  //  StrCopy(strP, "http://192.168.1.4");
  
  /* append server top level domain */
  /*  if (srv == EXPEDIA_SRV_DE) {
    StrCat(strP, map_servers[EXPEDIA_SRV_DE]);
  } else {
    StrCat(strP, map_servers[EXPEDIA_SRV_COM]);
  }
  */
  
  /* convert lat/lon to strings */
  format_number(lat, 7, lat_str);
  format_number(lon, 7, lon_str);
  
  /* complete URL */
  /*  StrPrintF(&strP[StrLen(strP)],
    "/pub/agent.dll?qscr=mrdt&ID=3XNsF.&CenP=%s,%s&Lang=%s&Alti=%u\
&Size=%u,%u&Offs=0.000000,0.000000",
lat_str, lon_str, reg_str, alti, width, height);*/

  StrPrintF(&strP[StrLen(strP)],
    "/cgi-bin/fetchosmap?lat=%s&lon=%s&scale=%lu",
    lat_str, lon_str, scale);

#ifdef DEBUG_URL
  WinDrawChars(strP, StrLen(strP), 0, 115);
#endif

}
