/*****************************************************************************
 *
 * $RCSfile: mapblast.c,v $
 *
 * Generate URL for Map Download from Mapblast
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004/03/23 20:53:18 $
 *
 * $Revision: 1.1 $
 *
 * $Log: mapblast.c,v $
 * Revision 1.1  2004/03/23 20:53:18  mp
 * moved from DownloadMap.c (for reference only)
 *
 *
 ****************************************************************************/

#include <PalmOS.h>
#include "fp.h"

#define NUM_DIGITS 7


/*****************************************************************************
 * FUNCTION:	gen_mapblast_url
 *
 * DESCRIPTION:	Generate Mapblast/Vicinity URL.
 *
 * PARAMETERS:	strP		-- ptr to string buffer
 *		lat		-- latitude (in degrees)
 *		lon		-- longitude (in degrees)
 *		scale		-- map scale
 *		width		-- width (in pixels)
 *		height		-- height (in pixels)
 *		reg		-- Expedia region
 *		srv		-- Expedia server 
 *
 * RETURNED:	%
 *****************************************************************************/
void gen_mapblast_url(Char *strP, double lat, double lon, UInt32 scale,
  UInt16 width, UInt16 height, Boolean metric)
{
  /*const Char	baseurl[] = "http://www.mapblast.com/myblastd/MakeMap.d?&CT=";*/
  const Char	baseurl[] = "http://www.vicinity.com/gif?&CT=";
  Char		url[160];	/* URL string */
  Char		lat_str[20];	/* latitude string */
  Char		lon_str[20];	/* longitude string */
  
  /* Add base URL */
  StrCopy(url, baseurl);
    
  /* convert lat/lon to strings */
  format_number(lat, NUM_DIGITS, lat_str);
  format_number(lon, NUM_DIGITS, lon_str);

  /* complete URL */
  StrPrintF(&strP[StrLen(strP)],
    "%s:%s:%lu&IC=&W=%u&H=%u&FAM=myblast&LB=.",
    lat_str, lon_str, scale, width, height);

  if (metric) {
    StrCat(strP, "&DU=KM");
  }

#ifdef DEBUG_URL
  WinDrawChars(strP, StrLen(strP), 0, 115);
#endif

}
