/*****************************************************************************
 *
 * $RCSfile: AboutForm.c,v $
 *
 * GPS4Palm About Form
 *
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-03-25 13:36:06 $
 *
 * $Revision: 1.3 $
 *
 * $Log: AboutForm.c,v $
 * Revision 1.3  2005-03-25 13:36:06  mp
 * fixed file header
 *
 * Revision 1.2  2004/12/10 19:52:07  mp
 * replaced DmGet1Resource by DmGetResource
 *
 * Revision 1.1  2004/11/24 21:17:55  mp
 * initial version
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include "ResourceDefines.h"
#include "Utils.h"
#include "AboutForm.h"
#include "common.h"

/* Global Variables */
extern PrefsType	gPrefs;			/* application preferences */


/****************************************************************************/
/**
 * \brief	Route Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean AboutFormHandleEvent(EventPtr eventP)
{
  Boolean	handled = false;
  FormPtr	frmP;
  MemHandle	msgH; 
  MemPtr	msgP;

  switch (eventP->eType) {
    case frmOpenEvent:
      frmP = FrmGetActiveForm();
      
      /* get message string from resource */
      msgH = DmGetResource(strRsc, AboutString);
      if (msgH != NULL) {
        /* string handle valid, lock it */
        msgP = MemHandleLock(msgH);
	
	/* display message string in field */
        SetFieldText(AboutField, msgP, false, true);
	UpdateScrollbar(FrmGetActiveForm(), AboutField, AboutScrollBar);
        
	/* unlock string handle */
	MemHandleUnlock(msgH);
	
	/* release string resource */
        DmReleaseResource(msgH);
      }
      FrmDrawForm(frmP);
      handled = true;
      break;

    case keyDownEvent:
     {
	WChar	c = eventP->data.keyDown.chr;

	// check for virtual char. On 2-byte system (like Japanese),
	// we could have a real character that maps to 
	// vchrPageUp or vchrPageDown
	//if (TxtGlueCharIsVirtual(eventP->data.keyDown.modifiers, c)) {
          if (c == vchrPageUp) {
            PageScroll(FrmGetActiveForm(), AboutField,
	      AboutScrollBar, winUp);
	    handled = true;
	  } else if (c == vchrPageDown) {
	    PageScroll(FrmGetActiveForm(), AboutField,
	      AboutScrollBar, winDown);
	    handled = true;
          }
	//}
      }
      break;
  
    case sclRepeatEvent:
      ScrollLines(FrmGetActiveForm(), AboutField, AboutScrollBar,
	eventP->data.sclRepeat.newValue - 
	eventP->data.sclRepeat.value, false);
      break;
      
    case ctlSelectEvent:
      switch (eventP->data.ctlSelect.controlID) {
        case AboutOKButton:
	  FrmGotoForm(gPrefs.form);
          handled = true;
          break;
      }

    case frmCloseEvent:
      /* Deallocate Fields' Text Memory */
      SetFieldText(AboutField, NULL, false, true);
      FldEraseField(GetObjectFromActiveForm(AboutField));
      handled = false;
      break;
  
    default:
      break;
  }

  return (handled);
}
