/*****************************************************************************
 *
 * $RCSfile: ApproachForm.c,v $
 *
 * GPS4Palm Approach Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004-11-25 19:47:37 $
 *
 * $Revision: 1.2 $
 *
 * $Log: ApproachForm.c,v $
 * Revision 1.2  2004-11-25 19:47:37  mp
 * updated descriptions
 *
 * Revision 1.1  2004/11/25 19:44:05  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>			
#include "ResourceDefines.h"
#include "Serial.h"
#include "ApproachForm.h"
#include "Utils.h"
#include "Data.h"
#include "GPS.h"
#include "fp.h"
#include "geo.h"
#include "common.h"

/* Global Variables */
extern PrefsType	gPrefs;			/* Preferences data structure */
extern DmOpenRef	gWaypointDB;		/* Wpt. Data Base reference */
extern GPSType		gGPSData;		/* GPS Data */


/* Static Functions */
static void ApproachFormInit(FormPtr form);
static void ApproachFormDeinit();


/*****************************************************************************
 * FUNCTION:	ApproachFormInit
 *
 * DESCRIPTION: Approach Form init function
 *
 * PARAMETERS:	form		- Form Pointer (unused)
 *
 * RETURNED:    %
 *
 ****************************************************************************/
static 
void ApproachFormInit(FormPtr form)
{
#pragma unused(form)
  // warning-- don't do any drawing in this routine.
  // Also, don't call FrmSetFocus from here (it must be called *after*
  // FrmDrawForm)
}


/*****************************************************************************
 * FUNCTION:	ApproachFormDeinit
 *
 * DESCRIPTION: Approach Form deinit function
 *
 * PARAMETERS:	%
 *
 * RETURNED:    %
 *
 ****************************************************************************/
static 
void ApproachFormDeinit(void)
{

}


/****************************************************************************/
/**
 * \brief	Approach Form event handler.
 *
 *		The Approach Form shows the Active Waypoint's ID
 *		and the current distance to the waypoint.
 *		If the GPS data is valid, the distance display is updated
 *		periodically. If enabled in the application preferences,
 *		an alarm sound is generated periodically.
 *
 * \param	eventP	Pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean ApproachFormHandleEvent(EventPtr eventP)
{
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  UInt16		rec;			/* record index */
  static double		wpt_lat;		/* waypoint latitude */
  static double		wpt_lon;		/* waypoint longitude */
  double		wpt_dst;		/* distance to waypoint */
  Boolean		handled = false;	/* flag: event handled */
  FormPtr		frmP;			/* form ptr */
  Char			str[10];		/* string buffer */
  Err			err;			/* error code */
	
  switch (eventP->eType) {
    case frmOpenEvent:
      frmP = FrmGetActiveForm();
      ApproachFormInit(frmP);

      /* find active waypoint */
      err = DmFindRecordByID(gWaypointDB, gPrefs.act_wpt.waypointID,
	&rec);
	
      if (!err) {
	/* active waypoint found */
	waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
	packed_waypoint =
	  (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
	UnpackWaypoint(&waypoint, packed_waypoint);

        /* get waypoint position */
	wpt_lat = semi2deg(waypoint.posn.lat);
	wpt_lon = semi2deg(waypoint.posn.lon);
	
	/* display WPT ID */
	MemSet(str, 7, 0);
	StrNCopy(str, (Char *)&waypoint.ident, 6);
	SetFieldText(ApproachWptField, str, false, true);
	
	/* display waypoint distance */
	wpt_dst = gc_dist_sphere(gGPSData.lat, gGPSData.lon,
          wpt_lat, wpt_lon) * calcR(gGPSData.lat);
	format_number(wpt_dst, 1, str);
	SetFieldText(ApproachDstField, str, false, true);
	
	MemHandleUnlock(waypointDBEntry);
      }
      
      FrmDrawForm(frmP);

      handled = true;
      break;

    case ctlSelectEvent:
      if (eventP->data.ctlSelect.controlID == ApproachOKButton) {
        
	/* return to previous form */
	FrmGotoForm(gPrefs.form);
	handled = true;
      }
      break;

    case nilEvent:
      handled = true;

      /* throw away anything in the buffer-- we want fresh data */
      DoReceiveFlush(gPortID, 1);

      /*
       * Get GPS Data and Waypoint Distance
       */
      ReadFromGPS();
      
      if (gGPSData.valid) {
	/* get distance to active waypoint */
	wpt_dst = gc_dist_sphere(gGPSData.lat, gGPSData.lon,
          wpt_lat, wpt_lon) * calcR(gGPSData.lat);
	format_number(wpt_dst, 1, str);
	SetFieldText(ApproachDstField, str, false, true);      
      }

      if (gPrefs.approach_sound) 
	SndPlaySystemSound(sndAlarm);
      
      break;
          
    case frmCloseEvent:
      /* Deallocate Fields' Text Memory */
      SetFieldText(ApproachWptField, NULL, false, true);
      SetFieldText(ApproachDstField, NULL, false, true);
		
      ApproachFormDeinit();
      
      handled = false;
      break;

  } /* switch (eventP->eType) */
  return(handled);
}
