/*****************************************************************************
 *
 * $RCSfile: Clip.c,v $
 *
 * GPS4Palm Clipboard Functions
 *
 * This program is Copyright (C) 07/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-02 17:29:25 $
 *
 * $Revision: 1.4 $
 *
 * $Log: Clip.c,v $
 * Revision 1.4  2005-04-02 17:29:25  mp
 * moved Position2Geodb() to Data.c/Data.h
 *
 * Revision 1.3  2005/04/02 10:10:08  mp
 * modified Position2Clipboard() and Position2Geodb() to return 'handled'
 *
 * Revision 1.2  2005/04/02 07:24:44  mp
 * added Position2Geodb()
 *
 * Revision 1.1  2005/03/25 13:40:48  mp
 * initial version
 *
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "common.h"
#include "Clip.h"
#include "Data.h"
#include "geo.h"	/* deg_to_str() */


/****************************************************************************/
/**
 * \brief	Add Time/Date and Position (in user selected unit)
 *		to Clipboard
 *
 * \param	gps_data	GPS Data (Time of Fix, Position, ...)
 * \param	units		User selected units
 *
 * \return	True if GPS data is valid.
 ****************************************************************************/
Boolean Position2Clipboard(GPSType gps_data, UnitsType units)
{
  DateTimeType	dt;			/* date/time structure */
  Char		lat_str[20];		/* latitude/easting */
  Char		lon_str[20];		/* longitude/northing */
  Char		field_str[5];		/* UTM zone and field */
  Char		clp_str[61];		/* time&date: 16 lat&lon: 45 */
    
  if (!gps_data.valid)
    return false;

  /* get time and date (output format: yy-mm-dd hh:mm) */
  TimSecondsToDateTime(gps_data.time, &dt);
  
  /* write date to string */
  DateToAscii(dt.month, dt.day, dt.year,
    dfYMDWithDashes,
    clp_str);

  /* append space */
  StrCat(clp_str, " ");
  
  /* write time to string */
  TimeToAscii(dt.hour, dt.minute,
    tfColon24h,
    &clp_str[StrLen(clp_str)]);
  
  /* append space */
  StrCat(clp_str, " ");
  
  /* convert position to string */
  deg_to_str(gps_data.lat, gps_data.lon, units.pos_unit,
             lat_str, lon_str, field_str, NULL, NULL);
  
  /* append UTM zone and field (or empty string) */
  StrCat(clp_str, field_str);
  
  if (StrLen(field_str) == 0)
    /* append space */
    StrCat(clp_str, " ");

  if (units.pos_unit == POS_UTM)
    StrCat(clp_str, "E");
  
  /* append Latitude or Easting */
  StrCat(clp_str, lat_str);
  
  /* append space */
  StrCat(clp_str, " ");

  if (units.pos_unit == POS_UTM)
    StrCat(clp_str, "N");

  /* append Longitude or Northing */
  StrCat(clp_str, lon_str);

  ClipboardAddItem(clipboardText, clp_str, StrLen(clp_str));
  
  return true;
}
