/*****************************************************************************
 *
 * $RCSfile: Data.c,v $
 *
 * GPS4Palm Data Base Functions
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-02 17:55:16 $
 *
 * $Revision: 1.23 $
 *
 * $Log: Data.c,v $
 * Revision 1.23  2005-04-02 17:55:16  mp
 * removed test code
 *
 * Revision 1.22  2005/04/02 17:28:21  mp
 * added CompareStringFunc() and Position2Geodb(), removed CompareIDFunc()
 *
 * Revision 1.21  2005/02/05 17:33:42  mp
 * added DeleteTracks()
 *
 * Revision 1.20  2004/12/12 21:01:08  mp
 * fixed track log wrap around
 *
 * Revision 1.19  2004/12/10 19:52:23  mp
 * replaced DmGet1Resource by DmGetResource
 *
 * Revision 1.18  2004/12/09 17:32:44  mp
 * replaced default track comment string by string resource,
 * implemented wrapping when track log database full
 *
 * Revision 1.17  2004/12/08 20:51:07  mp
 * added TrackIntervalCheck()
 *
 * Revision 1.16  2004/12/07 18:51:09  mp
 * added TrackWriteLog()
 *
 * Revision 1.15  2004/12/04 09:40:49  mp
 * added gTrackHdrDB, StoreTrkHdr(), StoreTrkPoint(), TrackDBName(), and
 * TrackIdent()
 *
 * Revision 1.14  2004/11/27 10:16:08  mp
 * replaced ErrFatalDisplay() by Die() for memory allocation failures
 *
 * Revision 1.13  2004/11/25 20:04:45  mp
 * added NotifyApproach()
 *
 * Revision 1.12  2004/11/25 16:59:22  mp
 * added CheckApproach()
 *
 * Revision 1.11  2004/11/24 21:03:43  mp
 * modified error handling in UpdateActWpt()
 *
 * Revision 1.10  2004/11/23 17:53:09  mp
 * removed unused variable
 *
 * Revision 1.9  2004/11/20 15:53:12  mp
 * added UpdateActWpt()
 *
 * Revision 1.8  2004/11/17 19:47:10  mp
 * code cleanup
 *
 * Revision 1.7  2004/11/16 21:05:53  mp
 * modified route/waypoint record write functions: pass record index instead
 * of handle, this allows to call DmGetRecord() / DmReleaseRecord
 *
 * Revision 1.6  2004/11/16 19:21:40  mp
 * added InsertRouteWpt(), DelRouteWpt(), and MoveRouteWpt(),
 * modified PackWaypoint() (removed debug output)
 *
 * Revision 1.5  2004/06/23 17:58:14  mp
 * Added comments, removed example functions from Sales application.
 *
 * Revision 1.4  2004/04/21 18:34:36  mp
 * modified macro DEG2SEMI()
 *
 * Revision 1.3  2004/03/12 21:16:26  mp
 * added InitializeWaypoints(void)
 *
 * Revision 1.2  2004/03/10 17:23:37  mp
 * modified due to changes in Data.h
 *
 * Revision 1.1  2004/03/07 14:28:49  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "Garmin.h"
#include "Data.h"
#include "geo.h"		/* gc_dist_sphere() */
#include "common.h"
#include "gpslib.h"		/* Palm2GarminTimeDiff */
#include "Utils.h"
#include "ResourceDefines.h"

/** Convert degrees (double) to semicircles (fix-point) */
#define DEG2SEMI(x)	((G_long)(x * ((1UL<<31)/180.0)))

/* Global Variables */
DmOpenRef	gWaypointDB;		/**< Waypoint Database Reference */
DmOpenRef	gRouteDB;		/**< Route Database Reference */
DmOpenRef	gTrackHdrDB;		/**< Track Header Database Reference */
DmOpenRef	gTrackDB = 0;		/**< Track Database Reference */

extern PrefsType	gPrefs;		/* Preferences data structure */
extern GPSType		gGPSData;	/* GPS Data */

/*
 * Track Log time interval in seconds
 *   (must match TrkIntTimeList in GPS4Palm.rcp)
 */
static const UInt16 log_interval_time[] = {
  1, 2, 5, 10, 20, 30, 60, 120
};

/*
 * Track Log distance interval in meters
 *   (must match TrkIntDstList in GPS4Palm.rcp)
 */
static const UInt16 log_interval_dist[] = {
  10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000
};

/* Memo Name (keep it short, this has to be entered by the user in GeoDB) */
static const Char memo_name[] = "Gps4";

/* Location Name (has to be edited anyway) */
static const Char location_name[] = "_GPS4Palm";


/*****************************************************************************
 * FUNCTION:	CheckDatabases
 *
 * DESCRIPTION: Check lock state of databases.
 *
 * PARAMETERS:	%
 *
 * RETURNED:    %
 ****************************************************************************/
#ifdef DEBUG_BUILD
void CheckDatabases(void)
{
  UInt8		highest;
  UInt32	count;
  UInt32	busy;
  
  return;
  DmGetDatabaseLockState(gWaypointDB, &highest, &count, &busy);
  if (highest || busy)
    ErrFatalDisplay("Waypoint db not quiescent");
  DmGetDatabaseLockState(gRouteDB, &highest, &count, &busy);
  if (highest || busy)
    ErrFatalDisplay("Route db not quiescent");
  DmGetDatabaseLockState(gTrackDB, &highest, &count, &busy);
  if (highest || busy)
    ErrFatalDisplay("Track db not quiescent");
}
#endif


/****************************************************************************/
/**
 * \brief	Compare Strings
 *
 *
 * PARAMETERS:	see DmComparF() in PalmOS Reference
 *
 * \return
 *		- -1 if str1 < str2
 *		-  1 if str1 > str2
 *		-  0 if str1 = str2
 ****************************************************************************/
Int16 CompareStringFunc(const Char *str1, const Char *str2, Int16 i, 
  SortRecordInfoPtr s1, SortRecordInfoPtr s2, MemHandle appInfoH)
{
  return StrCompare(str1, str2);
}


/****************************************************************************/
/**
 * \brief	Compare Waypoint Identifiers
 *
 * \note	The identifiers are not necessarily null-terminated,
 *		  therefore StrCompare() cannot be used.
 *
 * PARAMETERS:	see DmComparF() in PalmOS Reference
 *
 * \return
 *		- -1 if ident1 < ident2
 *		-  1 if ident1 > ident2
 *		-  0 if ident1 = ident2
 ****************************************************************************/
Int16 CompareWayptIdent(waypoint_t *rec1, packed_waypoint_t *rec2,
  Int16 i, SortRecordInfoPtr s1, SortRecordInfoPtr s2, MemHandle appInfoH)
{
  #pragma unused(i, s1, s2, appInfoH)
  UInt16	n;
  
  for (n=0; n<6; n++) {
    if (rec1->ident[n] < rec2->ident[n]) {
      return -1;
    } else if (rec1->ident[n] > rec2->ident[n]) {
      return 1;
    } else if ((rec1->ident[n] == '\0') && (rec2->ident[n] == '\0')) {
      break;
    }
  }
  
  /* Identifiers are equal */
  return 0;
}


/****************************************************************************/
/**
 * \brief	Compare Route Numbers
 *
 * \note	see DmComparF() in PalmOS Reference for parameters
 *
 * \return
 *		- -1 if nmbr1 < nmbr2
 *		-  1 if nmbr1 > nmbr2
 *		-  0 if nmbr1 = nmbr2
 ****************************************************************************/
Int16 CompareRteNmbr(route_t *rec1, route_t *rec2,
  Int16 i, SortRecordInfoPtr s1, SortRecordInfoPtr s2, MemHandle appInfoH)
{
  #pragma unused(i, s1, s2, appInfoH)
  
  if (rec1->nmbr < rec2->nmbr) {
    return -1;
  } else if (rec1->nmbr > rec2->nmbr) {
    return 1;
  }
  
  /* Identifiers are equal */
  return 0;
}



/****************************************************************************/
/**
 * \brief	Pack Waypoint and store it in Database
 *
 * \param	waypoint		ptr to waypoint structure
 * \param	index			waypoint record index
 *
 ****************************************************************************/
void PackWaypoint(waypoint_t *waypoint, UInt16 index)
{
  MemHandle	waypointDBEntry;	/* waypoint record handle */
  UInt16	length = 0;		/* record length */
  Char		*s;			/* record ptr */
  UInt16	offset = 0;		/* offset into record */
  
  /* get waypoint record */
  waypointDBEntry = DmGetRecord(gWaypointDB, index);
  
  /* figure out necessary size */
  length = (UInt16) (	sizeof(waypoint->ident) +
			sizeof(waypoint->posn) +
			sizeof(waypoint->dst) +
			sizeof(waypoint->smbl) +
			sizeof(waypoint->dspl) +
			StrLen(waypoint->cmnt) +
			1 /* 1 for string terminator */
		    );

  /* resize the MemHandle */
  if (MemHandleResize(waypointDBEntry, length) == 0) {
    /* copy the fields */
    s = (Char *) MemHandleLock(waypointDBEntry);
    offset = 0;
    DmWrite(s, offset, &waypoint->ident,
      sizeof(waypoint->ident));
    offset += sizeof(waypoint->ident);
    DmWrite(s, offset, &waypoint->posn.lat,
      sizeof(waypoint->posn.lat));
    offset += sizeof(waypoint->posn.lat);
    DmWrite(s, offset, &waypoint->posn.lon,
      sizeof(waypoint->posn.lon));
    offset += sizeof(waypoint->posn.lon);
    DmWrite(s, offset, &waypoint->dst,
      sizeof(waypoint->dst));
    offset += sizeof(waypoint->dst);
    DmWrite(s, offset, &waypoint->smbl,
      sizeof(waypoint->smbl));
    offset += sizeof(waypoint->smbl);
    DmWrite(s, offset, &waypoint->dspl,
      sizeof(waypoint->dspl));
    offset += sizeof(waypoint->dspl);
    DmStrCopy(s, offset, waypoint->cmnt);
    MemHandleUnlock(waypointDBEntry);
  } else {
    Die("Cannot resize waypoint DB record!");
  }
  
  /* release record */
  DmReleaseRecord(gWaypointDB, index, true /* dirty */);
}


/****************************************************************************/
/**
 * \brief	Store Route Header in Database
 *
 * \param	route		ptr to route structure
 * \param	index		route record index
 *
 ****************************************************************************/
void PackRouteHdr(route_t *route, UInt16 index)
{
  MemHandle	routeDBEntry;	/* route record handle */
  UInt16	length = 0;	/* record length */
  Char		*s;		/* record ptr */
  UInt16	offset = 0;	/* offset into record */
  
  routeDBEntry = DmGetRecord(gRouteDB, index);
  
  /* figure out necessary size */
  length = (UInt16) (	sizeof(route->nmbr) +
			sizeof(route->cmnt) +
			sizeof(route->items) +
			route->items * sizeof(route->wpt_rec_id[0])
		    );

  /* resize the MemHandle */
  if (MemHandleResize(routeDBEntry, length) == 0) {
    /* copy the fields */
    s = (Char *)MemHandleLock(routeDBEntry);
    offset = 0;

    DmWrite(s, offset, &route->nmbr,
      sizeof(route->nmbr));
    offset += sizeof(route->nmbr);
    DmStrCopy(s, offset, route->cmnt);
    offset += sizeof(route->cmnt);
    DmWrite(s, offset, &route->items,
      sizeof(route->items));
    MemHandleUnlock(routeDBEntry);
  } else {
    Die("Cannot resize route DB record (header)!");
  }

  /* release record */
  DmReleaseRecord(gRouteDB, index, true /* dirty */);
}


/****************************************************************************/
/**
 * \brief	Store Route Waypoint in Database (append)
 *
 * \param	wpt_rec_id	waypoint record ID
 * \param	index		route record index
 *
 ****************************************************************************/
void PackRouteWpt(UInt32 wpt_rec_id, UInt16 index)
{
  MemHandle	routeDBEntry;	/* route record handle */
  UInt16	items;		/* number of waypoints on route */
  UInt16	length = 0;	/* record length */
  UInt16	offset = 0;	/* offset into record */
  route_t	*route;		/* record ptr */
  
  /* get record handle */
  routeDBEntry = DmGetRecord(gRouteDB, index);
 
  /* figure out necessary size: original size + new wpt_rec_id */
  route = MemHandleLock(routeDBEntry);
  items = route->items + 1;
  MemHandleUnlock(routeDBEntry);
  
  length = (UInt16) (	sizeof(route->nmbr) +
			sizeof(route->cmnt) +
			sizeof(route->items) +
			(sizeof(route->wpt_rec_id[0]) * items)
		    );
    
  /* resize the MemHandle */
  if (MemHandleResize(routeDBEntry, length) == 0) {
    /* copy the fields */
    route = MemHandleLock(routeDBEntry);
    offset = length - sizeof(route->wpt_rec_id[0]);
    DmWrite(route, offset, &wpt_rec_id,
      sizeof(wpt_rec_id));
    
    DmWrite(route, OffsetOf(route_t, items), &items, sizeof(route->items));
    MemHandleUnlock(routeDBEntry);
  } else {
    Die("Cannot resize route DB record (waypoints)!");
  }

  /* release record */
  DmReleaseRecord(gRouteDB, index, true /* dirty */);
}


/****************************************************************************/
/**
 * \brief	Store Route Waypoint in Database (insert after selection)
 *
 * \param	rec		route record index
 * \param	wpt_rec_id	waypoint record ID
 * \param	wpt_index	waypoint array index
 *
 ****************************************************************************/
void InsertRouteWpt(UInt16 rec, UInt32 wpt_rec_id, UInt16 wpt_index)
{
  MemHandle	routeDBEntry;	/* route record handle */
  route_t	*route;		/* route record ptr */
  UInt16	items;		/* no. of waypoints */
  UInt16	length;		/* length of record */
  UInt16	offset;		/* offset within record */
  UInt16	i;		/* loop index */
  
  /* get record handle */
  routeDBEntry = DmGetRecord(gRouteDB, rec);
  
  /* figure out necessary size: original size + new wpt_rec_id */
  route = MemHandleLock(routeDBEntry);
  items = route->items + 1;
  MemHandleUnlock(routeDBEntry);
  
  length = (UInt16) (	sizeof(route->nmbr) +
			sizeof(route->cmnt) +
			sizeof(route->items) +
			(sizeof(route->wpt_rec_id[0]) * items)
		    );
  
  /* set offset to last item */
  offset = length - sizeof(route->wpt_rec_id[0]);
  
  /* resize the MemHandle */
  if (MemHandleResize(routeDBEntry, length) == 0) {
    /* lock record handle */
    route = MemHandleLock(routeDBEntry);
    
    /* move array items down */
    for (i=items-1; i>wpt_index+1; i--) {
      DmWrite(route, offset, &route->wpt_rec_id[i-1],
        sizeof(route->wpt_rec_id[i-1]));
      offset -= sizeof(route->wpt_rec_id[0]);
    }

    /* write new item */
    DmWrite(route, offset, &wpt_rec_id, sizeof(wpt_rec_id));
   
    /* store new number of items */
    DmWrite(route, OffsetOf(route_t, items), &items, sizeof(items));

    /* unlock record handle */
    MemHandleUnlock(routeDBEntry);

  } else {
    Die("Cannot resize route DB record (waypoints)!");
  }
  
  /* release record */
  DmReleaseRecord(gRouteDB, rec, true /* dirty */);
}


/****************************************************************************/
/**
 * \brief	Delete Route Waypoint from Database
 *
 * \param	rec		route record index
 * \param	wpt_index	waypoint array index
 *
 ****************************************************************************/
void DelRouteWpt(UInt16 rec, UInt16 wpt_index)
{
  MemHandle	routeDBEntry;	/* route record handle */
  route_t	*route;		/* route record ptr */
  UInt16	i;		/* loop index */
  UInt16	offset;		/* offset within record */
  UInt16	items;		/* new no. of items */
  UInt16	length = 0;	/* length of record */
  
  /* get record handle */
  routeDBEntry = DmGetRecord(gRouteDB, rec);

  /* offset to waypoint record ID to be deleted */
  offset = sizeof(route->nmbr) +
           sizeof(route->cmnt) +
	   sizeof(route->items) +
	   sizeof(route->wpt_rec_id[0]) * wpt_index;
    
  /* lock record */
  route = MemHandleLock(routeDBEntry);
  
  /* move array items up */
  for (i=wpt_index; i<route->items-1; i++) {
    DmWrite(route, offset, &route->wpt_rec_id[i+1],
      sizeof(route->wpt_rec_id[i+1]));
    offset += sizeof(route->wpt_rec_id[0]);    
  }
  
  /* calculate offset to route->items */
  offset = sizeof(route->nmbr) +
           sizeof(route->cmnt);
  
  /* new no. of items */
  items = route->items - 1;
  
  /* write new no. of items to record */
  DmWrite(route, offset, &items, sizeof(items));
  
  /* unlock record */
  MemHandleUnlock(routeDBEntry);
  
  /* release record */
  DmReleaseRecord(gRouteDB, rec, true /* dirty */);
  
  /* resize the MemHandle */
  length = (UInt16) (	sizeof(route->nmbr) +
			sizeof(route->cmnt) +
			sizeof(route->items) +
			(sizeof(route->wpt_rec_id[0]) * items)
		    );

  MemHandleResize(routeDBEntry, length);
}


/****************************************************************************/
/**
 * \brief	Move Route Waypoint in Database
 *
 * \param	rec		route record index
 * \param	wpt_index	waypoint array index
 * \param	up		direction: true - up / false - down
 *
 ****************************************************************************/
void MoveRouteWpt(UInt16 rec, UInt16 wpt_index, Boolean up)
{
  MemHandle	routeDBEntry;	/* route record handle */
  route_t	*route;		/* route record ptr */
  UInt16	offset;		/* offset within record */
  UInt32	wpt_rec_id;	/* unique waypoint record ID */

  /* get record handle */
  routeDBEntry = DmGetRecord(gRouteDB, rec);

  /* offset to waypoint record ID to be moved */
  offset = sizeof(route->nmbr) +
           sizeof(route->cmnt) +
	   sizeof(route->items) +
	   sizeof(route->wpt_rec_id[0]) * wpt_index;
    
  /* lock record */
  route = MemHandleLock(routeDBEntry);
  
  if (up) {
    wpt_rec_id = route->wpt_rec_id[wpt_index-1];
    DmWrite(route, offset-sizeof(wpt_rec_id),
      &route->wpt_rec_id[wpt_index], sizeof(wpt_rec_id));
    DmWrite(route, offset,
      &wpt_rec_id, sizeof(wpt_rec_id));
  } else {
    wpt_rec_id = route->wpt_rec_id[wpt_index+1];
    DmWrite(route, offset+sizeof(wpt_rec_id),
      &route->wpt_rec_id[wpt_index], sizeof(wpt_rec_id));
    DmWrite(route, offset,
      &wpt_rec_id, sizeof(wpt_rec_id));
  }
  
  /* unlock record */
  MemHandleUnlock(routeDBEntry);

  /* release record */
  DmReleaseRecord(gRouteDB, rec, true /* dirty */);
}

/**********************************************************************/
/**
 * Unpack waypoint retrieved from database.
 *
 * \note	The packed_waypoint must remain locked while
 *		the waypoint is in use.
 * \param	waypoint	ptr to unpacked waypoint structure
 * \param	packed_waypoint	ptr to packed waypoint structure
 *********************************************************************/
void UnpackWaypoint(waypoint_t *waypoint,
  const packed_waypoint_t *packed_waypoint)
{
  StrNCopy(waypoint->ident, packed_waypoint->ident, 6);
  waypoint->posn.lat = packed_waypoint->posn.lat;
  waypoint->posn.lon = packed_waypoint->posn.lon;
  waypoint->unused = 0;
  waypoint->dst = packed_waypoint->dst;
  waypoint->smbl = packed_waypoint->smbl;
  waypoint->dspl = packed_waypoint->dspl;
  StrCopy(waypoint->cmnt, packed_waypoint->cmnt);
}

/**********************************************************************/
/**
 * Store track header in database.
 *
 * \param	trkhdr		ptr to track header structure
 * \param	index		track header record index
 **********************************************************************/
void StoreTrkHdr(trk_header_t *trkhdr, UInt16 index)
{
  MemHandle	trkhDBEntry;	/* track point record handle */
  UInt16	length = 0;	/* record length */
  UInt16	offset = 0;	/* offset into record */
  trk_header_t	*trkP;		/* record ptr */  

  length = (UInt16) (	sizeof(trkhdr->dspl) +
			sizeof(trkhdr->color) +
			sizeof(trkhdr->trk_ident)
		    );  

  /* get record handle */
  trkhDBEntry = DmGetRecord(gTrackHdrDB, index);  
  
  /* resize the MemHandle */
  if (MemHandleResize(trkhDBEntry, length) == 0) {
    /* copy the fields */
    trkP = MemHandleLock(trkhDBEntry);
    offset = 0;
    DmWrite(trkP, offset, &trkhdr->dspl,
      sizeof(trkhdr->dspl));
    offset += sizeof(trkhdr->dspl);
    DmWrite(trkP, offset, &trkhdr->color,
      sizeof(trkhdr->color));
    offset += sizeof(trkhdr->color);
    DmWrite(trkP, offset, &trkhdr->trk_ident,
      sizeof(trkhdr->trk_ident));
    MemHandleUnlock(trkhDBEntry);
  } else {
    Die("Cannot resize track header DB record!");
  }
  
  /* release record */
  DmReleaseRecord(gTrackHdrDB, index, true /* dirty */);
}


/**********************************************************************/
/**
 * Store track point in database.
 *
 * \param	trkpoint	ptr to track point structure
 * \param	index		track point record index
 **********************************************************************/
void StoreTrkPoint(trk_point_t *trkpoint, UInt16 index)
{
  MemHandle	trkpDBEntry;	/* track point record handle */
  UInt16	length = 0;	/* record length */
  UInt16	offset = 0;	/* offset into record */
  trk_point_t	*trkP;		/* record ptr */

  length = (UInt16) (	sizeof(trkpoint->posn) +
			sizeof(trkpoint->time) +
			sizeof(trkpoint->new_trk)
		    );  

  /* get record handle */
  trkpDBEntry = DmGetRecord(gTrackDB, index);
  
  /* resize the MemHandle */
  if (MemHandleResize(trkpDBEntry, length) == 0) {
    /* copy the fields */
    trkP = MemHandleLock(trkpDBEntry);
    offset = 0;
    DmWrite(trkP, offset, &trkpoint->posn,
      sizeof(trkpoint->posn));
    offset += sizeof(trkpoint->posn);
    DmWrite(trkP, offset, &trkpoint->time,
      sizeof(trkpoint->time));
    offset += sizeof(trkpoint->time);
    DmWrite(trkP, offset, &trkpoint->new_trk,
      sizeof(trkpoint->new_trk));
    MemHandleUnlock(trkpDBEntry);
  } else {
    Die("Cannot resize track point DB record!");
  }
  
  /* release record */
  DmReleaseRecord(gTrackDB, index, true /* dirty */);
}


/****************************************************************************/
/**
 * \brief	Generate Track Point Data Base Name
 *
 * \param	dbname		pointer to destination string buffer
 * \param	trackID		track header unique record ID		
 *
 ****************************************************************************/
void TrackDBName(Char *dbname, UInt32 trackID)
{
  UInt16	i;			/* loop index */
  
  /* copy fixed part of track data base name */
  StrPrintF(dbname, "%s", kTrkDBName);
  
  for (i=0; i<6; i++) {
    StrPrintF(&dbname[StrLen(dbname)], "%c", nibble2hex((trackID >> 20) & 0xF));
    trackID <<= 4;
  }
}

/****************************************************************************/
/**
 * \brief	Generate Track Identifier from current Time and Date
 *
 * \param	ident		pointer to destination string buffer
 *
 ****************************************************************************/
void TrackIdent(Char *ident)
{
  UInt32	seconds;		/* current time */
  DateTimeType	dt;			/* date/time structure */
  MemHandle	strH;			/* string handle */ 
  MemPtr	strP;			/* string ptr */


  /*
   * create comment: <Track yy-mm-dd hh:mm>
   */
  MemSet(ident, sizeof(ident), 0);
  
  /* get string from resource */
  strH = DmGetResource(strRsc, TrackStr);
  if (strH != NULL) {
    /* string handle valid, lock it */
    strP = MemHandleLock(strH);
  
    /* copy string */
    StrCopy(ident, strP);

    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  }
  
  /* get current time and date */
  seconds = TimGetSeconds();
  TimSecondsToDateTime(seconds, &dt);
  
  /* write date to string */
  DateToAscii(dt.month, dt.day, dt.year, dfYMDWithDashes,
    &ident[StrLen(ident)]);

  /* append space */
  StrCat(ident, " ");
  
  /* write time to string */
  TimeToAscii(dt.hour, dt.minute, tfColon24h,
    &ident[StrLen(ident)]);
  
  /* append bracket */
  StrCat(ident, ">");
}


/****************************************************************************/
/**
 * \brief	Delete all track point databases
 *
 ****************************************************************************/
void DeleteTracks(void)
{
  DmSearchStateType	state;		/* DmSearchState */
  UInt16		card;		/* card no. */
  LocalID		currentDB = 0;	/* local ID of current DB */
  Err			err;		/* error code */

  do {
    err = DmGetNextDatabaseByTypeCreator(
      true /* newSearch */,
      &state /* stateInfoP */,
      kTrkDBType /* type */,
      kCreatorId /* creator */,
      false /* onlyLatestVers */,
      &card /* cardNoP */,
      &currentDB /* dbIDP */
    );
    
    if (!err) {
      DmDeleteDatabase(card, currentDB);
    }
  } while (!err);
}


/****************************************************************************/
/**
 * \brief	Add Position to Memo for Import to GeoDB
 *
 * \param	gps_data	GPS Data (Time of Fix, Position, ...)
 *
 * \return	True if GPS data is valid.
 *
 * \note	GeoDB expects the following format in the Memo:
 *
 *		< memo_name >\n
 *		< location_name >,< latitude >,< N|S >,< longitude >,< E|W >,
 *		< elevation >,< gmt_offset >,< E|W >,< daylight_time >
 * 
 *		Example:
 * 
 *		AA
 *		Adana,37:0.6,N,35:16.8,E,,2,E,0
 *
 +		(GPS4Palm only knows the position)
 *
 ****************************************************************************/
Boolean Position2Geodb(GPSType gps_data)
{
  Char		clp_str[100];		/* clipboard string buffer */
  Char		lat_str[20];		/* latitude string buffer */
  Char		lon_str[20];		/* longitude string buffer */
  DmOpenRef	memoDB;
  MemHandle	recH;
  MemPtr	recP;
  UInt16	index;

  if (!gps_data.valid)
    return false;
 
  /* convert double lat/lon to string */
  deg_to_geodb_str(gps_data.lat, gps_data.lon, lat_str, lon_str);
  
  /*
   * TODO:
   * Elevation (or "Altitude") is actually known, but not a 
   * Member of GPSType yet.
   */
  
  /* format output string */
  StrPrintF(clp_str, "%s\n%s,%s,%s,,,,,",
    memo_name,
    location_name,
    lat_str,
    lon_str);
  
/*
 * Note: optionally we could add it to clipboard  
  ClipboardAddItem(clipboardText, clp_str, StrLen(clp_str));
*/  

  /* Open Memo DB */
  memoDB = DmOpenDatabaseByTypeCreator('DATA', sysFileCMemo, dmModeReadWrite);
  
  /* Find sort position of new record in memo DB */  
  index = DmFindSortPosition(memoDB, clp_str, 0 /* newRecordInfo */,
    (DmComparF *)CompareStringFunc /* DmComparF *compar */, 0 /* other */);

  /* Create new memo record */
  recH = DmNewRecord(memoDB, &index, StrLen(clp_str) + 1);
  recP = MemHandleLock(recH);
  DmStrCopy(recP, 0, clp_str);
  MemHandleUnlock(recH);
  DmReleaseRecord(memoDB, index, true);

  /* Close Memo DB */
  DmCloseDatabase(memoDB);

  return true;
}


/****************************************************************************/
/**
 * \brief	Update active (route) waypoint according to distance to
 *		current and next waypoint.
 *
 *		The current implementation switches to the next route
 *		waypoint even if the current waypoint is just "passed by",
 *		i.e. the distance to the current waypoint is increasing
 *		while the distance to the next waypoint is decreasing. 
 *
 *		A different implementation could require the current
 *		waypoint to be "visited" with the proximity information
 *		stored in the waypoint record as the criterion. 
 *
 * \param	lat		current latitude
 * \param	lon		current longitude
 * \param	init		initialize function (after changing route)	
 *
 * \return
 *		- true  if next waypoint has been select as active waypoint
 *		- false if current waypoint is still active waypoint
 *
 * \note	After changing the active route, the finction must be called
 *		with init=true (lat/lon may be invalid). This marks the
 *		distance values stored previously invalid. The next call
 *		which must occur with a valid position updates the stored
 *		distance values. Any subsequent call will then provide
 *		a correct waypoint selection.
 * 
 ****************************************************************************/
Boolean UpdateActWpt(double lat, double lon, Boolean init)
{
  route_t		*route;			/* route */
  MemHandle		routeDBEntry;		/* route record handle */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  UInt16		rec;			/* record index */
  static Boolean	init_ok = false;	/* flag: initialization o.k. */
  static double		dst_act_old;		/* old dist. to active wpt. */
  static double		dst_next_old;		/* dist. to next wpt. */
  double		dst_act;		/* dist. to active waypoint */
  double		dst_next;		/* dist. to next waypoint */
  double		wpt_lat;		/* waypoint longitude */
  double		wpt_lon;		/* waypoint latitude */
  Boolean		updated = false;	/* flag: act. wpt. updated */
  Err			err;			/* error code */
  
  if (init) {
    init_ok = false;
    return false;
  }
  
  /* get active route */
  err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID,
    &rec);

  if (!err) {
    /* active route found */
    routeDBEntry = DmQueryRecord(gRouteDB, rec);
    
    /* lock route record */
    route = (route_t *)MemHandleLock(routeDBEntry);
    
    if (gPrefs.act_rte.wpt_index == route->items-1) {
      /* last waypoint already active */
      MemHandleUnlock(routeDBEntry);
      return false;
    }
    
    /* get active waypoint */
    err = DmFindRecordByID(gWaypointDB,
      route->wpt_rec_id[gPrefs.act_rte.wpt_index],
      &rec);

    if (!err) {
      /* active waypoint found */
      waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
      packed_waypoint =
	(packed_waypoint_t *)MemHandleLock(waypointDBEntry);
      UnpackWaypoint(&waypoint, packed_waypoint);

      wpt_lat = semi2deg(waypoint.posn.lat);
      wpt_lon = semi2deg(waypoint.posn.lon);

      MemHandleUnlock(waypointDBEntry);
      dst_act = gc_dist_sphere(lat, lon, wpt_lat, wpt_lon);
    } else {
      return false;
    }
    
    /* get next waypoint */
    err = DmFindRecordByID(gWaypointDB,
      route->wpt_rec_id[gPrefs.act_rte.wpt_index + 1],
      &rec);
      
    if (!err) {
      /* next waypoint found */
      waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
      packed_waypoint =
	(packed_waypoint_t *)MemHandleLock(waypointDBEntry);
      UnpackWaypoint(&waypoint, packed_waypoint);

      wpt_lat = semi2deg(waypoint.posn.lat);
      wpt_lon = semi2deg(waypoint.posn.lon);

      MemHandleUnlock(waypointDBEntry);
      dst_next = gc_dist_sphere(lat, lon, wpt_lat, wpt_lon);
    } else {
      return false;
    }    
    
    if (init_ok && (dst_act > dst_act_old) && (dst_next < dst_next_old)) {
      /* next waypoint selected */
      gPrefs.act_rte.wpt_index++;

      /* update active waypoint record ID */
      gPrefs.act_wpt.waypointID = 
	route->wpt_rec_id[gPrefs.act_rte.wpt_index];
      updated = true;

    } else {
      /* current waypoint still active */
      updated = false;
    }

    /* save current distances */
    dst_act_old  = dst_act;
    dst_next_old = dst_next;
    init_ok = true;
  
    /* unlock route record */
    MemHandleUnlock(routeDBEntry);
  } /* if (!err) */
  
  return updated;
}


/****************************************************************************/
/**
 * \brief	Write Track Point to Track Log Data Base
 *
 * \param	new_trk		new track or track segment
 *
 * \return	- true  if successful
 *              - false if data base is full and wrapping is disabled
 ****************************************************************************/
Boolean TrackWriteLog(Boolean new_trk)
{
  Char		dbname[dmDBNameLength];	/* data base name */
  MemHandle	trkDBEntry;		/* record handle */
  UInt16	index;			/* record index */
  trk_point_t	trkpoint;		/* track point */
  
  if (!gTrackDB) {
    /* data base not yet open */

    /* generate data base name */
    TrackDBName(dbname, gPrefs.act_trk.trackID);

    /* open track DB */
    gTrackDB = DmOpenDatabase(0 /* cardNo */,
      DmFindDatabase(0 /* cardNo */, dbname), dmModeReadWrite);
  }
  
  if ( DmNumRecords(gTrackDB) < dmMaxRecordIndex) {
  
    /* add at the end of the database */
    index = dmMaxRecordIndex;
  
    trkDBEntry = DmNewRecord(gTrackDB, &index, sizeof(trk_point_t));
    
    if (trkDBEntry) {
      DmReleaseRecord(gTrackDB, index, false /* dirty */);
    } else {
      Die("Cannot create track record!");
      return false;
    }
    
  } else {
    /* Data Base is full */
    
    if (!gPrefs.trk_log.wrap) {
      /* stop or wrap */
      if (FrmAlert(TrackFullAlert) == 1) {
	/* change to wrapping */
	gPrefs.trk_log.wrap = true;
      }
    }

    if (gPrefs.trk_log.wrap) {
      /* wrap around */

      /* delete first record, overwrite last record */
      DmRemoveRecord(gTrackDB, 0);
      
      /* add at the end of the database */
      index = dmMaxRecordIndex;

      /* seek last record (will be overwritten) */ 
      DmSeekRecordInCategory(gTrackDB, &index, 0 /* offset */,
	dmSeekBackward, dmAllCategories);

    } else {
      /* stop logging */
      gPrefs.act_trk.log_state = false;
      gPrefs.act_trk.valid     = false;
      DmCloseDatabase(gTrackDB);
      return false;

    }
  }
  
  /* copy track point data to struct */
  trkpoint.new_trk  = new_trk;
  trkpoint.posn.lat = deg2semi(gGPSData.lat);
  trkpoint.posn.lon = deg2semi(gGPSData.lon);
  trkpoint.time     = gGPSData.time - Palm2GarminTimeDiff;
  
  /* store track point in DB */
  StoreTrkPoint(&trkpoint, index);

  return true;
}


/****************************************************************************/
/**
 * \brief	check if track log interval reached
 *
 *		Checks if a new track point has to be logged depending on
 *		the selected track log mode (time or distance) and interval. 
 *		A new track segment is started upon initialization or after
 *		loosing and regaining coverage. 
 *
 * \param	init		init track interval accumulation and
 *				start new track segment
 * \param	new_trk		new track segment (returned by reference)
 *
 * \return
 *		- true  if track point log interval passed
 *		- false if track point log interval not yet passed
 * 
 ****************************************************************************/
Boolean TrackIntervalCheck(Boolean init, Boolean *new_trk)
{
  static Boolean	gps_valid;		/* GPS fix valid */
  static UInt16		acc_time;		/* accumulated time */
  static UInt16		acc_dist;		/* accumulated distance */
  static UInt32		prev_time;		/* previous time */
  static double		prev_lat;		/* previous latitude */
  static double		prev_lon;		/* previous longitude */ 
  static Boolean	init_flag = true;	/* flag: init */
  Boolean		log_now = false;	/* flag: log now */
  
  if (init || (gps_valid == false && gGPSData.valid)) {
    /* start new track segment as soon as GPS fix available */
    init_flag = true;
    acc_time  = 0;
    acc_dist  = 0;
  }
  
  if (gGPSData.valid) {
    if (init_flag) {
      *new_trk  = true;
      log_now   = true;
      init_flag = false;
    } else {
      *new_trk = false;
      
      if (gPrefs.trk_log.mode == 1) {
        /* log interval mode: time */
	acc_time += TimGetSeconds() - prev_time;
	
	if (acc_time >= log_interval_time[gPrefs.trk_log.ival]) {
	  acc_time = 0;
	  log_now = true;
	}
	
      } else {
        /* log interval mode: distance */
	acc_dist += (UInt16)(gc_dist_sphere(gGPSData.lat, gGPSData.lon,
          prev_lat, prev_lon) * calcR(gGPSData.lat));

        if (acc_dist >= log_interval_dist[gPrefs.trk_log.ival]) {
	  acc_dist = 0;
	  log_now = true; 
	}
	
      } /* if (gPrefs.trk_log.mode == 0) */
    } /* if (*new_trk) */
    
    /* save current position */
    prev_lat  = gGPSData.lat;
    prev_lon  = gGPSData.lon;
    
    /* save current time */
    prev_time = TimGetSeconds();
  } /* if (gGPSData.valid) */
  
  /* save current GPS fix state */
  gps_valid = gGPSData.valid;
  
  return log_now;  
}


/****************************************************************************/
/**
 * \brief	check if active waypoint is within proximity distance
 *
 *		If no proximity distance is stored in the waypoint data base
 *		(i.e. the proximity distance is zero), a default value is
 *		used. 
 *
 * \param	lat		current latitude
 * \param	lon		current longitude
 *
 * \return
 *		- true  if active waypoint is within proximity distance
 *		- false if active waypoint is not within proximity distance
 * 
 ****************************************************************************/
Boolean CheckApproach(double lat, double lon)
{
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  UInt16		rec;			/* record index */
  double		wpt_lat;		/* waypoint latitude */
  double		wpt_lon;		/* waypoint longitude */
  double		wpt_dst;		/* distance to waypoint */
  double		prox_dst;		/* wpt's proximity distance */
  Err			err;			/* error code */
  
  /* find active waypoint */
  err = DmFindRecordByID(gWaypointDB, gPrefs.act_wpt.waypointID,
    &rec);
	
  if (!err) {
    /* active waypoint found */
    waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
    packed_waypoint =
      (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    wpt_lat = semi2deg(waypoint.posn.lat);
    wpt_lon = semi2deg(waypoint.posn.lon);
    prox_dst = (double)waypoint.dst;
    MemHandleUnlock(waypointDBEntry);
  
    /* get distance to active waypoint */
    wpt_dst = gc_dist_sphere(lat, lon, wpt_lat, wpt_lon) * calcR(lat);
    
    /* use default if waypoint has no proximity distance set */
    if (prox_dst == 0.0)
      prox_dst = APPROACH_DST;
    
    /* compare current distance with proximity distance */
    if (wpt_dst <= prox_dst) {
      return true;
    }
  }
  
  return false;
}


/****************************************************************************/
/**
 * \brief	generate notification if Active Waypoint is within
 *		proximity distance
 *
 * 		Depending on the user settings, audible and/or visible
 *		notification is triggered. In audible only mode, just one
 *		alarm sound is played. Otherwise an alert form is opened
 *		and has to be acknowledged by the user. After achnowledgement,
 *		the notification will not be repeated for this waypoint.
 *
 ****************************************************************************/
void NotifyApproach(void)
{
  static Boolean		alarm_ok = false;	/* flag: alarm done */
  static UInt32			wpt_id_old = 0;		/* saved act. wpt. ID */

 if (gGPSData.valid && gPrefs.act_wpt.valid) {

    if (gPrefs.act_wpt.waypointID != wpt_id_old) {
      /* new active waypoint - reset approach alert */
      alarm_ok = false;
    }

    if ( CheckApproach(gGPSData.lat, gGPSData.lon) && !alarm_ok) {
      /*
       * Trigger approach notification, audible and/or visible.
       * The alert is generated once. To retrigger the detection, the
       * active waypoint ID has to change.
       */  
      if (gPrefs.approach_sound) {
	SndPlaySystemSound(sndAlarm);
      }
      if (gPrefs.approach_alert) {
	FrmGotoForm(ApproachForm);
      }

      alarm_ok = true;
    }

    /* save active waypoint ID for change detection */
    wpt_id_old = gPrefs.act_wpt.waypointID;
  }
}


/****************************************************************************/
/**
 * \brief	Open a database.
 *
 *		If it doesn't exist, create it.
 * 
 * \param	dbP	Database reference ptr
 * \param	type	Type of the database
 * \param	creator	Creator of the database
 * \param	mode	Open mode
 * \param	cardNo	Card number
 * \param	name	Database name
 * \param	created Flag: new database has been created (return value)
 *
 * \return	Error code of DmOpenDatabaseByTypeCreator() or
 *		DmCreateDatabase(), respectively.
 *		created	-- Flag: new database has been created
 *			   (returned by reference)
 *****************************************************************************/
Err OpenOrCreateDB(DmOpenRef *dbP, UInt32 type, UInt32 creator, 
  UInt16 mode, UInt16 cardNo, char *name, Boolean *created)
{
  Err err = errNone;
  
  *created = false;
  *dbP = DmOpenDatabaseByTypeCreator(type, creator, mode);
  if (! *dbP)
  {
    err = DmGetLastErr();
    if (err == dmErrCantFind)
      err = DmCreateDatabase(cardNo, name, creator, type, false);
    if (err != errNone) 
      return err;
    *created = true;
    
    *dbP = DmOpenDatabaseByTypeCreator(type, creator, mode);
    if (! *dbP) 
      return DmGetLastErr();
  }
  return err;
}


/****************************************************************************/
/**
 * \brief	Creates a Waypoint Database.
 *
 *              The Database contains the capitals of all states of the
 *		European Community (as of 03/2004).
 *		The waypoints were taken from http://wayhoo.com, which in turn
 *		uses GNS (http://gnswww.nima.mil/geonames/GNS/index.jsp).
 *
 *		Quoted from Wayhoo:
 *              Note: GNS coordinates have a low degree of precision,
 *		      latitude and longitude.
 *
 *		The Waypoint Identifiers (i.e. short names) are comprised of
 *		the ISO Country Code and the IATA (International Air
 *		Transport Association) Location ID.
 *		(see http://www.iata.org/codes/index.htm)
 *
 *		This is just an example after all.
 *
 *****************************************************************************/
void InitializeWaypoints(void)
{
  waypoint_t waypoints[] = {
    {"AT-VIE", {DEG2SEMI(48.20000), DEG2SEMI(16.36667)} , 0, 
      "Vienna, Austria (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"BE-BRU", {DEG2SEMI(50.83333), DEG2SEMI(4.33333)} , 0, 
      "Brussels, Belgium (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"DK-CPH", {DEG2SEMI(55.66667), DEG2SEMI(12.58333)} , 0, 
      "Copenhagen, Denmark (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"FI-HEL", {DEG2SEMI(60.17556), DEG2SEMI(24.93417)} , 0, 
      "Helsinki, Finland (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"FR-PAR", {DEG2SEMI(48.86667), DEG2SEMI(2.33333)} , 0, 
      "Paris, France (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"DE-BER", {DEG2SEMI(52.51667), DEG2SEMI(13.40000)} , 0, 
      "Berlin, Germany (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"GR-ATH", {DEG2SEMI(37.98333), DEG2SEMI(23.73333)} , 0, 
      "Athens, Greece (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"IE-DUB", {DEG2SEMI(53.33306), DEG2SEMI(-6.24889)} , 0, 
      "Dublin, Ireland (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"IT-ROM", {DEG2SEMI(41.90000), DEG2SEMI(12.48333)} , 0, 
      "Rome, Italy (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"LU-LUX", {DEG2SEMI(49.61167), DEG2SEMI(6.13000)} , 0, 
      "Luxembourg, Luxembourg (Capital)", 4000.0, sym_lrg_cty, dspl_smbl_only },
    {"NL-HAG", {DEG2SEMI(52.08333), DEG2SEMI(4.30000)} , 0, 
      "The Hague, Netherlands (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"PT-LIS", {DEG2SEMI(38.71667), DEG2SEMI(-9.13333)} , 0, 
      "Lisbon, Portugal (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"ES-MAD", {DEG2SEMI(40.40000), DEG2SEMI(-3.68333)} , 0, 
      "Madrid, Spain (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"SE-STO", {DEG2SEMI(59.33333), DEG2SEMI(18.05000)} , 0, 
      "Stockholm, Sweden (Capital)", 5000.0, sym_lrg_cty, dspl_smbl_only },
    {"GB-LON", {DEG2SEMI(51.50000), DEG2SEMI(-0.11667)} , 0, 
      "London, United Kingdom (Capital)", 10000.0, sym_lrg_cty, dspl_smbl_only }
  }; 
  
//    {"BS", {DEG2SEMI(52.2692), DEG2SEMI(10.52111)} , 0, 
//      "Braunschweig, Germany", 5000.0, sym_med_cty, dspl_smbl_only };
//    {"DE-BWE", {DEG2SEMI(52.319168), DEG2SEMI(10.556167)} , 0, 
//      "Airport Braunschweig, Germany ICAO:EDVE", 1000.0, sym_airport,
//      dspl_smbl_only };

  UInt16  numWaypoints = sizeof(waypoints) / sizeof(waypoints[0]);
  UInt16  i;
    
  for (i = 0; i < numWaypoints; i++) {
    UInt16	index;
    MemHandle	h;
    
    index = DmFindSortPosition(gWaypointDB,
      &waypoints[i],
      0 /* newRecordInfo */,
      (DmComparF *)CompareWayptIdent /* DmComparF *compar */,
      0 /* other */);
    
    h = DmNewRecord(gWaypointDB, &index, 1);
    
    if (h) {
      DmReleaseRecord(gWaypointDB, index, false /* dirty */);
      PackWaypoint(&waypoints[i], index);
    } else {
      Die("Cannot create waypoint record (init)!");
    }
  }
}
