/*****************************************************************************
 *
 * $RCSfile: Data.h,v $
 *
 * GPS4Palm Data Base Definitions (Waypoints)
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-02 17:25:36 $
 *
 * $Revision: 1.19 $
 *
 * $Log: Data.h,v $
 * Revision 1.19  2005-04-02 17:25:36  mp
 * added CompareStringFunc() and Position2Geodb(), removed CompareIDFunc()
 *
 * Revision 1.18  2005/02/05 17:33:42  mp
 * added DeleteTracks()
 *
 * Revision 1.17  2004/12/08 21:01:31  mp
 * added Track data description
 *
 * Revision 1.16  2004/12/08 20:50:25  mp
 * added TrackIntervalCheck()
 *
 * Revision 1.15  2004/12/07 17:15:33  mp
 * added TrackWriteLog()
 *
 * Revision 1.14  2004/12/04 09:39:18  mp
 * added trk_point_t, trk_header_t, StoreTrkHdr(), StoreTrkPoint(),
 * TrackDBName(), and TrackIdent()
 *
 * Revision 1.13  2004/11/25 20:04:28  mp
 * added NotifyApproach()
 *
 * Revision 1.12  2004/11/25 16:58:33  mp
 * added CheckApproach()
 *
 * Revision 1.11  2004/11/20 15:51:59  mp
 * added UpdateActWpt()
 *
 * Revision 1.10  2004/11/16 20:41:55  mp
 * modified PackWaypoint(), PackRouteHdr(), and PackRouteWpt()
 *
 * Revision 1.9  2004/11/16 18:51:35  mp
 * modified InsertRouteWpt()
 *
 * Revision 1.8  2004/11/15 19:51:53  mp
 * modified PackRouteHdr() and PackRouteWpt(), added InsertRouteWpt(),
 * DelRouteWpt(), and MoveRouteWpt()
 *
 * Revision 1.7  2004/11/11 18:47:21  mp
 * moved functions to 'database' section
 *
 * Revision 1.6  2004/06/23 17:56:58  mp
 * added routes
 *
 * Revision 1.5  2004/04/27 18:57:03  mp
 * removed examples, added route DB support
 *
 * Revision 1.4  2004/03/12 20:56:58  mp
 * added InitializeWaypoints()
 *
 * Revision 1.3  2004/03/10 17:23:03  mp
 * Removed member waypointID from waypoint_t and packed_waypoint_t
 *
 * Revision 1.2  2004/03/09 19:11:22  mp
 * Added #include "Garmin.h". Moved some functions to "database"
 * code section.
 *
 * Revision 1.1  2004/03/07 14:26:39  mp
 * initial version
 *
 ****************************************************************************/
#ifndef DATA_H
#define DATA_H

/* Garmin Data Types and Data Structures */
#include "Garmin.h"

/* GPS4Palm Data Types and Data Structures */
#include "common.h"

/*
 * The database functions are in their own segment
 */
#define DATA_SECTION   __attribute__ ((section ("database")))

/*
Databases:
  Waypoints: (sorted by identifier)
    D104_Wpt_Type: packed
    
  Routes: (sorted by route number)
    D201_Rte_Hdr_Type
    UInt16  numItems: number of Waypoints on Route
    UInt32  Wpt_Rec_Id:	PalmOS Database Record ID of Waypoint
    [...]: more Wpt_Rec_Id's    
    
  Tracks: (not sorted - new items are appended)
    Track Header Database:
      D310_Trk_Hdr_Type
    Track Point Database(s):
      D300_Trk_Point_Type
    
    Note: Only track header database exists; it contains one record for each
          track. One track point database exists for each track. Its name is
	  derived from the unique record ID of the associated track header
	  record. The track point data base can contain 0..dmMaxRecordIndex
	  records, one for each track point.

*/

typedef D104_Wpt_Type waypoint_t;

/*
 * Waypoint DB record
 */
typedef struct {
  char			ident[6];	/* identifier (string)            */
  Semicircle_Type	posn;		/* position                       */
  float			dst;		/* proximity distance (meters)    */
  Symbol_Type		smbl;		/* symbol ID                      */
  byte			dspl;		/* display option                 */
  char			cmnt[1];        /* comment, actually may be       */
                                        /* longer than 1, but no more     */
					/* than 40                        */
					/* (including null terminator)    */
} packed_waypoint_t;

typedef D201_Rte_Hdr_Type	route_header_t;

typedef D300_Trk_Point_Type	trk_point_t;

typedef D310_Trk_Hdr_Type	trk_header_t;

/*
 * Route DB record
 *
 * Note: The first two members are compliant to the D201_Rte_Hdr_Type,
 *       but additional space for the string delimiter of the comment
 *       is provided.
 *       This also provides alignment of the fields "items" and
 *       "wpt_rec_id" to the proper boundaries.
 */
typedef struct {
  G_byte	nmbr;			/* route number */
  G_char	cmnt[21];		/* comment */
  UInt16	items;			/* number of waypoints on route   */
  UInt32	wpt_rec_id[1];		/* array of waypoint record IDs,  */
  					/* actually may be longer than 1  */
} route_t;


extern DmOpenRef	gWaypointDB;
extern DmOpenRef	gRouteDB;
extern DmOpenRef	gTrackHdrDB;
extern DmOpenRef	gTrackDB;


Int16 CompareStringFunc(const Char *str1, const Char *str2,
  Int16 i,  SortRecordInfoPtr s1, SortRecordInfoPtr s2,
  MemHandle appInfoH)					DATA_SECTION;
Int16 CompareWayptIdent(waypoint_t *rec1, packed_waypoint_t *rec2,
  Int16 i, SortRecordInfoPtr s1, SortRecordInfoPtr s2,
  MemHandle appInfoH)					DATA_SECTION;
Int16 CompareRteNmbr(route_t *rec1, route_t *rec2,
  Int16 i, SortRecordInfoPtr s1, SortRecordInfoPtr s2,
  MemHandle appInfoH)					DATA_SECTION;

void PackWaypoint(waypoint_t *waypoint, UInt16 index)	DATA_SECTION;
void UnpackWaypoint(waypoint_t *waypoint,
  const packed_waypoint_t *packed_waypoint)		DATA_SECTION;
void PackRouteHdr(route_t *route, UInt16 index)		DATA_SECTION;
void PackRouteWpt(UInt32 wpt_rec_id, UInt16 index)	DATA_SECTION;
void InsertRouteWpt(UInt16 rec, UInt32 wpt_rec_id,
  UInt16 wpt_index)					DATA_SECTION;
void DelRouteWpt(UInt16 rec, UInt16 wpt_index)		DATA_SECTION;
void MoveRouteWpt(UInt16 rec, UInt16 wpt_index,
  Boolean up)						DATA_SECTION;
void StoreTrkHdr(trk_header_t *trkhdr, UInt16 index)	DATA_SECTION;
void StoreTrkPoint(trk_point_t *trkpoint, UInt16 index)	DATA_SECTION;
void TrackDBName(Char *dbname, UInt32 trackID)		DATA_SECTION;
void TrackIdent(Char *ident)				DATA_SECTION;
void DeleteTracks(void)					DATA_SECTION;
Boolean Position2Geodb(GPSType gps_data)		DATA_SECTION;
Boolean UpdateActWpt(double lat, double lon,
  Boolean init)						DATA_SECTION;
Boolean TrackWriteLog(Boolean new_trk)			DATA_SECTION;
Boolean TrackIntervalCheck(Boolean init,
  Boolean *new_trk)					DATA_SECTION;
Boolean CheckApproach(double lat, double lon)		DATA_SECTION;
void NotifyApproach(void)				DATA_SECTION;

Err OpenOrCreateDB(DmOpenRef *dbP, UInt32 type, UInt32 creator, UInt16 mode,
  UInt16 cardNo, char *name, Boolean *created)		DATA_SECTION;
void InitializeWaypoints(void)				DATA_SECTION;

#ifdef DEBUG_BUILD
void CheckDatabases(void)				DATA_SECTION;
#endif
#endif
