/*****************************************************************************
 *
 * $RCSfile: GPS.c,v $
 *
 * GPS4Palm Main functions
 * (PilotMain, AppStart, AppStop, EventLoop, ApplicationHandleEvent and others)
 *
 * Based on the example "GPS" from the book
 * "Palm Programming, the Developer's Guide, 2nd edition"
 * by O'Reilly.
 * (Copyright (c) 1998-2001, Neil Rhodes and Julie McKeehan  neil@pobox.com)
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-28 10:15:46 $
 *
 * $Revision: 1.49 $
 *
 * $Log: GPS.c,v $
 * Revision 1.49  2005-05-28 10:15:46  mp
 * fixed check for gpslib
 *
 * Revision 1.48  2005/05/13 18:31:30  mp
 * removed variables filename and gAutoPwrOff
 *
 * Revision 1.47  2005/05/06 13:32:59  mp
 * moved Bluetooth code to Serial.c
 *
 * Revision 1.46  2005/04/24 14:09:05  mp
 * fixed comment
 *
 * Revision 1.45  2005/04/20 20:58:25  mp
 * implemented opening of Bluetooth Virtual Serial Port
 * with connecting to known device (address)
 *
 * Revision 1.44  2005/04/08 14:45:28  mp
 * added check for High-Density Display Feature Set
 *
 * Revision 1.43  2005/03/24 09:18:36  mp
 * added check for FiveWay controller
 *
 * Revision 1.42  2005/02/26 14:55:19  mp
 * fixed doxygen-tags
 *
 * Revision 1.41  2005/02/26 14:51:04  mp
 * added 3.5 New Feature Set availability check
 *
 * Revision 1.40  2005/02/19 20:04:12  mp
 * changed filename[32] to filename[FILENAMEMAXLEN]
 *
 * Revision 1.39  2005/02/19 14:37:42  mp
 * moved checking VFS Feature Set from FileselForm.c
 *
 * Revision 1.38  2005/02/06 15:02:00  mp
 * added RTS serial mux control with SrmControl()
 *
 * Revision 1.37  2005/02/05 17:35:47  mp
 * added deletion of orphaned tracks, improved serial port handling
 *
 * Revision 1.36  2004/12/09 17:24:44  mp
 * added form switching with hard keys
 *
 * Revision 1.35  2004/12/07 17:15:02  mp
 * added closing of track DB upon application stop
 *
 * Revision 1.34  2004/12/04 09:37:18  mp
 * modified default Active Track preferences
 *
 * Revision 1.33  2004/12/02 23:00:38  mp
 * added default values for gPrefs.act_trk to application preferences
 *
 * Revision 1.32  2004/12/01 17:45:43  mp
 * added default values for track log preferences
 *
 * Revision 1.31  2004/11/30 21:04:01  mp
 * added TrackForm event handler, modified DB open/close
 *
 * Revision 1.30  2004/11/25 20:05:51  mp
 * added ApproachForm event handler and default settings
 *
 * Revision 1.29  2004/11/24 21:10:40  mp
 * moved static function declarations from header to implementation file,
 * added AboutForm event handler function
 *
 * Revision 1.28  2004/11/23 17:51:23  mp
 * removed unused variables, added missing return value
 *
 * Revision 1.27  2004/11/20 15:55:31  mp
 * added deletion of invalid Route DB
 *
 * Revision 1.26  2004/11/19 17:55:09  mp
 * added initialization of gPrefs.auto_wpt
 *
 * Revision 1.25  2004/11/13 15:54:36  mp
 * reordered includes, added RouteEditForm
 *
 * Revision 1.24  2004/11/11 21:38:01  mp
 * added default assignement for Active Route
 *
 * Revision 1.23  2004/04/29 17:42:09  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.22  2004/04/27 18:55:25  mp
 * added opening of route DB
 *
 * Revision 1.21  2004/04/22 18:39:18  mp
 * added check for gpslib
 *
 * Revision 1.20  2004/04/17 12:46:13  mp
 * moved kBaudRate to Serial.h, added gpslib.h
 *
 * Revision 1.19  2004/04/06 20:39:02  mp
 * corrected preferences version check in AppStart()
 *
 * Revision 1.18  2004/03/20 13:37:52  mp
 * added call of FrmCloseAllForms() to AppStop()
 *
 * Revision 1.17  2004/03/13 15:00:17  mp
 * added clearing of NavigationForm TimeField in case of no GPS
 *
 * Revision 1.16  2004/03/13 14:08:25  mp
 * added update of gGPSData.valid in case of no GPS
 *
 * Revision 1.15  2004/03/13 12:39:15  mp
 * moved call of Set_UTM_Parameters() to AppStart()
 *
 * Revision 1.14  2004/03/12 20:56:22  mp
 * added call of init_radius() to AppStart()
 *
 * Revision 1.13  2004/03/10 17:24:15  mp
 * added gPrefs.act_wpt.valid to preferences
 *
 * Revision 1.12  2004/03/09 19:07:21  mp
 * Cleaned up/added comments to #include directives.
 *
 * Revision 1.11  2004/03/07 14:23:18  mp
 * added waypoint database open/close
 *
 * Revision 1.10  2004/03/03 19:03:58  mp
 * added WayptEditForm event handler
 *
 * Revision 1.9  2004/03/01 17:22:15  mp
 * Fixed selection of connection profile (if New Serial Manager present).
 *
 * Revision 1.8  2003/12/30 20:19:54  mp
 * added MathLib
 *
 * Revision 1.7  2003/12/01 20:12:53  mp
 * added default values for new application preferences
 *
 * Revision 1.6  2003/11/20 20:42:45  mp
 * added event handler for MapOptsForm, added some Preferences defaults
 *
 * Revision 1.5  2003/11/18 18:44:25  mp
 * Added current Form and Auto Power-Off status to Application Preferences.
 * Implemented switch to previous form (stored in Preferences) at
 *   application start.
 * Cleaned up code and added comments.
 *
 * Revision 1.4  2003/10/20 17:15:10  mp
 * removed gShowingLostSatellite, changed 'No GPS' status display to
 * UpdateStatus(), changed SetFieldTextFromStr() to SetFieldText()
 *
 * Revision 1.3  2003/10/18 16:18:12  mp
 * added FileSelForm, added load/save of preferences
 *
 * Revision 1.2  2003/10/15 19:07:54  mp
 * moved HandleMessage etc. to HandleMessage.c
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#ifdef HAVE_FIVEWAY_SDK
#include <palmOne_68K.h>
#endif
#include <SerialMgrOld.h>
#include "GPS.h"
#include "Data.h"		/* Waypoint DB create/open/close */
#include "ResourceDefines.h"	/* Resource IDs */
#include "MathLib.h"		/* MathLib Find/Open/Close/Remove */
#include "gpslib.h"		/* GPS I/O library */
#include "HandleMessage.h"	/* UpdateStatus(), HandleMessage() */
#include "MainForm.h"		/* MainFormHandleEvent() */
#include "PortForm.h"		/* PortFormHandleEvent() */
#include "SkyviewForm.h"	/* SkyviewFormHandleEvent() */
#include "NavigationForm.h"	/* NavigationFormHandleEvent() */
#include "FileselForm.h"	/* FileSelFormHandleEvent() */
#include "MiscOptsForm.h"	/* MiscOptsFormHandleEvent() */
#include "MapOptsForm.h"	/* MapOptsFormHandleEvent() */
#include "WaypointForm.h"	/* WaypointFormHandleEvent() */
#include "WayptEditForm.h"	/* WayptEditFormHandleEvent() */
#include "RouteForm.h"		/* RouteFormHandleEvent() */
#include "RouteEditForm.h"	/* RouteEditFormHandleEvent() */
#include "TrackForm.h"		/* TrackFormHandleEvent() */
#include "AboutForm.h"		/* AboutFormHandleEvent() */
#include "ApproachForm.h"	/* ApproachFormHandleEvent() */
#include "DebugForm.h"		/* DebugFormHandleEvent() */
#include "MapForm.h"		/* MapFormHandleEvent */
#include "Utils.h"		/* misc. */
#include "Serial.h"		/* Serial Manager/New Serial Manager wrappers */
#include "geo.h"		/* Geodesic/Geometric Utility Functions */
#include "utm.h"		/* Set_UTM_Parameters() */
#include "common.h"

/** Minimum Version for 2.0 Capabilities */
#define kOS20MinVersion		0x02003000	

/** Minimum PalmOS Version */
#define kMinimumVersion	0x03000000

/** Minimum Version for 3.5 New Feature Set */
#define kOS35MinVersion		0x03503000	

#define GpsLibName	"gpslib"


/* #define USE_HARD_KEYS */	/* set define to switch main forms with keys */  


/* ---------------------------------------- */
/* Global timing variables (initialization) */
/* ---------------------------------------- */

/** tickCount of last time we read data from GPS */
UInt32		gLastSuccessfulReception = 0;		

/** tickCount of last time we displayed GPS data on the Palm device */
UInt32		gLastTimeDisplay = 0;

/** tickCount of the next scheduled read */
UInt32		gNextReadTime = 0;


#ifdef DEBUG_FORM
/* Start of Debugging */

/* number of checksum errors */
UInt32		gChkErrs = 0; 

/* minimum number of characters received */
UInt32		gMinReceived = 0xFFFFFFFF;

/* maximum number of characters received */
UInt32		gMaxReceived = 0;

/* End   of Debugging */
#endif

/** Global flag */
Boolean	gFormOpened = false;

/** VFS Feature Set available */
Boolean gVfsManagerFtr;

/** OS 3.5 Feature Set available */
Boolean gFtrSet35;

/** High-Density Display Feature Set available */
Boolean gHdFtrSet;

/** FiveWay Controller available */
Boolean		gFiveWayAvail;
static Boolean	haveFiveWay = false;

/* GPS Data */
extern GPSType gGPSData;

/** Preferences data structure */
PrefsType gPrefs;

/* four-character port name or logical port number (New Serial Manager) */
extern UInt32 gSerialPort;

/* Variables used by Connection Manager */
extern Char **gProfiles;
extern UInt16 gNumProfiles;


/* ------------------------------------- */
/* Global timing variables (declaration) */
/* ------------------------------------- */

/** Ticks per second (OS/Hardware specific) */
UInt16 gTicksPerSecond;

/**
 *  if we go this long without updating the time
 *  then update as soon as we get a valid time
 *  (without waiting for an even 5-second time)
 */
UInt32 gMaxTicksWithoutTime;

/**
 *  if we go this long without communicating with GPS,
 *  we've lost it and need to notify the user
 */
UInt32	gTicksToLoseGPS;

/* Static Functions */
static Err AppStart(void);
static void AppStop(void);
static void OS2(void);
static Boolean HaveFiveWay(void);
static Boolean ApplicationHandleEvent(EventPtr event);
static void EventLoop(void);
static long TimeUntilNextRead(void);
static void WarnUserIfHaventHeardFromGPSRecently(void)	MSGHANDLER_SECTION;
static UInt32 WaitForData(UInt32 bufferSize)		MSGHANDLER_SECTION;
static Boolean ReadAvailableData(Char *data,
  UInt32 bufferSize, UInt32 numBytesPending)		MSGHANDLER_SECTION;


/****************************************************************************
 *
 * FUNCTION:	AppStart
 *
 * DESCRIPTION: Loads Application Preferences.
 *              Checks if Serial Library is available and opens serial port,
 *		checks if Math Library is available and
 *		shows appropriate error message on failure.
 *              Initializes timing parameters and enables serial port mux.
 *
 * GLOBALS:	gPrefs			-- Application Preferences
 *		gWaypointDB		-- Waypoint Database reference
 *		gNewSerialManager	-- New Serial Manager present
 *		gPortID			-- Serial Port ID
 ****************************************************************************/
static 
Err AppStart(void)
{
  UInt32	vfsMgrVersion;			/* VFS Manager Version */
  UInt32	version;			/* feature version no. */
  UInt16	prefsSize;			/* preferences struct size */
  Int16		prefsVersion = noPreferenceFound;
  UInt16	cnc_index;			/* CncMgr profile index */
  UInt32	host_port;			/* Host serial port */
  Err		err = 0;			/* error code */
  Boolean	created;			/* flag: DB created */	
  LocalID	dbID;				/* local DB ID */
  
  prefsSize = sizeof(PrefsType);

  /* Get Application Preferences (unsaved) */
  prefsVersion = PrefGetAppPreferences(kCreatorId, kPrefsId, &gPrefs,
     &prefsSize, false);

  if ( (prefsVersion != kVersion) || (prefsVersion == noPreferenceFound) ) {
    /* no preferences or wrong version: set defaults */ 
    StrCopy(gPrefs.filename, "");
    gPrefs.mag = 1;
    gPrefs.xofs = 0;
    gPrefs.yofs = 0;
    gPrefs.dx = 0;
    gPrefs.dy = 0;
    gPrefs.dw = 0;
    gPrefs.dh = 0;
    gPrefs.form = GPSMainForm;
    gPrefs.pwroff = AUTOPWROFF;
    gPrefs.auto_wpt = true;
    gPrefs.approach_alert = true;		/* Approach alert form        */
    gPrefs.approach_sound = true;		/* Approach alert sound       */
    gPrefs.serprefs.gps_port = 0;		/* GPS serial port/CncMgr ID  */ 
    gPrefs.serprefs.host_port = 0;		/* Host serial port/CncMgrID  */
    gPrefs.serprefs.rts_ctrl = false;		/* Multiplexer control by RTS */
    gPrefs.trk_log.mode = 0;			/* Track Log Mode: 	Dist. */
    gPrefs.trk_log.ival = 0;			/* Track Log Interval:	10 m  */
    gPrefs.trk_log.wrap = 0;			/* Track Log Full:	stop  */
    gPrefs.units.pos_unit = POS_D;		/* Position format:     DD.DDD */
    gPrefs.units.dst_unit = DST_KM;		/* Distance unit:       km      */
    gPrefs.units.alt_unit = ALT_M;		/* Altitude unit:       m       */
    gPrefs.units.spd_unit = SPD_KMH;		/* Speed unit:          km/h    */
    gPrefs.mapprefs.scale = 20000;		/* Map scale:           20.000  */
    gPrefs.mapprefs.download = DLD_A;		/* Map download policy: auto    */
    gPrefs.mapprefs.select = SEL_AUTO;		/* Map selection:       auto    */
    gPrefs.mapprefs.fetch = FETCH_READY;	/* FetchMap state:	ready   */
    gPrefs.act_wpt.valid = false;		/* Active waypoint:     n/a     */
    gPrefs.act_rte.valid = false;		/* Active route:        n/a     */
    gPrefs.act_trk.valid = false;		/* Active track:	n/a     */ 
    gPrefs.act_trk.log_state = false;		/* Active track log:	off     */
  }

  /* check if VFS Manager Feature Set is available */
  err = FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfsMgrVersion);
  gVfsManagerFtr = (err == 0) ? true : false; 
  err = 0;

  /* check if High-Density Display Feature Set is available */
  gHdFtrSet = (FtrGet(sysFtrCreator, sysFtrNumWinVersion, &version) == 0) &&
                (version >= 4);

  if (GetRomVersion() >= kOS20MinVersion) {
    /* PalmOS 2.x or higher specific operations */
    OS2();
  }

  gFiveWayAvail = HaveFiveWay();
  
  gFtrSet35 = (GetRomVersion() < kOS35MinVersion) ? false : true;

  err = OpenOrCreateDB(&gWaypointDB, kWaypointDBType /* type */,
    kCreatorId /* creator */, dmModeReadWrite, 0 /* cardNo */,
    kWaypointDBName /* name */, &created);

  if (created) {
    InitializeWaypoints();
  
    /*
     * If a new Waypoint DB has been created, the Route DB
     * (if any exists) is invalid and must be deleted.
     */
     dbID = DmFindDatabase(0 /* cardNo */, kRouteDBName);
     
     if (dbID) {
       DmDeleteDatabase(0 /* cardNo */, dbID);
     }
  }
  
  err = OpenOrCreateDB(&gRouteDB, kRouteDBType /* type */,
    kCreatorId /* creator */, dmModeReadWrite, 0 /* cardNo */,
    kRouteDBName /* name */, &created);

  err = OpenOrCreateDB(&gTrackHdrDB, kTrkHdrDBType /* type */,
    kCreatorId /* creator */, dmModeReadWrite, 0 /* cardNo */,
    kTrkHdrDBName /* name */, &created);
  
  /* if created, delete all tracks */
  if (created) {
    DeleteTracks();
  }
  
  err = Set_UTM_Parameters(	SEMIMAJOR_AXIS,
      			    	FLATTENING,
			    	0 /* Override_Zone */);

  if (err != UTM_NO_ERROR)
     return err;

  dbID = DmFindDatabase(0, "gpslib");
  if ( !dbID ) {
    FrmAlert(NoGpslibAlert);
    return sysErrLibNotFound;
  }

  /* check if New Serial Manager is available */
  gNewSerialManager = (FtrGet(sysFileCSerialMgr,
    sysFtrNewSerialPresent, &version) == 0) && (version != 0);
  
  /* check if Connection Manager feature set is available */
  gCncManagerFtr = (FtrGet(kCncFtrCncMgrCreator,
    kCncFtrCncMgrVersion, &version) == 0) && (version >= 0x00040001);
  
  if (gNewSerialManager) {
    if (gCncManagerFtr) {
    
      /* get GPS Connection Manager profile ID from preferences */
      if (CncProfileGetIndex(gPrefs.serprefs.gps_port, &cnc_index) == errNone) {

	/* get list of Connection Manager profiles */
	err = CncGetProfileList(&gProfiles, &gNumProfiles);

        if (!err) {
	  /* get port name and baud rate from current profile */
	  err = CncGetProfileInfo(gProfiles[cnc_index],
            &gSerialPort, &gPrefs.serprefs.gps_baud, 0, 0, 0, 0, 0, 0);			

          /* free Connection Manager profile list */
          FreeCncProfileList();
        }
      }

      /* get Host Connection Manager profile ID from preferences */
      if (CncProfileGetIndex(gPrefs.serprefs.host_port, &cnc_index) == errNone) {

	/* get list of Connection Manager profiles */
	err = CncGetProfileList(&gProfiles, &gNumProfiles);

        if (!err) {
	  /* get port name and baud rate from current profile */
	  err = CncGetProfileInfo(gProfiles[cnc_index],
            &host_port, &gPrefs.serprefs.host_baud, 0, 0, 0, 0, 0, 0);			

          /* free Connection Manager profile list */
          FreeCncProfileList();
        }
      }

    } else {
      /* no Connection Managere feature set available */
      if (gPrefs.serprefs.gps_port)
        gSerialPort = gPrefs.serprefs.gps_port;
	
      if (gPrefs.serprefs.host_port)
        host_port = gPrefs.serprefs.host_port;
    }
        
    if (!err) {
      /*
       * Note:
       * -----
       * We do not try to open the host port now, because
       * some interfaces (IrDA, Bluetooth) have to be enabled first,
       * but should be disabled to save power until they are
       * really needed.
       */
       
      /* set Serial Port to be used by gpslib */
      GpslibSetSerialPort(host_port);
      
      /* open GPS port */
      if (gSerialPort == sysFileCVirtRfComm) {
	/*
	 * Bluetooth Virtual Serial Port --
	 * Connect to known device (address)
	 */
        err = BluetoothOpen(gSerialPort, gPrefs.serprefs.gps_baud,
	  gPrefs.serprefs.gps_btaddr, &gPortID);

      } else {
        /* open non-BT serial port */
        err = SrmOpen(gSerialPort, gPrefs.serprefs.gps_baud, &gPortID);
      }
    }
    
    if (!err) {
      FrmGotoForm(gPrefs.form);    
    } else { 
      err = 0;
      FrmGotoForm(GPSPortForm);
    }

  } else {
    /* use default settings */
    err = SysLibFind("Serial Library", &gPortID);
    if (err == 0)
      err = SerOpen(gPortID, 0, kBaudRate);

    if (err == serErrAlreadyOpen) {
      FrmAlert(SerialInUseAlert);
      DoClose(gPortID);
    } else if (err) {
      FrmAlert(CantOpenSerialAlert);
    } else {
      err = DoSetReceiveBuffer(gPortID, gSerialBuffer,
	sizeof(gSerialBuffer));
      FrmGotoForm(gPrefs.form);
    }
  }

  if (!err) {
    err = SysLibFind(MathLibName, &MathLibRef);
    if (err)
      err = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
    
    if (err) {
      FrmAlert(NoMathLibAlert);
      DoClose(gPortID);
    } else {
      err = MathLibOpen(MathLibRef, MathLibVersion);
    
      if (err) {
        FrmAlert(NoMathLibAlert);
        DoClose(gPortID);
      }
    }
  }
  
  if (!err) {
    /* Initialize lookup table earth_radius[latitude] */
    init_radius();

#if defined(SERMUX_VISOR_HS) | \
    defined(SERMUX_VISOR_KBD) | \
    defined(SERMUX_PALMIII_RTS)
    sermux_enable(true);
#else
    sermux_enable(gPrefs.serprefs.rts_ctrl);
#endif
    sermux_select(false);

    
    gTicksPerSecond = SysTicksPerSecond();
    gMaxTicksWithoutTime = 6 * gTicksPerSecond;
    gTicksToLoseGPS = 15 * gTicksPerSecond;
  }
  
  return err;
}

/*****************************************************************************
 * FUNCTION:	AppStop
 *
 * DESCRIPTION: Restores serial buffer size and closes serial port.
 *		Disables serial port mux.
 *		Saves Application Preferences. 
 *
 ****************************************************************************/
static void AppStop(void)
{
  Int	usecount;
  Err	err;

  FrmCloseAllForms();
  
  /* Close Waypoint Database */
  err = DmCloseDatabase(gWaypointDB);

  /* Close Route Database */
  err = DmCloseDatabase(gRouteDB);

  /* Close Track Header Database */
  err = DmCloseDatabase(gTrackHdrDB);

  /* Close Track Point Database (if open) */
  if (gTrackDB)
    err = DmCloseDatabase(gTrackDB);

  /* Close Serial Port */
  if (gPortID) {
    /* restore the default buffer before closing the serial port */
    DoSetReceiveBuffer(gPortID, NULL, 0);
    DoClose(gPortID);
  }
  
  /* Disable Serial Port Mux Control */
  sermux_enable(false);
  
  /* Set Application Preferences (unsaved) */
  PrefSetAppPreferences(kCreatorId, kPrefsId, kVersion, &gPrefs,
     sizeof(PrefsType), false);
  
  err = MathLibClose(MathLibRef, &usecount);
  ErrFatalDisplayIf(err, "Can't close MathLib");
  if (usecount == 0)
    SysLibRemove(MathLibRef);
  
}


/****************************************************************************/
/**
 * \brief	PalmOS 2.x or higher specific operations
 *
 * \note	Copied from plucker-1.8
 *
 ****************************************************************************/
static void OS2(void)
{
#if defined( HAVE_FIVEWAY_SDK ) || defined( HAVE_HANDSPRING_SDK )
    Err       err;
    UInt32    value;
#endif
#ifdef HAVE_FIVEWAY_SDK
    err = FtrGet( navFtrCreator, navFtrVersion, &value );
    if ( err == errNone )
        haveFiveWay = true;
#endif
#ifdef HAVE_HANDSPRING_SDK
    err = FtrGet( sysFileCSystem, sysFtrNumUIHardwareHas5Way, &value );
    if ( err == errNone ) {
        haveFiveWay = true;
        err = FtrGet( hsFtrCreator, hsFtrIDNavigationSupported, &value );
        if ( err == errNone ) {
            haveHsNav = true;
        }
    }
#endif

}


/****************************************************************************/
/**
 * \brief	Check if a FiveWay controller is present
 *
 * \note	Copied from plucker-1.8
 *
 * \returned	true if FiveWay controller is available, else false
 ****************************************************************************/
static Boolean HaveFiveWay(void)
{
#if defined( HAVE_FIVEWAY_SDK ) || defined( HAVE_HANDSPRING_SDK )
    return haveFiveWay;
#else
    return false;
#endif
}


/*****************************************************************************
 * FUNCTION:	TimeUntilNextRead
 *
 * DESCRIPTION: Calculate time until next read from serial port is required.
 *
 ****************************************************************************/
static 
long TimeUntilNextRead(void)
{
  if (!gFormOpened) {
    return evtWaitForever;
  } else {
    Int32 timeRemaining;

    timeRemaining = gNextReadTime - TimGetTicks();

    if (timeRemaining < 0)
      timeRemaining = 0;
    return timeRemaining;
  }
}


/*****************************************************************************
 * FUNCTION:	WarnUserIfHaventHeardFromGPSRecently
 *
 * DESCRIPTION: The name says it all. On the position form, all fields are
 *		cleared, on the Sky View form, the last data set is kept.
 *		The warning is shown on the StatusField of the current form.
 *
 ****************************************************************************/
static 
void WarnUserIfHaventHeardFromGPSRecently(void)
{
  UInt16	form = FrmGetActiveFormID();
  
  if ((TimGetTicks() - gLastSuccessfulReception) > gTicksToLoseGPS) {
    UpdateStatus( STAT_NO_GPS );
    gGPSData.valid = false;
    
    switch (form) {
    
      case GPSMainForm:
	SetFieldText(TimeField, NULL, false, true);
	SetFieldText(LatitudeField, NULL, false, true);
	SetFieldText(LongitudeField, NULL, false, true);
	SetFieldText(SpeedField, NULL, false, true);
	SetFieldText(AltField, NULL, false, true);
	SetFieldText(CMGField, NULL, false, true);
	SetFieldText(MagField, NULL, false, true);
	break;
      
      case SkyviewForm:
	SetFieldText(TimeField, NULL, false, true);
	SetFieldText(PDOPField, NULL, false, true);
	SetFieldText(HDOPField, NULL, false, true);
	SetFieldText(VDOPField, NULL, false, true);
	SetFieldText(FixField, NULL, false, true);
	SetFieldText(SatsField, NULL, false, true);
        break;
      
      case NavigationForm:
  	SetFieldText(TimeField, NULL, false, true);
        break;
    } // switch (form) 
  } // if ((TimGetTicks() - gLastSuccessfulReception) > gTicksToLoseGPS)
}


/*****************************************************************************
 * FUNCTION:	WaitForData
 *
 * DESCRIPTION: Wait for data available in serial buffer. Fill read buffer
 *              or wait for inter-character gap (0.5 s)
 *
 * PARAMETERS:	bufferSize	-- buffer size assigned to serial port
 *
 * RETURNED:    number of bytes available
 ****************************************************************************/
static 
UInt32 WaitForData(UInt32 bufferSize)
{
  UInt32 numBytesPending = 0;
  Err	err;
  const Int32 kOneHalfSecond = gTicksPerSecond / 2;

  // We'll fill our read buffer, or 1/2 second between
  // bytes, whichever comes first.
  err = DoReceiveWait(gPortID, bufferSize, kOneHalfSecond);
  if (err == serErrLineErr) {
    DoReceiveFlush(gPortID, 1);	// will clear the error
    return 0;			
  } 
  if (err != serErrTimeOut)
    ErrFatalDisplayIf(err != 0, "DoReceiveWait");
  err = DoReceiveCheck(gPortID, &numBytesPending);
  if (err == serErrLineErr) {
    DoReceiveFlush(gPortID, 1);	// will clear the error
    return 0; 			
  }
  ErrFatalDisplayIf(err != 0, "DoReceiveCheck Fail");
  
  return numBytesPending;
}


/*****************************************************************************
 * FUNCTION:	ReadAvailableData
 *
 * DESCRIPTION: Read data from serial buffer
 *
 * PARAMETERS:	data		-- pointer to string buffer
 *		bufferSize	-- buffer size assigned to serial port
 *		numBytesPending	-- number of bytes available
 *
 * RETURNED:    always 0
 ****************************************************************************/
static 
Boolean ReadAvailableData(Char *data, UInt32 bufferSize, 
  UInt32 numBytesPending)
{
   Err		err;
   UInt32	numBytes;

   // read however many bytes are waiting
   if (numBytesPending > bufferSize)
     numBytesPending = bufferSize;
   numBytes = DoReceive(gPortID, data, numBytesPending, 0, &err);
   if (err == serErrLineErr) {
     DoReceiveFlush(gPortID, 1);	// will clear the error
   } else
     data[numBytes] = '\0';		// null-terminate theData
   return 0;
}




/**********************************************************************/
/**
 * \fn		Boolean ReadFromGPS(void)
 *
 * \brief	Reads data from GPS receiver.
 *
 *              If no messages are received for a certain time,
 *              a warning is generated.
 *              A large block of data is read from the serial buffer
 *              in one go. Then the data is parsed. Only complete
 *              messages (starting with $GP and ending with checksum) 
 *              and valid checksum are passed to the message handler
 *              function.
 *
 * \return	Flag which shows if message has been handled
 *              successfully. (TBD: all msgs, first msg, ???)
 *
 ***********************************************************************/
Boolean ReadFromGPS(void)
{
  Err		err;
  Boolean	updatedDisplay = false;
  Boolean	messageComplete = false;
  UInt32	numBytesPending;
  char		theData[(80+2)*5+1];	// three lines (80 chars with <CR><LF>
					//  + one for null byte
  	
  // if we've gone too long without hearing from the GPS
  // tell the user
  WarnUserIfHaventHeardFromGPSRecently();
	
  numBytesPending = WaitForData(sizeof(theData)-1);

#ifdef DEBUG_FORM
  if (numBytesPending < gMinReceived) {
    gMinReceived = numBytesPending;
  }
  if (numBytesPending > gMaxReceived) {
    gMaxReceived = numBytesPending;
  }
#endif  
  
  if (numBytesPending > 0) {
    char		*startOfMessage;	/* ptr to start of message */
    char		*endOfMessage;		/* ptr to end   of message */
    unsigned char	chk;			/* checksum */

    Char		exp[3];			/* expected checksum (string) */
    Char		act[3];			/* actual checksum   (string) */

    err = ReadAvailableData(theData, sizeof(theData) - 1, numBytesPending);
    if (err)
      return false;

    /* start parsing at start of buffer */
    startOfMessage = theData;
           
    /* loop while start of message can be found */
    while ((startOfMessage = StrStr(startOfMessage, "$GP")) != NULL) {
      
      /* start at start of message, skip '$' for checksum calculation */
      endOfMessage = startOfMessage + 1;
      messageComplete = false;
      chk = 0;
      
      /* loop until end of buffer (or end of message detected) */
      while (*endOfMessage != '\0') {
	
	if (*endOfMessage == '*') {
	  /* checksum delimiter found */
	  
	  if (( *(endOfMessage+1) != '\0') &&
	      ( *(endOfMessage+2) != '\0')) {
	    /* checksum found */
	    
	    /* expected checksum */
	    exp[0] = *(endOfMessage+1);
	    exp[1] = *(endOfMessage+2);
	    
	    /* actual checksum */
	    act[0] = nibble2hex((chk & 0xF0) >> 4);
	    act[1] = nibble2hex( chk & 0x0F);
	    
	    if ((exp[0] == act[0]) && (exp[1] == act[1])) {
	      /* checksum ok, handle message */
	      updatedDisplay = HandleMessage(startOfMessage);
	    }
	    
#ifdef DEBUG_FORM
	    else
	    {
	      /* increment checksum error counter */
	      if (gChkErrs != 0xFFFFFFFF) {
	        gChkErrs++;
              }
            }
#endif
	    
	    // end of message found 
	    messageComplete = true;
	    
	  } /* checksum received */
          
	  // search for start of next message
	  //   after current message
	  startOfMessage = endOfMessage;
	  
	  break;
	} /* if (*endOfMessage == '*') */
	
	/* update checksum */
	chk ^= *endOfMessage;
	
	/* next character */
	endOfMessage++;
      } /* while (*endOfMessage != '\0') */
      
      // end of buffer contents
      if (!messageComplete)
        break;
    }
  }

  return updatedDisplay;
}


/*****************************************************************************
 * FUNCTION:	ApplicationHandleEvent
 *
 * DESCRIPTION: Set Application Event Handler to Form specific Handler.
 *
 * PARAMETERS:	event	-- pointer to event (only frmLoadEvent handled)
 *
 * RETURNED:    Event handled as boolean.
 ****************************************************************************/
static Boolean ApplicationHandleEvent(EventPtr event)
{
  FormPtr	form;
  UInt16	formId;
  Boolean	handled = false;

  if (event->eType == frmLoadEvent) {
    /* Load the form resource specified in the event then activate the form. */
    formId = event->data.frmLoad.formID;
    form = FrmInitForm(formId);
    FrmSetActiveForm(form);

    switch (formId) {
      case GPSMainForm:
	      FrmSetEventHandler(form, MainFormHandleEvent);
	      break;

      case SkyviewForm:
	      FrmSetEventHandler(form, SkyviewFormHandleEvent);
	      break;

      case NavigationForm:
	      FrmSetEventHandler(form, NavigationFormHandleEvent);
	      break;

      case MapForm:
	      FrmSetEventHandler(form, MapFormHandleEvent);
	      break;

      case GPSPortForm:
	      FrmSetEventHandler(form, PortFormHandleEvent);
	      break;

      case FileSelForm:
	      FrmSetEventHandler(form, FileSelFormHandleEvent);
	      break;

      case MiscOptsForm:
	      FrmSetEventHandler(form, MiscOptsFormHandleEvent);
	      break;

      case MapOptsForm:
	      FrmSetEventHandler(form, MapOptsFormHandleEvent);
	      break;

      case WaypointForm:
	      FrmSetEventHandler(form, WaypointFormHandleEvent);
	      break;

      case WayptEditForm:
	      FrmSetEventHandler(form, WayptEditFormHandleEvent);
	      break;

      case RouteForm:
	      FrmSetEventHandler(form, RouteFormHandleEvent);
	      break;

      case RouteEditForm:
              FrmSetEventHandler(form, RouteEditFormHandleEvent);
              break;

      case TrackForm:
	      FrmSetEventHandler(form, TrackFormHandleEvent);
	      break;

      case AboutForm:
	      FrmSetEventHandler(form, AboutFormHandleEvent);
	      break;
       
      case ApproachForm:
              FrmSetEventHandler(form, ApproachFormHandleEvent);
	      break;
     
#ifdef DEBUG_FORM
      case DebugForm:
	      FrmSetEventHandler(form, DebugFormHandleEvent);
	      break;
#endif		
    }
    handled = true;
  }

  return handled;
}


/*****************************************************************************
 * FUNCTION:	EventLoop
 *
 * DESCRIPTION: Main Event Loop
 *
 ****************************************************************************/
static 
void EventLoop(void)
{
	EventType	event;
	UInt16		error;
	
	do {
		/* Wait for OS Event or time-out (to read from serial port) */
		EvtGetEvent(&event, TimeUntilNextRead());
		
		/* Reset Auto Power-Off Timer if Auto Power-Off is disabled */
		if (! gPrefs.pwroff) {
		  EvtResetAutoOffTimer();
		}

#if 0
#ifdef USE_HARD_KEYS
                /* use the hard keys to switch between main forms */
		if ((event.eType == keyDownEvent) && 
		    (event.data.keyDown.modifiers & commandKeyMask)) {
		  
		  switch (event.data.keyDown.chr) {
		    case vchrHard1:
		      event.eType = menuEvent;
		      event.data.menu.itemID = PositionMenu;
		      EvtAddEventToQueue(&event);
		      break;
		      
		    case vchrHard2:
		      event.eType = menuEvent;
		      event.data.menu.itemID = SkyviewMenu;
		      EvtAddEventToQueue(&event);
		      break;
		    
		    case vchrHard3:
		      event.eType = menuEvent;
		      event.data.menu.itemID = NavigationMenu;
		      EvtAddEventToQueue(&event);
		      break;
		    
		    case vchrHard4:
		      event.eType = menuEvent;
		      event.data.menu.itemID = MapMenu;
		      EvtAddEventToQueue(&event);
		      break; 
		  }
		} 		
#endif
#endif

		/* Dispatch Event */
		if (! SysHandleEvent(&event))
		  if (! MenuHandleEvent(0, &event, &error))
		    if (! ApplicationHandleEvent(&event))
		      FrmDispatchEvent(&event);
			
	} while (event.eType != appStopEvent);
}


/****************************************************************************/
/**
 * \fn		UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
 *
 * \brief 	Application Entry Function
 *
 * \param	cmd		Launch Code
 * \param	cmdPBP		Command Parameter Block Pointer
 * \param	launchFlags	Launch Flags (set by OS)
 *
 * \return	always 0
 *
 ****************************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
  UInt16 error;		

  error = RomVersionCompatible (kMinimumVersion, launchFlags);
  if (error)
	  return error;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    if (!AppStart()) {
      EventLoop();
      AppStop();
    }
  }
  
  return 0;
}
