/*****************************************************************************
 *
 * $RCSfile: MainForm.c,v $
 *
 * GPS4Palm Main Form (Position Form)
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-02 10:09:43 $
 *
 * $Revision: 1.20 $
 *
 * $Log: MainForm.c,v $
 * Revision 1.20  2005-04-02 10:09:43  mp
 * modified Position2Clipboard() and Position2Geodb() to return 'handled'
 *
 * Revision 1.19  2005/04/02 07:26:59  mp
 * added PosgeodbMenu event handling
 *
 * Revision 1.18  2005/03/25 13:44:41  mp
 * added PosclipMenu event handling
 *
 * Revision 1.17  2005/01/23 18:40:43  mp
 * added PortsMenu event handling
 *
 * Revision 1.16  2004/12/09 17:31:11  mp
 * removed debug specific code
 *
 * Revision 1.15  2004/12/08 21:02:43  mp
 * added track logging, modified nilEvent handling
 *
 * Revision 1.14  2004/11/30 20:43:48  mp
 * added TrackMenu event handling
 *
 * Revision 1.13  2004/11/25 20:14:20  mp
 * fixed typo
 *
 * Revision 1.12  2004/11/25 20:07:49  mp
 * added NotifyApproach()
 *
 * Revision 1.11  2004/11/25 16:58:06  mp
 * added function call UpdateActWpt()
 *
 * Revision 1.10  2004/11/24 21:12:40  mp
 * moved static function declarations from header to implementation file,
 * added AboutForm menu event handling
 *
 * Revision 1.9  2004/04/29 20:10:26  mp
 * modified comments for doxygen
 *
 * Revision 1.8  2004/02/28 13:24:56  mp
 * added WaypointMenu
 *
 * Revision 1.7  2004/01/18 19:01:53  mp
 * added FrmHideObject/FrmShowObject calls for switching between Latitude/Longitude
 * and UTM position display mode
 *
 * Revision 1.6  2003/12/31 15:05:35  mp
 * added comments, cleaned up code
 *
 * Revision 1.5  2003/11/20 20:44:29  mp
 * added status redraw on frmOpenEvent
 *
 * Revision 1.4  2003/11/18 20:52:56  mp
 * Added Auto Power-Off disable, moved serial port setup to GPS.c,
 * added current form to Application Preferences, modified Fields.
 *
 * Revision 1.3  2003/10/20 17:23:38  mp
 * added field deallocation on frmCloseEvent
 *
 * Revision 1.2  2003/10/15 19:12:33  mp
 * modified event loop, added menu item Debug
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>			
#include "ResourceDefines.h"
#include "Serial.h"
#include "Utils.h"
#include "MainForm.h"
#include "GPS.h"
#include "HandleMessage.h"
#include "Utils.h"
#include "Data.h"		/* UpdateActWpt(), NotifyApproach() */
#include "Clip.h"
#include "common.h"

/* Preferences data structure */
extern PrefsType gPrefs;

/* Auto Power-Off */
extern UInt16	gAutoPwrOff;

/* GPS Data */
extern GPSType		gGPSData;

/* Static Functions */
static void MainFormInit(void);


static 
void MainFormInit(void)
{
	gFormOpened = true;
}

/****************************************************************************/
/**
 * \brief	Position Form event handler
 *
 * \param	event		pointer to event structure
 *
 * \return	event handled flag
 ****************************************************************************/
Boolean MainFormHandleEvent(EventPtr event)
{
	Boolean		handled = false;
	Boolean		updatedDisplay;
	FormPtr		frmP;		/* form ptr */

	switch (event->eType) {
	
        case menuEvent:
          switch (event->data.menu.itemID) {
      
            case SkyviewMenu:
	      FrmGotoForm(SkyviewForm);
	      handled = true;
	      break;
            
	    case NavigationMenu:
	      FrmGotoForm(NavigationForm);
	      handled = true;
	      break;

	    case MapMenu:
	      FrmGotoForm(MapForm);
	      handled = true;
	      break;

            case MiscOptsMenu:
	      FrmGotoForm(MiscOptsForm);
	      handled = true;
	      break;
	       
            case MapOptsMenu:
	      FrmGotoForm(MapOptsForm);
	      handled = true;
	      break;
	    
	    case PortsMenu:
	      if (gNewSerialManager) {
	        FrmGotoForm(GPSPortForm);
	      }
	      handled = true;
	      break;
	    
            case WaypointMenu:
	      FrmGotoForm(WaypointForm);
	      handled = true;
	      break;

            case RouteMenu:
	      FrmGotoForm(RouteForm);
	      handled = true;
	      break;
	       
            case TrackMenu:
	      FrmGotoForm(TrackForm);
	      handled = true;
	      break;
	    
	    case PosclipMenu:
	      handled = Position2Clipboard(gGPSData, gPrefs.units);
	      break;
	      
	    case PosgeodbMenu:
	      handled = Position2Geodb(gGPSData);
	      break;
	      
            case AboutMenu:
	      FrmGotoForm(AboutForm);
	      handled = true;
	      break;
	       
#ifdef DEBUG_FORM
            case DebugMenu:
	      FrmGotoForm(DebugForm);
	      handled = true;
	      break;
#endif
	  }
	  break;
	  
	case nilEvent:
                handled = true;
		/* throw away anything in the buffer -- we want fresh data */
		DoReceiveFlush(gPortID, 1);
		
		/* read data from GPS */
		updatedDisplay = ReadFromGPS();
                
		if (gGPSData.valid && gPrefs.act_rte.valid && gPrefs.auto_wpt) {
	          /* update active waypoint */
		  UpdateActWpt(gGPSData.lat, gGPSData.lon, false /* init */);
		}
				
		/* Active Waypoint approach detection and display */
		NotifyApproach();
		
		/* Track Logging */
		if (gPrefs.act_trk.log_state) {
		  Boolean new_trk;
		  		  
		  if (TrackIntervalCheck(false /* init */, &new_trk)) {
		    TrackWriteLog(new_trk);
		  }
		}		
		break;
		
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		gPrefs.form = GPSMainForm;
		MainFormInit();
		
		if (gPrefs.units.pos_unit == POS_UTM) { 
		  FrmHideObject(frmP, FrmGetObjectIndex(frmP, LatLabel));
		  FrmHideObject(frmP, FrmGetObjectIndex(frmP, LonLabel));
		  FrmShowObject(frmP, FrmGetObjectIndex(frmP, EastLabel));
		  FrmShowObject(frmP, FrmGetObjectIndex(frmP, NorthLabel));
		} else {
		  FrmHideObject(frmP, FrmGetObjectIndex(frmP, EastLabel));
		  FrmHideObject(frmP, FrmGetObjectIndex(frmP, NorthLabel));
		  FrmShowObject(frmP, FrmGetObjectIndex(frmP, LatLabel));
		  FrmShowObject(frmP, FrmGetObjectIndex(frmP, LonLabel));
		}   

		FrmDrawForm(frmP);
		UpdateStatus(STAT_REDRAW);
		handled = true;
		break;
		
	case frmCloseEvent:
        	/* Deallocate Fields' Text Memory */
	  	SetFieldText(PositionStatusField, NULL, false, true);
		SetFieldText(TimeField, NULL, false, true);
	  	SetFieldText(LatitudeField, NULL, false, true);
	  	SetFieldText(LongitudeField, NULL, false, true);
	  	SetFieldText(SpeedField, NULL, false, true);
	  	SetFieldText(AltField, NULL, false, true);
	  	SetFieldText(CMGField, NULL, false, true);
	  	SetFieldText(MagField, NULL, false, true);
		handled = true;
		break;
	}
	return(handled);
} /* MainFormHandleEvent() */
