/*****************************************************************************
 *
 * $RCSfile: MapForm.c,v $
 *
 * GPS4Palm Map Form
 *
 * Application event loop for Palm GPS
 *   - determines image file type (GIF/PNG) and calls
 *     appropriate decoding function
 *   - allows zooming and panning of image
 *
 * created: 2002-12-15
 *
 * History:
 * 2007-07-10:  deep = 16 for scrDisplayMode (fs)
 *              set screen depth for depth > 8 disabled : resets Palm TX
 *              set palette for depth > 8 disabled : sometimes resets Palm TX
 *
 * 2003-05-03:  imported from Palm Image Viewer (piv)
 *		removed Handera specific code
 *		removed unused code
 *		(temporarily) disabled calls to (currently) unimplemented forms
 *
 * 2002-12-15:	first version, taken from the program PiNGer (pngmain.c)
 *		  which is part of the ZBoxZ application suite (Rev. 0.30)
 *		  by Tom Zerucha.
 *		Moved dobitmap() to separate file, restructured event loop,
 *		  added some (file open and ZLib) checks, added some comments.
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * Changes for Palm TX are Copyright (C) 07/2007 Frank Saurbier
 * <frank.saurbier@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2007-10-05 22:29:03 $
 *
 * $Revision: 1.37.2.1 $
 *
 * $Log: MapForm.c,v $
 * Revision 1.37.2.1  2007-10-05 22:29:03  mp
 * added Frank's changes (fixes reset bug on TX)
 *
 * Revision 1.37  2005/05/16 19:16:18  mp
 * removed test code
 *
 * Revision 1.36  2005/05/16 17:16:30  mp
 * removed mapprefs.fetch handling in case of FS open failure
 *
 * Revision 1.35  2005/05/16 15:31:42  mp
 * added zoom to calculation of crosshair coordinates
 *
 * Revision 1.34  2005/05/15 12:52:55  mp
 * moved file handling functions to file_io.h
 * changed from global file reference/handle to file descriptor (fd)
 * structure passed to functions
 * map search/download code rewritten
 *
 * Revision 1.33  2005/04/17 18:35:19  mp
 * improved High-Density Display support
 *
 * Revision 1.32  2005/04/08 14:50:24  mp
 * added High-Density Display support
 *
 * Revision 1.31  2005/04/02 10:09:43  mp
 * modified Position2Clipboard() and Position2Geodb() to return 'handled'
 *
 * Revision 1.30  2005/04/02 07:26:59  mp
 * added PosgeodbMenu event handling
 *
 * Revision 1.29  2005/03/25 13:45:11  mp
 * added PosclipMenu event handling
 *
 * Revision 1.28  2005/03/24 09:17:30  mp
 * added support for FiveWay controller for panning
 *
 * Revision 1.27  2005/02/25 18:39:39  mp
 * replaced gCardNo by gPrefs.filecard, fixed VFS flag handling
 *
 * Revision 1.26  2005/02/19 19:56:53  mp
 * modified size of filename[]
 *
 * Revision 1.25  2005/02/19 18:25:27  mp
 * modified status messages while reading file
 *
 * Revision 1.24  2005/02/19 14:35:54  mp
 * added VFS support to search_map()
 *
 * Revision 1.23  2005/02/19 14:00:14  mp
 * fixed File Stream open
 *
 * Revision 1.22  2005/02/19 12:20:24  mp
 * fixed support for multiple cards (File Streaming API), added VFS support
 *
 * Revision 1.21  2005/01/23 18:41:06  mp
 * added PortsMenu event handling
 *
 * Revision 1.20  2004/12/19 10:05:36  mp
 * added call to UpdateStatus()
 *
 * Revision 1.19  2004/12/10 19:52:56  mp
 * replaced DmGet1Resource by DmGetResource
 *
 * Revision 1.18  2004/12/09 17:30:31  mp
 * replaced strings by string resources
 *
 * Revision 1.17  2004/12/08 20:49:47  mp
 * added track logging
 *
 * Revision 1.16  2004/11/30 20:44:35  mp
 * added TrackMenu event handling
 *
 * Revision 1.15  2004/11/25 20:15:08  mp
 * added finction calls: NotifyApproach() and UpdateActWpt()
 *
 * Revision 1.14  2004/11/24 21:13:58  mp
 * moved static function declarations from header to implementation file,
 * added AboutForm menu event handling
 *
 * Revision 1.13  2004/11/23 17:50:15  mp
 * removed unused variables, added #include "Utils.h"
 *
 * Revision 1.12  2004/11/21 11:33:22  mp
 * removed debug specific change
 *
 * Revision 1.11  2004/04/29 18:36:50  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.10  2004/03/12 20:55:20  mp
 * moved invocation of init_radius() to GPS.c
 *
 * Revision 1.9  2004/03/11 21:28:10  mp
 * replaced gLon/gLat/gGPSok by equivalent gGPSData members
 *
 * Revision 1.8  2004/02/28 17:21:43  mp
 * added menuEvent WaypointMenu
 *
 * Revision 1.7  2003/12/30 19:48:40  mp
 * replaced printDouble() by format_number()
 *
 * Revision 1.6  2003/12/30 14:48:54  mp
 * renamed static functions
 *
 * Revision 1.5  2003/12/28 18:22:12  mp
 * Major parts of position display and map selection code rewritten.
 *
 * Revision 1.4  2003/11/20 20:57:14  mp
 * Main changes: added debug output (map center/current position/scale),
 * added invocation of FetchMap (some test code still in place),
 * added decoding of map parameters from filename.
 *
 * Revision 1.3  2003/10/18 16:14:34  mp
 * re-enabled interaction with FileSelForm
 *
 * Revision 1.2  2003/10/15 19:11:28  mp
 * renamed lat/lon to gLat/gLon, added extern declaration of gLat/gLon
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Libraries/PalmOSGlue/PalmOSGlue.h>
#ifdef HAVE_FIVEWAY_SDK
#include <palmOne_68K.h>
#endif
#include <Extensions/ExpansionMgr/VFSMgr.h>


#include "stringil.h"
#include "ResourceDefines.h"
#include "MapForm.h"
#include "FileselForm.h"
#include "dobitmap.h"
#include "pngdec.h"
#include "gifread.h"
#include "common.h"
#include "Serial.h"
#include "GPS.h"
#include "HandleMessage.h"
#include "Data.h"		/* UpdateActWpt(), NotifyApproach() */
#include "Clip.h"
#include "geo.h"
#include "georef.h"
#include "fp.h"
#include "Utils.h"
#include "getdir.h"
#include "file_io.h"
#include "fiveway.h"

#define ON_MAP(x,y) \
  ( (x >= dx+CROSS_X) && (x <= dx+dw-CROSS_X) && \
    (y >= dy+CROSS_Y) && (y <= dy+dh-CROSS_X))
#define ON_MAP_WAYP(x,y) \
  ( (x >= dx) && (x <= dx+dw) && \
    (y >= dy) && (y <= dy+dh))



#define FW_ZOOM	0		/**< FiveWay Mode: Zoom */
#define FW_PAN	1		/**< FiveWay Mode: Pan */
//#define VERBOSE

/* Preferences data structure */
extern PrefsType gPrefs;

/* GPS Data */
/*   defined and modified in HandleMessage() (HandleMessage.c) */
extern GPSType gGPSData;

/* FiveWay Controller available */
extern Boolean gFiveWayAvail;

/* High-Density Display Feature Set available */
extern Boolean		gHdFtrSet;

/* OS 3.5 Feature Set available */
extern Boolean		gFtrSet35;

extern unsigned short int w, h, bandh;

/* returns palette if color, etc. */
extern unsigned int mapmax;
extern RGBColorType map[256];



// ABMH Distance to Wpt
#include "MathLib.h"
#include "Sinetab.h"

RGBColorType rgbcolor;


/** Constants for distance unit conversions and display */
extern const struct {
  char		*unit;
  double	conv;
} dst_c[];
/** Constants for speed unit conversions and display */
extern const struct {
  char		*unit;
  double	conv;
} spd_c[];

// abmh
int abmh_debug(char *pos, unsigned long l, char *stri);

int mapformwayptredraw;

int scroll_scale;
int scroll_redraw_i;
#define scroll_redraw_count 100
double old_lat, old_lon;

/****************************************************************************/
/**
 * \brief	Display waypoint ID and update waypoint position
 *
 * \param	lat		latitude returned by reference
 * \param	lon		longitude returned by reference
 *
 ****************************************************************************/
static
void get_wpt(double *lat, double *lon)
{
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  UInt16		rec;			/* record index */
  Err			err;			/* error code */
  Char			tmp[7];			/* waypoint identifier */ 
  MemHandle		strH;			/* string handle */
  MemPtr		strP;			/* string ptr */
  
  /* default "Waypoint Identifier" */ 
  /* get string from resource */
  strH = DmGetResource(strRsc, NoWptStr);

  if (strH != NULL) {
    /* string handle valid, lock it */
    strP = MemHandleLock(strH);
  
    /* copy string */
    StrCopy(tmp, strP);

    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  }

  /* display active waypoint */
  err = DmFindRecordByID(gWaypointDB, gPrefs.act_wpt.waypointID,
    &rec);
	
  if (!err) {
    /* active waypoint found */
    waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
    packed_waypoint =
      (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    MemSet((Char *)tmp, 7, 0);
    StrNCopy(tmp, (Char *)&waypoint.ident, 6);
    *lat = semi2deg(waypoint.posn.lat);
    *lon = semi2deg(waypoint.posn.lon);
    MemHandleUnlock(waypointDBEntry);
  }

  //  CtlSetLabel((ControlPtr)GetObjectFromActiveForm(NavigationWptIDSel), tmp);
}



/*
 * Note:
 * 'color' is also used to determine if
 * WinPushDrawState() and WinPopDrawState() may be called,
 * which require the 3.5 New Feature Set!
 */
/** passed to PNG (maybe others) to generate greyscale or colors */
Boolean color;

/*
 * gPrefs Structure
 *   modified within MapFormHandleEvent(), doGIF(), doPNG(), and dobitmap()
 *   The filename is loaded from/saved to the application preferences.
 */
extern PrefsType gPrefs;

/* Static Functions */
static void MapFormInit(void)			MAPFORM_SECTION;
static void MapFormDeinit(FormPtr frmP)		MAPFORM_SECTION;
static void callFetchmap(void)			MAPFORM_SECTION;
static void eraseFilename(char *filename)		MAPFORM_SECTION;
static void updateCross(Int32 x, Int32 y,
  Int32 *xo, Int32 *yo,
  double map_lat, double map_lon,
  UInt32 map_scale, int mag, UInt32 deep)	MAPFORM_SECTION;


static void MapFormInit(void)
{
	gFormOpened = true;
	mapformwayptredraw = 0;
}


static void MapFormDeinit(FormPtr frmP)
{
#pragma unused(frmP)  

}

static void eraseFilename(char *filename)
{     
  MemHandle		strH;			/* string handle */
  MemPtr		strP;			/* string ptr */
  
  /* get string from resource */
  strH = DmGetResource(strRsc, DecodeStr);
  
  if (strH != NULL) {
    /* string handle valid, lock it */
    strP = MemHandleLock(strH);
  
    WinEraseChars(strP, strlen(strP), 0, 0);
    WinEraseChars(filename, strlen(filename), 0, 12);

    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  }
}


/*****************************************************************************
 * FUNCTION:	callFetchmap
 *
 * DESCRIPTION: Calls the application FetchMap to download new map.
 *		Before switching to FetchMap, the serial port is closed,
 *		the port multiplexer is disabled (mobile phone is default)
 *		and the application preferences are saved.
 *		The required map parameters are passed in the parameter block
 *		to FetchMap. After completion of FetchMap, GPS4Palm is opened
 *		again, but execution starts from PilotMain (as if started from
 *		the Application Launcher again).   
 *
 ****************************************************************************/
static void callFetchmap(void)
{
  MapParamType *	map_param;		/* map parameters */
  LocalID		dbID;			/* local DB ID of FetchMap */
  Err			err;			/* error code */
  
  map_param = MemPtrNew(sizeof(MapParamType));
  map_param->lat    = gGPSData.lat;
  map_param->lon    = gGPSData.lon;
  map_param->scale  = gPrefs.mapprefs.scale;	/* from Preferences */
  if (gHdFtrSet) {
    map_param->width  = 320;
//    map_param->height = 480; /* (fs) ich will die grossen: sollte "option" fetchmap werden */
    map_param->height = 320;
  } else {
    map_param->width  = 160;
    map_param->height = 160; 
  }


  dbID = DmFindDatabase(0, "FetchMap");
  if ( !dbID ) {
    FrmAlert(FetchMapAlert);
  } else {
    MemPtrSetOwner(map_param, 0);
    
    /* close serial port before leaving application */
    if (gPortID) {
      /* restore the default buffer before closing the serial port */
      DoSetReceiveBuffer(gPortID, NULL, 0);
      DoClose(gPortID);
    }
    sermux_enable(false);

    /* Set Application Preferences (unsaved) */
    PrefSetAppPreferences(kCreatorId, kPrefsId, kVersion, &gPrefs,
       sizeof(PrefsType), false);

   /* Invoke FetchMap to download map */
   /* Note: Execution actually does not return from SysUIAppSwitch */
   err = SysUIAppSwitch(0,                           /* cardNo */
			dbID,                        /* Database ID */
			sysAppLaunchCmdNormalLaunch, /* cmd */
			map_param);                  /* cmdPBP */

    sermux_enable(true);
  }
}

/*****************************************************************************
 * FUNCTION:	updateCross
 *
 * DESCRIPTION: Display/Move crosshair if map is new or position has changed.
 *
 * RETURNS:	True if updated, else false.
 *
 ****************************************************************************/

static void updateCross(Int32 x, Int32 y, Int32 *xo, Int32 *yo,
                           double map_lat, double map_lon, UInt32 map_scale,
			   int mag, UInt32 deep)
{
  Int32		x_old = *xo;		/* previous x-coordinate */
  Int32		y_old = *yo;		/* previous y-coordinate */
  Int32     cr_x, cr_y;	    /* length of cross for 16bit depth - const CROSS_X -_Y too small */

#ifdef DEBUG
  Char		s[20];			/* string buffer */
  Coord		y_top = 110;		/* y-coordinate of text */
  Coord		y_dst = 10;		/* y-distance for text lines */
#endif


  if (gFtrSet35) {
    WinPushDrawState();
    /* restore default colors */
    WinSetForeColor(UIColorGetTableEntryIndex(UIObjectForeground));
    WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
    WinSetTextColor(UIColorGetTableEntryIndex(UIFieldText));
  }
  
  if (gHdFtrSet) {
    WinSetCoordinateSystem(kCoordinatesNative);
  }

  /*  if (deep < 16)
      {*/
  // ABMH
  //  cr_y = cr_x = 16;
    cr_x = CROSS_X;
    cr_y = CROSS_Y;
    /*  }
	else cr_y = cr_x = 8;*/


  /* redraw map at old crosshair position */
  dobitmap(x_old-cr_x, y_old-cr_y, mag, x_old-cr_x, y_old-cr_y, 
    2*cr_x+1, 2*cr_y+1, deep);
  
  /* draw new crosshair */ 
  //  WinSetForeColor(UIColorGetTableEntryIndex(UIObjectForeground));
  //  WinSetForeColor (0);
  //  WinSetForeColor(white)
  rgbcolor.r = rgbcolor.g = rgbcolor.b = 255; // white
  //  white = WinRGBToIndex(&rgbcolor);
  //  rgbcolor.r = rgbcolor.g = rgbcolor.b = 0; // black
  //  black = WinRGBToIndex(&rgbcolor);
  WinSetForeColor(WinRGBToIndex(&rgbcolor));
  WinDrawLine(x-cr_x, y, x+cr_x, y);
  WinDrawLine(x+1, y-cr_y, x+1, y+cr_y);
  rgbcolor.r = rgbcolor.g = rgbcolor.b = 0; // black
  WinSetForeColor(WinRGBToIndex(&rgbcolor));
  WinDrawLine(x, y-cr_y, x, y+cr_y);
  WinDrawLine(x-cr_x, y+1, x+cr_x, y+1);
  //  WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
  //  WinSetForeColor (255);
  //  WinInvertLine(x-cr_x, y+1, x+cr_x, y+1);
  //  WinInvertLine(x+1, y-cr_y, x+1, y+cr_y);


#ifdef DEBUG
  if (gHdFtrSet) {
    y_top = WinScaleCoord(y_top, false /* ceiling */);
    y_dst = WinScaleCoord(y_dst, false /* ceiling */);
  }
    
  /* TBD: debug only, remove later or create proper text overlay */
  StrPrintF(s, "%lu", map_scale);
  WinDrawChars(s, StrLen(s), 0, y_top);
  y_top += y_dst;
  format_number(map_lat, 7, s);
  WinDrawChars(s, StrLen(s), 0, y_top);
  y_top += y_dst;
  format_number(map_lon, 7, s);
  WinDrawChars(s, StrLen(s), 0, y_top);
  y_top += y_dst;
  format_number(gGPSData.lat, 7, s);
  WinDrawChars(s, StrLen(s), 0, y_top);
  y_top += y_dst;
  format_number(gGPSData.lon, 7, s);
  WinDrawChars(s, StrLen(s), 0, y_top);
#endif

  /*
    optimation: to prevent push/pop to be used too often, this is done within the calling function
  if (gFtrSet35) {
    WinPopDrawState();
  }
  */
}


/****************************************************************************/
/**
 * \brief	Clear screen and display message "No map available!"
 *
 ****************************************************************************/
void noMap()
{
  RectangleType	r;				/* rectangle */
  Coord		maxx, maxy;			/* max. display extent */
  MemHandle	strH;				/* string handle */
  MemPtr	strP;				/* string ptr */

  /*----------------------------------------------------
   * No map available
   *   The hopeless case - either we tried and failed or
   *   we are not even allowed to try! So we just have
   *   to inform the user... too bad!
   *  
   *----------------------------------------------------*/

  if (gFtrSet35) {
    WinPushDrawState();
    
    /* restore default colors */
    WinSetForeColor(UIColorGetTableEntryIndex(UIObjectForeground));
    WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
    WinSetTextColor(UIColorGetTableEntryIndex(UIFieldText));
  }

  if (gHdFtrSet) {
    WinSetCoordinateSystem(kCoordinatesNative);
  }

  WinGetDisplayExtent(&maxx, &maxy);
  r.topLeft.x = r.topLeft.y = 0;
  r.extent.x = maxx;
  r.extent.y = maxy;
  WinEraseRectangle(&r, 0);

  /* get string from resource */
  strH = DmGetResource(strRsc, NoMapStr);

  if (strH != NULL) {
    /* string handle valid, lock it */
    strP = MemHandleLock(strH);

    WinDrawChars(strP, StrLen(strP), (maxx-FntCharsWidth(strP, StrLen(strP))) / 2, (maxy-FntCharHeight()) / 2);

    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  } /* if (strH != NULL) */

  if (gFtrSet35) {
    WinPopDrawState();
  }  
}




static void copy_of_cvt_ultoa(unsigned long,char *);
/*****************************************************************************
 * FUNCTION:	cvt_ultoa
 *
 * DESCRIPTION:	Convert unsigned long value to string.
 *
 * PARAMETERS:	value	-- unsigned long value
 *		strP	-- ptr to string buffer for returning result
 *
 * RETURNED:	%
 *****************************************************************************/
static void copy_of_cvt_ultoa(unsigned long value, char *strP)
{
  int i;
  char *cP;
  char tmpstr[16];		/* a 32 bit long can give up to 10 digits */

    if( value <= 9 )		/* a quick special case */
    {
	*strP++ = value + '0';
	*strP = '\0';
	return;
    }

    /* We convert the number backwards */
    for( cP = tmpstr+9, i = 10; i-- > 0; )
    {
	*cP-- = (value % 10) + '0';
	value /= 10;
    }

    tmpstr[10] = '\0';

    for( cP = tmpstr; *cP && (*cP == '0'); )
	++cP;
    
    StrCopy(strP,cP);
}

/****************************************************************************/
/**
 * \brief	Map Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	event handled flag
 ****************************************************************************/
Boolean MapFormHandleEvent(EventPtr eventP)
{
  static FileDescrType	fd;			/* File Descriptor */
  static UInt32		deep, odeep;		/* orig./curr. display depth */
  EventType		event;			/* event to queue */
  static int		xo, yo;			/* image offset (img coords) */
  static int		mag;			/* magnification (1..8) */
  static Coord		dm;			/* magnification (1..8) */
  static int		refresh;		/* refresh flag (0..2) */
  Char			tbuf[4];		/* character buffer */
  static RectangleType	r;			/* rectangle */
  static Coord		dx, dy;
  static Coord		maxx, maxy;		/* max. display extent */
  Err			err;			/* error code */
  Boolean		handled = false;	/* event handled flag */
  Boolean		updatedDisplay;		/* updated display flag */
  static double		map_lat;		/* Lat. at map center */
  static double		map_lon;		/* Lon. at map center */
  static UInt32		map_scale;		/* Map scale */
  static Int32		x_old = CROSS_X;	/* x (saved crosshair pos.) */ 
  static Int32		y_old = CROSS_Y;	/* y (saved crosshair pos.) */
  Int32			x;			/* x (new crosshair pos.) */
  Int32			y;			/* y (new crosshair pos.) */
  static Int32		dw;			/* draw region width */
  static Int32		dh;			/* draw region height */
  static Boolean	decode_ok = false;	/* image file decoded flag */
  static Boolean	map_valid;		/* map valid flag */
  static UInt8		fiveway_mode = FW_ZOOM;	/* 5-Way Controller mode */
  static Boolean	cross_visible = false;	/* crosshair visible flag */
  
  
  double		brg;			/* bearing */

  dm = mag;
  
  switch (eventP->eType) {
    case frmOpenEvent:
    
      /* init form */
      MapFormInit();
      
      /* init map file status */
      map_valid = false;
      decode_ok = false;

      /* save current form in application prefs */
      gPrefs.form = MapForm;
      
      /* Erase Form */
      WinEraseWindow();
      
      UpdateStatus(STAT_REDRAW);
      
      if (StrCompare(gPrefs.filename, "") == 0) {
        /* No file selected (yet) */
	
	if (gPrefs.mapprefs.select == SEL_AUTO) {
          /* Auto map selection */ 
	  noMap();
	  
	} else {
	  /* User Map selection */
	  FrmGotoForm(FileSelForm);
	  
	}
	break;
      
      }
            
      {
         Char	tmp[] = "Decoding";
	 WinDrawChars(tmp, strlen(tmp), 0, 0);
      }
    
      fd.fh = NULL;
      fd.fr = 0;
        
      if (!(gPrefs.filecard & VFS_FLAG)) {	
	Char	tmp[] = "File Stream Database";
	
	/* use file streaming API */
	WinDrawChars(gPrefs.filename, strlen(gPrefs.filename), 0, 12);
	WinDrawChars(tmp, StrLen(tmp), 0, 36);

	fd.fh = FileOpen( gPrefs.filecard, /* card nr */
			  gPrefs.filename,
			  'DATA',  /* type */
			  'BOXR',  /* creator */
                    	  fileModeReadOnly | fileModeAnyTypeCreator, /* mode */
			  &err);   /* err */

	/*
	 * Show alert and open file browser if file cannot be opened!
	 * This can happen in both SEL_AUTO and SEL_USER mode, e.g. if
	 * a file is broken or has been removed (while still set as default
	 * in application preferences).
	 */
	if ((fd.fh == 0) || (err != 0)) {
          FrmCustomAlert(NoFileAlert, gPrefs.filename, NULL, NULL);
          
	  if (gPrefs.mapprefs.select == SEL_USER) { 
	    /* let user select another map */
	    FrmGotoForm(FileSelForm);
	    eraseFilename(gPrefs.filename);
	  }
	  
	  /* clear filename */
	  StrCopy(gPrefs.filename, "");
	  
	  break;
	}
      } else {
	Char		tmp[] = "File on Virtual File System";
	MemPtr		pathName;
	UInt32		vfs_vol = gPrefs.filecard & ~VFS_FLAG;
	
	/* use VFS API */
	WinDrawChars(gPrefs.filename, strlen(gPrefs.filename), 0, 12);
	WinDrawChars(tmp, StrLen(tmp), 0, 36);
	

	/*
	if (gPrefs.mapprefs.scale == 8500) {
	  StrCopy(FILEDIR, FILEDIRpre);
	  StrCat(FILEDIR, "/8500");
	} else if (gPrefs.mapprefs.scale == 18000) {
	  StrCopy(FILEDIR, FILEDIRpre);
	  StrCat(FILEDIR, "/18k");
	} else {
	  StrCopy(FILEDIR, FILEDIRpre);
	}
	*/

	/*
	{
	  static first_start = 1;
	  if (first_start)
      StrCopy(FILEDIR, FILEDIRpre);
	}
	*/

        /* generate absolute file path */
	//        pathName = MemPtrNew(sizeof(FILEDIR) + StrLen(gPrefs.filename) + 2);
	/* abmh 2009
        pathName = MemPtrNew(StrLen(FILEDIR) + StrLen(gPrefs.filename) + 2);
	StrCopy(pathName, FILEDIR);
	StrCat(pathName, "/");
	StrCat(pathName, gPrefs.filename);

	err = VFSFileOpen(vfs_vol, pathName, vfsModeRead, &fd.fr);
	MemPtrFree(pathName);
	*/



        pathName = MemPtrNew(StrLen(FILEDIRpre) + StrLen(gPrefs.filename) + 2);
	StrCopy(pathName, FILEDIRpre);
	StrCat(pathName, "/");
	StrCat(pathName, gPrefs.filename);
	err = VFSFileOpen(vfs_vol, pathName, vfsModeRead, &fd.fr);
	MemPtrFree(pathName);
	 if (!fd.fh && !fd.fr) {
	   char scaleStr[16];
	   Int16 numpos, i;
	   
	   copy_of_cvt_ultoa((unsigned long) gPrefs.mapprefs.scale, scaleStr);
	   StrCopy(FILEDIR, FILEDIRpre);
	   StrCat(FILEDIR, "/");
	   StrCat(FILEDIR, scaleStr);
	   StrCat(FILEDIR, ".x");
	   numpos = StrLen(FILEDIR) - 1;
	   
	   for (i = 0; i < 10; i++) {
	     FILEDIR[numpos] = i + 48;

	     pathName = MemPtrNew(StrLen(FILEDIR) + StrLen(gPrefs.filename) + 2);
	     StrCopy(pathName, FILEDIR);
	     StrCat(pathName, "/");
	     StrCat(pathName, gPrefs.filename);
	     err = VFSFileOpen(vfs_vol, pathName, vfsModeRead, &fd.fr);
	     MemPtrFree(pathName);
	     if (!fd.fh && !fd.fr) {
	     } else {
	       break;
	     }
	   }
	 }
 


      }

      if (!fd.fh && !fd.fr) {
	/* Error: neither File Streaming nor VFS API could open file */ 
	FrmCustomAlert(NoFileAlert, gPrefs.filename, NULL, NULL);
        
	if (gPrefs.mapprefs.select == SEL_USER) {

	  /* let user select another map */
	  FrmGotoForm(FileSelForm);
          eraseFilename(gPrefs.filename);
	
	} else {
	  noMap();
	  
	}

	/* clear filename */
	StrCopy(gPrefs.filename, "");

	break;
      }
    
      /*
       * At this point, we have a valid file handle!
       */
      
      deep = 1;
      ScrDisplayMode(scrDisplayModeGetSupportedDepths, NULL, NULL,
      			&odeep, NULL);
      ScrDisplayMode(scrDisplayModeGetSupportsColor, NULL, NULL,
      			NULL, &color);


// fs Support for 16 bit Display
      if (odeep & 0x8000)
	deep = 16; // pixel selection for Palm TX
      else 
      if (odeep & 0x80)
	deep = 8;
      else if (odeep & 8)
	deep = 4;
      else if (odeep & 2)
	deep = 2;

      ScrDisplayMode(scrDisplayModeGet, NULL, NULL, &odeep, NULL);

/*********************/
/* **TEST**** T3 *** */
/* * TX works too ** */
/*********************/
// use the current bpp of screen, overwrite the max possible bpp above
deep = odeep;
// hope we can use the calculation of the possible bpp for a device
// later in an option dialog, to allow selection by user

      /* read first 4 characters from file to determine file type */
      fread(fd, tbuf, 4);
      rewind(fd);

      if (!strncmp(tbuf, "\211PNG", 4)) {
	/* map file is a PNG */ 
	/* check if ZLib is installed */
	if (DmFindDatabase (0, /* card no. */ "ZLib") == 0) {
          FrmAlert(NoZLibAlert);
          if (gPrefs.mapprefs.select == SEL_USER) {
	    FrmGotoForm(FileSelForm);
	  } else {
	    gPrefs.mapprefs.fetch = FETCH_STOP;
	    StrCopy(gPrefs.filename, "");
	  }
	  eraseFilename(gPrefs.filename);
	  return true;
	}

	if (doPNG(fd)) {
	  FrmAlert(AbortPNGAlert);
	  eraseFilename(gPrefs.filename);
          if (gPrefs.mapprefs.select == SEL_USER) {
	    FrmGotoForm(FileSelForm);
	  } else {
	    gPrefs.mapprefs.fetch = FETCH_STOP;
	    noMap();
	  }         
	  return true;
	}
      
      } else if (!strncmp(tbuf, "GIF8", 4)) {
	/* map file is a GIF */
	if (doGIF(fd)) {
	  cleanGIF();
	  FrmAlert(AbortGIFAlert);
	  eraseFilename(gPrefs.filename);
          if (gPrefs.mapprefs.select == SEL_USER) {
	    FrmGotoForm(FileSelForm);
	  } else {
	    gPrefs.mapprefs.fetch = FETCH_STOP;
	    noMap();
	  }
	  return true; 
	}
      
      } else {
        /* map file is neither PNG nor GIF */
	FrmAlert(WrongFileAlert);
	eraseFilename(gPrefs.filename); 
	if (gPrefs.mapprefs.select == SEL_USER) {
          FrmGotoForm(FileSelForm);
	} else {
	  gPrefs.mapprefs.fetch = FETCH_STOP;
	  noMap();
	}
	return true;
      }
      decode_ok = true;
      
#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "decode ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif
      /* determine map parameters */
      map_valid = georef_filename(gPrefs.filename, &map_lat, &map_lon, &map_scale);
      
#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "map valid ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif      /* allow fetching of new map if required */
      gPrefs.mapprefs.fetch = FETCH_READY;

#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "mappref fetch ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif      /* save draw state (if color mode) */
      if (color) {
	WinPushDrawState();
      }
#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "push draw ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif
      if (gHdFtrSet) {
        WinSetCoordinateSystem(kCoordinatesNative);
#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "Hd coordsys ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
// (fs) next the Palm TX resets, special coding for 16bit display.
// Err
}
#endif
      } 

// TX resets at this block every time!
// so don't do ScrDisplayMode for 16bpp screen - could be a Palm TX
// this is only a work around
if (deep < 16) /* (fs) don't reset my TX with this */
{

      /* set new display depth */
      // next line resets the Palm TX, every bpp does this!!!! (fs 2007)
      ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &deep, NULL);
#ifdef VERBOSE
{
  Char		tmp[60];
  
  StrPrintF(tmp, "newdepth ok - not for Palm TX");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(250);
}
#endif
} /* (fs) now the Palm TX is still alive */



      WinEraseWindow();

#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "erase ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif

// (fs) TX in 16bit does not like a pallete, many resets - but not always, can't figure out what happens
// check deep first, and set palette for displaydepth < 16 bits
// err
  if (deep < 16)
  {
      /* set custom color palette */
      if (color) 
      {
	WinPalette(winPaletteSet, 0, mapmax, map);
#ifdef VERBOSE
{
  Char		tmp[60];	
  
  StrPrintF(tmp, "palette ok - no 16 bit mode");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(150);
}
#endif
      } // if (color)
  } // if (deep < 16)

else
{

/* (fs) test with a palette, can't use it in 16bit mode */
/********************************************************
{

RGBColorType col [256];
int i;
// col = (RGBColorType*)malloc(sizeof(RGBColorType) * 256);
// Set color palette to match BGR233 /
for (i = 0; i < 256; i++)
{
col[i].index = 0;
col[i].r = (((i >> 0) & 0x7) << 5) | (((i >> 0) & 0x7)
<< 2) | (((i >> 0) & 0x7) >> 1);
col[i].g = (((i >> 3) & 0x7) << 5) | (((i >> 3) & 0x7)
<< 2) | (((i >> 3) & 0x7) >> 1);
col[i].b = (((i >> 6) & 0x3) << 6) | (((i >> 6) & 0x3)
<< 4) | (((i >> 6) & 0x3) << 2) | ((i >> 6) & 0x3);
}
WinPalette(winPaletteSet, 0,256, col);
WinSetForeColorRGB(NULL,col); 

}
***********************************************************   switched  OFF */




/***********************************************************   switched  OFF

// (fs) set palette to default
// set default color palette works, but has no effect

	WinPalette(winPaletteSetToDefault, 0, 255, NULL);

#ifdef VERBOSE
{
  Char		tmp[60];	
  
  StrPrintF(tmp, "default palette in 16 bit mode");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(150);
}
#endif

***********************************************************   switched  OFF */



} // else deep < 16

      WinGetDisplayExtent(&maxx, &maxy);

      /*
       * Image size is set by doGIF/doPNG
       * w: image width
       * h: image height
       *
       * The image offset (xo, yo) is calculated to center the
       * image on the display. 
       */
      xo = w / 2 - maxx / 2;
      yo = h / 2 - maxy / 2;
      dx = 0;
      dy = 0;
      mag = 0;
      refresh = 1;
      dm = mag;
            
      if (color) {
        WinPopDrawState();
      }

      handled = true;
      /* frmOpenEvent */
      break;
   
    case keyDownEvent:
#if 0
      abmh_debug("keyDownEvent", gFiveWayAvail, "");

      if ( gFiveWayAvail ) {
	if (FiveWayCenterPressed(eventP)) {
          /* toggle FiveWay-Mode */
	  //          fiveway_mode = (fiveway_mode == FW_ZOOM) ? FW_PAN : FW_ZOOM;

	  if (gGPSData.scroll) {
	    abmh_debug("scroll", scroll_scale, "");

	    scroll_scale += 3;
	    if (scroll_scale > 16) {
	      gGPSData.scroll = 0;	     
	    }
	  } else {
	    gGPSData.scroll = 1;
	    scroll_scale = 1;
	    
	    if (!gGPSData.valid) {
	      gGPSData.valid = 1;
	      gGPSData.lat = map_lat;
	      gGPSData.lon = map_lon;
	    }
	  }
          handled = true;

	} else if (FiveWayKeyPressed(eventP, Right)) {          /* Pan Right */
	  //	  xo += maxx / 2;
	  //	  refresh = 1;
	  if (gGPSData.valid) {
	    gGPSData.lon += map_scale/8859.0*0.013438/2/scroll_scale;
	  }
	  handled = true;

	} else if (FiveWayKeyPressed(eventP, Left)) {          /* Pan Left */
	  //	  xo += maxx / 2;
	  //	  refresh = 1;
	  if (gGPSData.scroll) {
	    gGPSData.lon -= map_scale/8859.0*0.013438/2/scroll_scale;
	  }
	  handled = true;


	} else if (FiveWayKeyPressed(eventP, Up)) {
          /* Pan Up */
	    //	  yo -= maxx / 2;
	    //	  refresh = 1;
	    //	  handled = true
	  if (gGPSData.scroll) {
	    gGPSData.lat += map_scale/8859.0*0.008986/2/scroll_scale;
	  }


	} else if (FiveWayKeyPressed(eventP, Down)) {
	  /* Pan Down */
	    //	  yo += maxx / 2;
	    //	  refresh = 1;
	    //          handled = true;
	  if (gGPSData.scroll) {
	    gGPSData.lat -= map_scale/8859.0*0.008986/3/scroll_scale;
	  }

	}
      }

#endif

      if (!handled && TxtGlueCharIsVirtual(eventP->data.keyDown.modifiers,
                                           eventP->data.keyDown.chr)) {

	abmh_debug("eventP->data.keyDown.chr", eventP->data.keyDown.chr, "");

	switch (eventP->data.keyDown.chr) {
	  
#if 0
	  /* Zoom Display */
	  case '1'...'8':
            mag = event.data.keyDown.chr - '0';
            refresh = 2;
            break;
	  case pageUpChr:
            mag++;
            refresh = 2;
            break;
	  case pageDownChr:
            mag--;
            refresh = 2;
            break;

	  case vchrLaunch:
            event.eType = appStopEvent;
	    EvtAddEventToQueue(&event);
            break;
#endif

	case 317:
	  if (gGPSData.scroll) {
	    abmh_debug("scroll", scroll_scale, "");

	    if (scroll_scale < 4) {
	      scroll_scale *= 4;
	    } else {
	      scroll_scale *= 2;
	    }

	    if (scroll_scale > 16) {
	      gGPSData.scroll = 0;	     
	    }
	  } else {
	    gGPSData.scroll = 1;
	    scroll_scale = 1;
	    
	    if (!gGPSData.valid) {
	      gGPSData.valid = 1;
	      old_lat = gGPSData.lat = map_lat;
	      old_lon = gGPSData.lon = map_lon;
	    } else {
	      old_lat = gGPSData.lat;
	      old_lon = gGPSData.lon;
	    }
	  }
          handled = true;
	  scroll_redraw_i = scroll_redraw_count;
	  break;
	case 309:
	  // ->
	  if (gGPSData.valid) {
	    gGPSData.lon += map_scale/8859.0*0.013438/2/scroll_scale;
	    goto update_pos;
	  }
	  handled = true;
	  scroll_redraw_i = scroll_redraw_count;
	  break;
	case 308:
	  // <-
	  if (gGPSData.scroll) {
	    gGPSData.lon -= map_scale/8859.0*0.013438/2/scroll_scale;
	    goto update_pos;
	  }
	  handled = true;
	  scroll_redraw_i = scroll_redraw_count;
	  break;
	case 11:
          /* Pan Up */
	  if (gGPSData.scroll) {
	    gGPSData.lat += map_scale/8859.0*0.008986/2/scroll_scale;
	    goto update_pos;
	  }
	  handled = true;
	  scroll_redraw_i = scroll_redraw_count;
	  break;
	case 12:
	  /* Pan Down */
	  if (gGPSData.scroll) {
	    gGPSData.lat -= map_scale/8859.0*0.008986/2/scroll_scale;
	    goto update_pos;
	  }
	  handled = true;
	  scroll_redraw_i = scroll_redraw_count;
	  break;
	} /* switch (eventP->data.keyDown.chr) */
	handled = true;
      }
      break;

    
    /* Pan Display */
    case penDownEvent:
    map_redraw:
      {
	Coord	px = eventP->screenX;
	Coord	py = eventP->screenY;

	mapformwayptredraw = 0;

        /*
	 * Scale Standard screen coordinates to Native coordinates
	 * if using High-Density Display.
	 */
	if (color && gHdFtrSet) {
	  WinPushDrawState();
	  WinSetCoordinateSystem(kCoordinatesNative);
	  px = WinScaleCoord(px, false /* ceiling */);
	  py = WinScaleCoord(py, false /* ceiling */);
	  WinPopDrawState();
	}
      	
	if (py > maxy) {
          handled = true;
	  break;
	}

	xo += px - maxx / 2;
	yo += py - maxy / 2;
	refresh = 1;
      }
      handled = true;
      break;

    case menuEvent:
      switch (eventP->data.menu.itemID) {
        case PositionMenu:
	  FrmGotoForm(GPSMainForm);
	  handled = true;
	  break;

        case SkyviewMenu:
	  FrmGotoForm(SkyviewForm);
	  handled = true;
	  break;

	case NavigationMenu:
	  FrmGotoForm(NavigationForm);
	  handled = true;
	  break;

        case MiscOptsMenu:
	  FrmGotoForm(MiscOptsForm);
	  handled = true;
	  break;
	       
        case MapOptsMenu:
	  FrmGotoForm(MapOptsForm);
	  handled = true;
	  break;

	case PortsMenu:
	  if (gNewSerialManager) {
	    FrmGotoForm(GPSPortForm);
	  }
	  handled = true;
	  break;
	    
        case WaypointMenu:
	  FrmGotoForm(WaypointForm);
	  handled = true;
	  break;
	       
        case RouteMenu:
	  FrmGotoForm(RouteForm);
	  handled = true;
	  break;

        case TrackMenu:
	  FrmGotoForm(TrackForm);
	  handled = true;
	  break;
	       	       
	case PosclipMenu:
	  handled = Position2Clipboard(gGPSData, gPrefs.units);
	  break;

	case PosgeodbMenu:
	  handled = Position2Geodb(gGPSData);
	  break;

	case AboutMenu:
	  FrmGotoForm(AboutForm);
	  handled = true;
	  break;
	      	       
#ifdef DEBUG_FORM
        case DebugMenu:
	  FrmGotoForm(DebugForm);
	  handled = true;
	  break;
#endif

	case FileOpenMenu:
	  FrmGotoForm(FileSelForm);
	  handled = true;
	  break;
      
        case FetchMapMenu:
	  /* allow fetching of new map as soon as position is available */
	  gPrefs.mapprefs.fetch = FETCH_READY;

	  /* force searching for new map (see nilEvent) */
	  StrCopy(gPrefs.filename, "");

	  handled = true;
	  break;	  
	  
      } /* switch (eventP->data.menu.itemID) */
      break;

    case nilEvent:
    update_pos:
      handled = true;

scroll_redraw_i++;
if (gGPSData.scroll) {
  if (scroll_redraw_i > scroll_redraw_count) {
    scroll_redraw_i = 0;
    break;
  }
 }
      
      /* throw away anything in the buffer-- we want fresh data */
      DoReceiveFlush(gPortID, 1);

      if (!gGPSData.scroll) {
	/* read data from GPS */
	updatedDisplay = ReadFromGPS();

	if (gGPSData.valid && gPrefs.act_rte.valid && gPrefs.auto_wpt) {
	  /* update active waypoint */
	  UpdateActWpt(gGPSData.lat, gGPSData.lon, false /* init */);
	}

	/* Active Waypoint approach detection and display */
	NotifyApproach();
      
	/* Track Logging */
	if (gPrefs.act_trk.log_state) {
	  Boolean new_trk;

	  if (TrackIntervalCheck(false /* init */, &new_trk)) {
	    TrackWriteLog(new_trk);
	  }
	}		
      }

      if (map_valid && decode_ok) {
	/* map parameters o.k. and map decoded */ 
	
	/* calculate crosshair offset from map center */
	calcxy(&x, &y, gGPSData.lon, gGPSData.lat, map_lon, map_lat, map_scale);

	/* calculate screen coordinates from offset */
	x = dx + dw/2 + x/mag;
	y = dy + dh/2 - y/mag;
      }

      /*----------------------------------------------------
       * Map has been decoded successfully
       *----------------------------------------------------*/
      if (decode_ok) {
      

	// ABMH waypoints
#if 0
max_rows = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);
      rec = 0;
  /* seek to record at top of the table */
  DmSeekRecordInCategory(gWaypointDB, &rec /* index */, top /* offset */,
    dmSeekForward, dmAllCategories);
      /* get next record */
      waypointDBEntry = DmQueryNextInCategory(gWaypointDB, &rec, 
        dmAllCategories);

      /* unpack waypoint */
      packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
      UnpackWaypoint(&waypoint, packed_waypoint);
   /* convert semicircles to floating point */
    lat = semi2deg(waypoint.posn.lat);
    lon = semi2deg(waypoint.posn.lon);

      MemHandleUnlock(waypointDBEntry);


#endif

	WinPushDrawState();
	WinSetCoordinateSystem(kCoordinatesNative);


	mapformwayptredraw--;
      if (gPrefs.mapdispwayptname &&
	  (mapformwayptredraw <= 0)) {

	static write_kml = 1;

	UInt16		recc;			/* record nr */
	UInt16		rec;			/* record nr */
	UInt16          numrec;
	MemHandle		waypointDBEntry;	/* waypoint record handle */
	packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
	waypoint_t		waypoint;		/* unpacked waypoint */

	double lat, lon;
	Int32 x, y;
	int i;
	Char           ident2[7];

	// for kml writting only
	Err err = errNone;
	UInt32 volumeIterator = 0;
	UInt16 firstVolume = 0;
	FileRef f;
	UInt8 bufferP[255];
	UInt32 written = 0;
	// /kml writting


	mapformwayptredraw = 15;


	if (write_kml) {
	  // Get first card
	  volumeIterator = vfsIteratorStart;
	  err = VFSVolumeEnumerate(&firstVolume, &volumeIterator);
	  if (!(err == expErrEnumerationEmpty)) {
	    f = NULL;
	    err = VFSFileOpen(firstVolume, "/gps4wayp.kml", vfsModeWrite|vfsModeCreate|vfsModeTruncate, &f);
	    
	    StrCopy(bufferP, "<kml xmlns=\"http://earth.google.com/kml/2.0\">\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	    StrCopy(bufferP, "<Folder>\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	    StrCopy(bufferP, "  <name>GPS4Treo Waypts</name>\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	    StrCopy(bufferP, "  <open>1</open>\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);

	    StrCopy(bufferP, "\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	  } else {
	    write_kml = 0;
	  }
	}


	rec = 0;
	/* seek to record at top of the table */
	DmSeekRecordInCategory(gWaypointDB, &rec, 0,
			       dmSeekForward,
			       dmAllCategories);

	numrec = DmNumRecordsInCategory(gWaypointDB,
					dmAllCategories);

	//      abmh_debug("WptDis00", numrec, "");
	for (recc = 0; recc < numrec; recc++) {
	  rec = recc;
	  /* get next record */
	  waypointDBEntry = DmQueryNextInCategory(gWaypointDB,
						  &rec, 
						  dmAllCategories);
	  /* unpack waypoint */
	  packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
	  UnpackWaypoint(&waypoint, packed_waypoint);
	  /* convert semicircles to floating point */
	  lat = semi2deg(waypoint.posn.lat);
	  lon = semi2deg(waypoint.posn.lon);

	  /* calculate crosshair offset from map center */
	  calcxy(&x, &y, lon, lat,
		 map_lon, map_lat, map_scale);

	  /* calculate screen coordinates from offset */
	  x = dx + dw/2 + x/mag;
	  y = dy + dh/2 - y/mag;

	    for (i = 0; i < 6; i++) {
	      ident2[i] =  waypoint.ident[i];
	      if (waypoint.ident[i] == 0) {
		break;
	      }
	    }
	    ident2[6] = 0;

	    //	  abmh_debug("WptDis01x", x, "");
	    //	  abmh_debug("WptDis01y", y, "");
	    WinDrawChars(ident2, StrLen(ident2), x, y);	  	   	  
	  if (ON_MAP_WAYP(x,y)) {

	    //	  abmh_debug("WptDis01", 0, "");
	    WinDrawChars(ident2, StrLen(ident2), x, y);	  	   	  
	  }
	  //	  abmh_debug("WptDis01", rec, ident2);



	  if (write_kml) {
	    StrCopy(bufferP, "  <Placemark>\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	    
	    StrCopy(bufferP, "    <name>");
	    StrCat(bufferP, ident2);
	    if ((waypoint.cmnt[0] != '<') && 
		(waypoint.cmnt[0] != 0)) {
	      StrCat(bufferP, " - ");	  
	      StrCat(bufferP, waypoint.cmnt);
	    }
	    StrCat(bufferP, "</name>\n");	  
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);

	    StrCopy(bufferP, "    <Point><coordinates>");
	    format_number(lon, 7, &bufferP[StrLen(bufferP)]);
	    StrCat(bufferP, ",");
	    format_number(lat, 7, &bufferP[StrLen(bufferP)]);
	    StrCat(bufferP, ",0</coordinates></Point>\n");	  
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);

	    StrCopy(bufferP, "  </Placemark>\n");
	    VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	  }

	  MemHandleUnlock(waypointDBEntry);	  
	}

	if (write_kml) {
	  write_kml = 0;
	  StrCopy(bufferP, "</Folder>\n");
	  VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	  StrCopy(bufferP, "</kml>\n");
	  VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	  VFSFileClose(f);  
	}
      }

	// ABMH clock & distance
	{
	  UInt32	seconds;		/* current time */
	  DateTimeType	dt;			/* date/time structure */
	  UInt8 bufferP[255];
	  double		dist;			/* distance to waypoint */
	  Char	dist_str[8];

	  static double		wpt_lat;		/* lat of waypoint */
	  static double		wpt_lon;		/* lon of waypoint */

	  static Int32		trk;			/* old track */
	  PointType	p1;		/* point 1 */  
	  int r, lx, ly;
	  //see function beginning	  double		brg;			/* bearing */
	  double		turn;			/* turn */
	  double	tmp;


	  seconds = TimGetSeconds();
	  
	  TimSecondsToDateTime(seconds, &dt);
	  
	  /* write time to string */
	  TimeToAscii(dt.hour, dt.minute,
		      tfColon24h,
		      &bufferP);

	  if (gGPSData.valid) {
	    if (gPrefs.act_rte.valid || gPrefs.act_wpt.valid) {
	      /*  get waypt position */
	      get_wpt(&wpt_lat, &wpt_lon);


	      /*
	       * Distance to Waypoint
	       */
	      if (gGPSData.scroll) {
		StrCat(bufferP, " ** scroll **");
		dist = gc_dist_sphere(gGPSData.lat, gGPSData.lon, old_lat, old_lon);	      } else {
		dist = gc_dist_sphere(gGPSData.lat, gGPSData.lon, wpt_lat, wpt_lon);
	      }
	      /* calculate distance in km */
	      dist = dist * calcR(gGPSData.lat) / 1000.0;
	      dist *= dst_c[gPrefs.units.dst_unit].conv;
	      	      
	      /* adapt number of decimals to value */
	      if (dist < 2.0) {
		//    format_number(dist, 3, dist_str);
		//    ABMH
		format_number(dist, 2, dist_str);
	      } else if (dist < 10.0) {
		format_number(dist, 2, dist_str);
	      } else if (dist < 100.0) {
		format_number(dist, 1, dist_str);
	      } else {
		format_number(dist, 0, dist_str);
	      }
	      
	      /* append unit to string */
	      StrCat(dist_str, dst_c[gPrefs.units.dst_unit].unit);
	      
	      StrCat(bufferP, " ");
	      StrCat(bufferP, dist_str);

	      //	      WinPushDrawState();
	      //	      WinSetCoordinateSystem(kCoordinatesNative);
	      // show the direction to target
	      if (gGPSData.scroll) {
	      } else {
		/*
		 * Remove old pointer
		 */
		r = 50;
		p1.x = 160;
		p1.y = 160;
		lx = p1.x + r * icos(trk) / 32768;
		ly = p1.y + r * isin(trk) / 32768;
		
		// abmh
		WinEraseLine(p1.x, p1.y, lx, ly);
		WinEraseLine(p1.x+1, p1.y, lx+1, ly);
		WinEraseLine(p1.x, p1.y+1, lx, ly+1);
		
		
		/*
		 * Bearing (= Direction to Waypoint)
		 */
		brg = gc_course_sphere(gGPSData.lat, gGPSData.lon, wpt_lat, wpt_lon);
		brg = rad2deg(brg);
		
		/*
		 * Turn = bearing - track (pos: right/neg: left)
		 */
		turn = brg - gGPSData.cmg;
		trk = (Int32)turn - 90;

		trk = (Int32)(brg - gGPSData.cmg) - 90;
		
		
		lx = p1.x + r * icos(trk) / 32768;
		ly = p1.y + r * isin(trk) / 32768;
		
		//abmh
		WinDrawLine(p1.x, p1.y, lx, ly);
		WinDrawLine(p1.x, p1.y+1, lx, ly+1);
		WinDrawLine(p1.x+1, p1.y, lx+1, ly);
		

#if 0
		StrCat(bufferP, " ");
		if (turn < 0.0) {
		  format_number(-turn, 0, &bufferP[StrLen(bufferP)]);
		  //		bufferP[StrLen(bufferP) - 1] = 0;
		  StrCat(bufferP, "°L");
		} else {
		  format_number(turn, 0, &bufferP[StrLen(bufferP)]);
		  //		bufferP[StrLen(bufferP) - 1] = 0;
		  StrCat(bufferP, "°R");
		}
		//WinPopDrawState();
#endif


		StrCat(bufferP, " ");
		// Speed
		/* convert to preferred unit */
		tmp = gGPSData.sog * spd_c[gPrefs.units.spd_unit].conv;
		
		/* convert to string, adapt number of decimals to value */
		if (tmp < 50.0) {
		  format_number(tmp, 1, &bufferP[StrLen(bufferP)]);
		} else {
		  format_number(tmp, 0, &bufferP[StrLen(bufferP)]);
		}
		
		/* append unit to string */
		StrCat(bufferP, spd_c[gPrefs.units.spd_unit].unit);
	      }
	    }

	  }


	  StrCat(bufferP, "                  ");
	  if (y < 160) {
	    WinDrawChars(bufferP, StrLen(bufferP), 0, 300);	  	    
	  } else {
	    WinDrawChars(bufferP, StrLen(bufferP), 0, 0);	   /// 150
	  }
	  {
	    static lasty = 10;
	    if (lasty != (y < 160)) {
	      lasty = (y < 160);
	      WinPopDrawState();
	      goto map_redraw;
	    }
	  }
	  
	}

	WinPopDrawState();

	/*----------------------------------------------------
	 * Hide Crosshair
	 *  - if GPS Status is not o.k.
	 *  - if map parameters are unknown
	 *  - if crosshair is off the map
	 *  - if map is no longer in use
	 *----------------------------------------------------*/
	if (!gGPSData.valid || !map_valid || !ON_MAP(x, y) || (*gPrefs.filename == 0)) {

          if (cross_visible) {
	    if (color && gHdFtrSet) {
	      WinPushDrawState();
	      WinSetCoordinateSystem(kCoordinatesNative);
#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "pushdraw before do");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(250);
}
#endif	    }

	    /* redraw map at old crosshair position */
	    /*
            dobitmap(x_old-CROSS_X, y_old-CROSS_Y, mag,
	             x_old-CROSS_X, y_old-CROSS_Y, 
                     2*CROSS_X+1, 2*CROSS_Y+1, deep);
	    */
	    // ABMH: do not remove the entire crosshair but only the center of it
            dobitmap(x_old-3, y_old-3, mag,
	             x_old-3, y_old-3, 
                     2*3+1, 2*3+1, deep);


#ifdef VERBOSE
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "redraw ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(250);
}
#endif	    if (color && gHdFtrSet) {
	      WinPopDrawState();
	    }

	    cross_visible = false;
          } /* if (cross_visible) */
	  
	} else {
	  /*----------------------------------------------------
	   * Update Crosshair
	   *  - if GPS Fix Status is o.k.
	   *  - if map parameters are valid
	   *  - if crosshair is on the map
	   *----------------------------------------------------*/
          updateCross(x, y, &x_old, &y_old, map_lat, map_lon, map_scale, 
	              mag, deep);

	  if (gGPSData.scroll) {
	  } else {
	    int lx, ly;

	    // plot a line within the crosshair to indicate the bearing
	    lx = x + 16 * icos(gGPSData.cmg - 90) / 32768;
	    ly = y + 16 * isin(gGPSData.cmg - 90) / 32768;
	  
	    //abmh
	    rgbcolor.r = 255;
	    rgbcolor.g = rgbcolor.b = 0; // red
	    WinSetForeColor(WinRGBToIndex(&rgbcolor));
	    WinDrawLine(x, y, lx, ly);
	    WinDrawLine(x, y+1, lx, ly+1);
	    //	    WinDrawLine(x+1, y, lx+1, ly);
	  }

	  /*
	    optimation: to prevent push/pop to be used too often, this is done here and not in updateCross
	  */
	  if (gFtrSet35) {
	    WinPopDrawState();
	  }


	  /* save current crosshair position */	
	  x_old = x;
	  y_old = y;
	  cross_visible = true;

	} /* if (<hide_crosshair_conditions>) */
      } /* if (decode_ok) */
      
      /*----------------------------------------------------
       *
       * If current map is selcted by user,
       * keep it -- no matter if it shows
       * the current position or not.
       *
       *----------------------------------------------------*/
      if (gPrefs.mapprefs.select == SEL_USER)
        break; 
      
      /*----------------------------------------------------
       *
       * If the current position is unknown,
       * do neither search nor download map.
       *
       *----------------------------------------------------*/  
      if (!gGPSData.valid)
        break;


//abmh 2009
//gPrefs.mapprefs.fetch = FETCH_READY; // always try to search for a map

      switch (gPrefs.mapprefs.fetch) {
        case FETCH_READY:
	  /*------------------------------------------------
	   * Search map if position not on current map or
	   * current map scale does not match preferences
	   * any more or no map selected at all.
	   *------------------------------------------------*/
	  if ((map_valid && 
	       (!ON_MAP(x, y) || gPrefs.mapprefs.scale != map_scale)) ||
	      (StrCompare(gPrefs.filename, "") == 0)) {
	    if (search_map_all_dirs(gPrefs.filename, &gPrefs.filecard,
				    gGPSData.lon, gGPSData.lat, maxx, maxy)) {
	      /* suitable map found */
	      gPrefs.mapprefs.fetch = FETCH_READY;
	      FrmGotoForm(MapForm);
	      
	    } else {
	      /* no suitable map found */
	      switch (gPrefs.mapprefs.download) {
	        case DLD_O:
		  /* download off - keep current map */
		  break;
		
		case DLD_A:
		  /* download auto - fetch new map */
		  gPrefs.mapprefs.fetch = FETCH_START;
		  callFetchmap();
		  break;
		
		case DLD_Q:
		  /* download query - user has to allow/deny download */
		  if ( FrmAlert(QueryDldAlert) == 0 ) {
        	    /* download allowed - fetch map */
		    gPrefs.mapprefs.fetch = FETCH_START;
		    callFetchmap();

		  } else {
		    /* download denied - don't ask again, keep current map */
		    gPrefs.mapprefs.fetch = FETCH_STOP;

		  }
		  break;
	      } /* switch (gPrefs.mapprefs.download) */
	    } /* if ( search_map() ) */
	  } /* if (other map needed) */
	  
	  /* FETCH_READY */
	  break;
	  
	case FETCH_START:
	  /*------------------------------------------------
	   * A new map should just have been downloaded
	   * and has to be searched for.
	   *------------------------------------------------*/
	   if (search_map_all_dirs(gPrefs.filename, &gPrefs.filecard,
	        gGPSData.lon, gGPSData.lat, maxx, maxy)) {
	     /* suitable map found - change to new map */
	     gPrefs.mapprefs.fetch = FETCH_READY;
	     FrmGotoForm(MapForm);
	     
	   } else {
	     /*
	      * No suitable map has been found.
	      * This means download failed or
	      * a different map is needed now
	      * (position or scale changed).
	      * In either case, fetching is stopped now
	      * and the current map is kept.
	      */
	     gPrefs.mapprefs.fetch = FETCH_STOP;
	   
	  }
	  /* FETCH_START */
	  break;
	  
	case FETCH_STOP:
	  /*------------------------------------------------
	   * Search, but do not fetch map.
	   * To leave this state, the user has to select
	   * the menu entries Map->Fetch or Map->Select.
	   *------------------------------------------------*/
	  if (search_map(gPrefs.filename, &gPrefs.filecard,
	        gGPSData.lon, gGPSData.lat, maxx, maxy)) {
	    /* suitable map found - change to new map */
	    gPrefs.mapprefs.fetch = FETCH_READY;
	    FrmGotoForm(MapForm);
	    
	  }
	  /* FETCH_STOP */
	  break;
      
      } /* switch (gPrefs.mapprefs.fetch) */
        
      /* nilEvent */
      break;
   
    case frmCloseEvent:
      /* close file */
      fclose(fd);
      
      /* restore color palette (if color mode) */
      if (color) {
        WinPalette(winPaletteSetToDefault, 0, 256, NULL);
      }
      /* restore original display depth */
      ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &odeep, NULL);

      /* restore draw state */
      if (color) {
        WinPopDrawState();
      }
      if (decode_ok) {
        cleanPNG();                   /* works for both PNG & GIF */
        decode_ok = false;
      }
      MapFormDeinit(FrmGetActiveForm());
      FrmEraseForm(FrmGetActiveForm());
      handled = false;
      break;
  
  } /* switch (eventP->eType) */


  /*----------------------------------------------------------------*/
  /* Refresh Display                                                */
  /*----------------------------------------------------------------*/
  if (refresh) {
    /* save draw state (if color mode) */
    if (color) {
      WinPushDrawState();
    }
    if (gHdFtrSet) {
      WinSetCoordinateSystem(kCoordinatesNative);
    }

    /* limit magnification to valid range */
    if (mag < 1)
      mag = 1;
    if (mag > 8)
      mag = 8;

    /* limit offset of visible region within image to valid range */
    if (xo < 0)
      xo = 0;
    if (yo < 0)
      yo = 0;
    if (xo + maxx * mag > w)
      xo = w - (maxx + 1) * mag + 1;
    if (yo + maxy * mag > h)
      yo = h - (maxy + 1) * mag + 1;
    if (xo < 0)
      xo = 0;
    if (yo < 0)
      yo = 0;

    /* restore default background color */
    if (gFtrSet35) {
      WinPushDrawState();
      WinSetBackColor(UIColorGetTableEntryIndex(UIFormFill));
    }

    /* clear image display area */
    r.topLeft.x = r.topLeft.y = 0;
    r.extent.x = maxx;
    r.extent.y = maxy;
    WinEraseRectangle(&r, 0);

    if (gFtrSet35) {
      WinPopDrawState();
    }

    refresh = 0;

#ifdef VERBOSE
    WinDrawChars("Redrawing...", 12, 0, 0);
    {
      char tmp[20];

      StrPrintF(tmp, "%d;%d", xo, yo);
      WinDrawChars(tmp, StrLen(tmp), 0, 15);
    }
    SysTaskDelay(100);
#endif

    /* center small images on display */
    if (maxx > w/mag) {
      dx = (maxx - w/mag) / 2;
    } else {
      dx = 0;
    }

    if (maxy > h/mag) {
      dy = (maxy - h/mag) / 2;
    } else {
      dy = 0;
    }

    dw = (maxx > w / mag) ? w / mag : maxx;
    dh = (maxy > h / mag) ? h / mag : maxy;

    /* do full redraw */
    mapformwayptredraw = 0;
    dobitmap(xo,	/* x-offset of draw region within image */
	     yo,	/* y-offset of draw region within image */
	     mag,	/* sub-sampling factor */
	     dx,	/* x-offset of drawing area on screen */
	     dy,	/* y-offset of drawing area on screen */
	     dw,	/* x-extent of drawing area */
             dh,	/* y-extent of drawing area */
	     deep	/* color depth */
    );

    if (color) {
      WinPopDrawState();
    }
  }

  return handled;
} /* MapFormHandleEvent(EventPtr eventP) */
