/*****************************************************************************
 *
 * $RCSfile: MapOptsForm.c,v $
 *
 * GPS4Palm Map Options Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-01-30 21:22:30 $
 *
 * $Revision: 1.7 $
 *
 * $Log: MapOptsForm.c,v $
 * Revision 1.7  2005-01-30 21:22:30  mp
 * added PortsMenu event handling
 *
 * Revision 1.6  2004/11/30 20:45:02  mp
 * added TrackMenu event handling
 *
 * Revision 1.5  2004/11/24 21:18:17  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.4  2004/11/23 17:49:11  mp
 * removed unused variable
 *
 * Revision 1.3  2004/04/29 18:42:31  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.2  2004/02/28 17:21:43  mp
 * added menuEvent WaypointMenu
 *
 * Revision 1.1  2003/11/20 20:38:49  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "ResourceDefines.h"
#include "Serial.h"
#include "Sinetab.h"
#include "MapOptsForm.h"
#include "GPS.h"
#include "Utils.h"
#include "common.h"

extern UInt16		gAutoPwrOff;
extern Boolean		gFormOpened;
extern PrefsType	gPrefs;

/* Locals */
#define	SCALE_ITEMS	30

/* Caution: Make sure the entries match the list ScaleList in Resources.rcp Src/GPS4Palm.rcp ! */
static UInt32	ScaleMappings[SCALE_ITEMS] = {
  		  1000, 1500, 2000, 3000, 5000, 7500,
		  10000, 15000, 20000, 30000, 50000, 75000,
		  100000, 150000, 200000, 300000, 500000, 750000,
		  1000000, 1500000, 2000000, 3000000, 5000000, 7500000,
		  10000000, 15000000, 20000000, 30000000, 50000000, 75000000
		};


/* Static Functions */
static void MapOptsFormInit(void)		MAPOPTSFORM_SECTION;
static UInt16 mapToPosition(
  UInt32 *mappingArray, UInt32 value,
  UInt16 mappings, UInt32 defaultItem)		MAPOPTSFORM_SECTION;


/*
 * mapToPosition:
 *   Get list position from list item (reverse array lookup).
 *
 *
 * Parameters:
 * -----------
 *	mappingArray		Pointer to array
 *	value			Lookup value
 *	mappings		Total number of items in array
 *	defaultItem		Default Item
 *
 * Returns:
 * --------
 *	Position of 'value' in array or default item if not found.
 */
static UInt16 mapToPosition(UInt32 *mappingArray, UInt32 value,
					UInt16 mappings, UInt32 defaultItem)
{
  UInt16 i;

  i = 0;
  while (mappingArray[i] != value && i < mappings)
	i++;
  if (i >= mappings)
	return defaultItem;

  return i;
}


static 
void MapOptsFormInit(void)
{    
  gFormOpened = true;
}


/****************************************************************************/
/**
 * \brief	Map Options Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	event handled flag
 ****************************************************************************/
Boolean MapOptsFormHandleEvent(EventPtr eventP)
{
	FormPtr		frmP = FrmGetActiveForm();	/* Form Ptr */
	ListPtr		lstP;				/* List Ptr */
	UInt16		sel;				/* List Selection */
	Boolean		handled;

	handled = false;
	switch (eventP->eType)
	{
        case menuEvent:
          switch (eventP->data.menu.itemID) {
      
            case PositionMenu:
	      FrmGotoForm(GPSMainForm);
	      handled = true;
	      break;

            case SkyviewMenu:
	      FrmGotoForm(SkyviewForm);
	      handled = true;
	      break;

            case NavigationMenu:
	      FrmGotoForm(NavigationForm);
	      handled = true;
	      break;

            case MapMenu:
	      FrmGotoForm(MapForm);
	      handled = true;
	      break;

            case MiscOptsMenu:
	      FrmGotoForm(MiscOptsForm);
	      handled = true;
	      break;

	   case PortsMenu:
	     if (gNewSerialManager) {
	       FrmGotoForm(GPSPortForm);
	     }
	     handled = true;
	     break;

            case WaypointMenu:
	      FrmGotoForm(WaypointForm);
	      handled = true;
	      break;

            case RouteMenu:
	      FrmGotoForm(RouteForm);
	      handled = true;
	      break;
            
	    case TrackMenu:
	      FrmGotoForm(TrackForm);
	      handled = true;
	      break;

	    case AboutMenu:
	      FrmGotoForm(AboutForm);
	      handled = true;
	      break;

	  }
          break;
	  
	case ctlSelectEvent:
          /* Map Download Policy Push Buttons */
	  switch (eventP->data.ctlSelect.controlID) {
	    case DldAPushb:
	      gPrefs.mapprefs.download = DLD_A;
	      break;
	    
	    case DldQPushb:
	      gPrefs.mapprefs.download = DLD_Q;
	      break;
	      
	    case DldOPushb:
	      gPrefs.mapprefs.download = DLD_O;
	      break;

	  case MapDispWayptName:
	    gPrefs.mapdispwayptname = GetControlValueFromId(MapDispWayptName);

	  case MapDispWaypt:
	    gPrefs.mapdispwaypt = GetControlValueFromId(MapDispWaypt);

	  }
  
	  break;

	case popSelectEvent:
	  /* Map Scale Popup List */
	  sel = eventP->data.popSelect.selection;
	  frmP = FrmGetActiveForm();
	  FrmSetFocus(frmP, noFocus);
	  gPrefs.mapprefs.scale = ScaleMappings[sel];
	  handled = false;
	  break;

	
	case frmOpenEvent:
		MapOptsFormInit();
		FrmDrawForm(FrmGetActiveForm());
                
		// ABMH
		/* MapDispWayptName */
		FrmSetControlValue(frmP,
				   FrmGetObjectIndex(frmP, MapDispWayptName),
				   (gPrefs.mapdispwayptname) ? 1 : 0);

		/* MapDispWaypt */
		FrmSetControlValue(frmP,
				   FrmGetObjectIndex(frmP, MapDispWaypt),
				   (gPrefs.mapdispwaypt) ? 1 : 0);

		/* Set Map Download Policy Push Buttons from Preferences */ 
		switch (gPrefs.mapprefs.download) {
		  case DLD_A:
		    SetControlValueFromId(DldAPushb, 1);
		    break;
		    
		  case DLD_Q:
		    SetControlValueFromId(DldQPushb, 1);
		    break;
		    
		  case DLD_O:
		    SetControlValueFromId(DldOPushb, 1);
		    break;	    
		}

		/* Init the scale pop-up from Preferences */
		lstP = FrmGetObjectPtr(frmP, 
		         FrmGetObjectIndex(frmP, ScaleList));

		/*  Reverse array lookup (value -> index) */
		sel = mapToPosition(ScaleMappings,	   /* array */
				    gPrefs.mapprefs.scale, /* Prefs. value */
				    SCALE_ITEMS,	   /* array size */
				    20000);		   /* default value */
      
		LstSetSelection (lstP, sel);
		CtlSetLabel(FrmGetObjectPtr(frmP,
		 FrmGetObjectIndex(frmP, ScalePopup)),
        	   LstGetSelectionText(lstP, sel));	

		handled = true;
		break;

	case frmCloseEvent:
		/* Deallocate Fields' Text Memory */
	  	//SetFieldText(NavigationStatusField, NULL, false, true);
		break;
				
	}
	return(handled);
} /* MapOptsFormHandleEvent() */
