/*****************************************************************************
 *
 * $RCSfile: MiscOptsForm.c,v $
 *
 * GPS4Palm Misc Options Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-13 18:32:05 $
 *
 * $Revision: 1.13 $
 *
 * $Log: MiscOptsForm.c,v $
 * Revision 1.13  2005-05-13 18:32:05  mp
 * removed variable gAutoPwrOff
 *
 * Revision 1.12  2005/01/30 21:22:13  mp
 * added PortsMenu event handling
 *
 * Revision 1.11  2004/12/01 17:44:58  mp
 * added TrkModePopup, TrkInt(Time|DST)Popup, and TrkFullPopup
 *
 * Revision 1.10  2004/11/30 20:44:48  mp
 * added TrackMenu event handling
 *
 * Revision 1.9  2004/11/25 17:01:40  mp
 * added ApproachAlertCheckbox and ApproachSoundCheckbox handling
 *
 * Revision 1.8  2004/11/24 21:18:17  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.7  2004/11/23 17:48:52  mp
 * removed unused variables
 *
 * Revision 1.6  2004/11/19 17:57:11  mp
 * added AutoWaypointCheckbox and modified AutoPwrOffCheckbox
 *
 * Revision 1.5  2004/04/29 18:47:16  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.4  2004/02/28 17:21:43  mp
 * added menuEvent WaypointMenu
 *
 * Revision 1.3  2004/01/18 11:46:04  mp
 * added position display option UTM
 *
 * Revision 1.2  2003/11/20 20:50:06  mp
 * Added setting of Controls from Preferences at frmOpenEvent and
 * setting of Prefernces on ctlSelectEvent.
 * Removed nilEvent and obsolete Controls (Sermux and Backlight).
 *
 * Revision 1.1  2003/11/16 17:45:02  mp
 * initial version
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "ResourceDefines.h"
#include "Serial.h"
#include "Sinetab.h"
#include "MiscOptsForm.h"
#include "GPS.h"
#include "Utils.h"
#include "common.h"

extern Boolean		gFormOpened;
extern PrefsType	gPrefs;

/* Static Functions */
static void MiscOptsFormInit(void);

static 
void MiscOptsFormInit(void)
{  
  
  gFormOpened = true;
}


/****************************************************************************/
/**
 * \brief	Misc Options Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	event handled flag
 ****************************************************************************/
Boolean MiscOptsFormHandleEvent(EventPtr eventP)
{
  FormPtr		frmP = FrmGetActiveForm();	/* Form Ptr */
  ListPtr		lstP;				/* List Ptr */
  UInt16		n;				/* no. of items */
  Boolean		handled;			/* event handled flag */

  handled = false;
  switch (eventP->eType) {

    case frmOpenEvent:
      MiscOptsFormInit();
      FrmDrawForm(FrmGetActiveForm());

      /* Set Auto Power-Off Check Box from Preferences */
      FrmSetControlValue(frmP,
	FrmGetObjectIndex(frmP, AutoPwrOffCheckbox), gPrefs.pwroff);

      /* Set Auto Waypoint Check Box from Preferences */
      FrmSetControlValue(frmP,
	FrmGetObjectIndex(frmP, AutoWaypointCheckbox),
	(gPrefs.auto_wpt) ? 1 : 0);

      /* Set Approach Alert Check Box from Preferences */
      FrmSetControlValue(frmP,
	FrmGetObjectIndex(frmP, ApproachAlertCheckbox),
	(gPrefs.approach_alert) ? 1 : 0);

      /* Set Approach Sound Check Box from Preferences */
      FrmSetControlValue(frmP,
	FrmGetObjectIndex(frmP, ApproachSoundCheckbox),
	(gPrefs.approach_sound) ? 1 : 0);

      /* Set Track Log Mode Popup from Preferences */
      lstP = GetObjectFromActiveForm(TrkModeList);
      LstSetSelection(lstP, gPrefs.trk_log.mode);
      CtlSetLabel(GetObjectFromActiveForm(TrkModePopup),
         LstGetSelectionText(lstP, gPrefs.trk_log.mode));

      /* Show Track Log Interval Popup from Preferences */
      switch (gPrefs.trk_log.mode) {
	case 0:
	  FrmHideObject(frmP, FrmGetObjectIndex(frmP, TrkIntTimePopup));
	  FrmShowObject(frmP, FrmGetObjectIndex(frmP, TrkIntDstPopup));
	  lstP = GetObjectFromActiveForm(TrkIntDstList);
	  LstSetSelection(lstP, gPrefs.trk_log.ival);
	  CtlSetLabel(GetObjectFromActiveForm(TrkIntDstPopup),
             LstGetSelectionText(lstP, gPrefs.trk_log.ival));	      
	  break;

	case 1:
	  FrmHideObject(frmP, FrmGetObjectIndex(frmP, TrkIntDstPopup));
	  FrmShowObject(frmP, FrmGetObjectIndex(frmP, TrkIntTimePopup));
	  lstP = GetObjectFromActiveForm(TrkIntTimeList);
	  LstSetSelection(lstP, gPrefs.trk_log.ival);
	  CtlSetLabel(GetObjectFromActiveForm(TrkIntTimePopup),
             LstGetSelectionText(lstP, gPrefs.trk_log.ival));	      
	  break;
      }

      /* Set Track Full Popup from Preferences */
      lstP = GetObjectFromActiveForm(TrkFullList);
      LstSetSelection(lstP, gPrefs.trk_log.wrap);
      CtlSetLabel(GetObjectFromActiveForm(TrkFullPopup),
         LstGetSelectionText(lstP, gPrefs.trk_log.wrap));

      /* Set Position Unit Push Buttons from Preferences */ 
      switch (gPrefs.units.pos_unit) {
	case POS_D:
	  SetControlValueFromId(PosDPushb, 1);
	  break;

	case POS_DM:
	  SetControlValueFromId(PosDMPushb, 1);
	  break;

	case POS_DMS:
	  SetControlValueFromId(PosDMSPushb, 1);
	  break;	    

	case POS_UTM:
	  SetControlValueFromId(PosUTMPushb, 1);
	  break;	    
      }

      /* Set Distance Unit Push Buttons from Preferences */ 
      switch (gPrefs.units.dst_unit) {
	case DST_KM:
	  SetControlValueFromId(DstKmPushb, 1);
	  break;

	case DST_MI:
	  SetControlValueFromId(DstMiPushb, 1);
	  break;

	case DST_NM:
	  SetControlValueFromId(DstNmPushb, 1);
	  break;	    
      }

      /* Set Altitude Unit Push Buttons from Preferences */ 
      switch (gPrefs.units.alt_unit) {
	case ALT_M:
	  SetControlValueFromId(AltMPushb, 1);
	  break;

	case ALT_FT:
	  SetControlValueFromId(AltFtPushb, 1);
	  break;
      }

      /* Set Speed Unit Push Buttons from Preferences */ 
      switch (gPrefs.units.spd_unit) {
	case SPD_KMH:
	  SetControlValueFromId(SpdKmhPushb, 1);
	  break;

	case SPD_MS:
	  SetControlValueFromId(SpdMsPushb, 1);
	  break;

	case SPD_MIH:
	  SetControlValueFromId(SpdMihPushb, 1);
	  break;

	case SPD_KTS:
	  SetControlValueFromId(SpdKtsPushb, 1);
	  break;
      }

      handled = true;
      break;

    case menuEvent:
      switch (eventP->data.menu.itemID) {

	case PositionMenu:
	  FrmGotoForm(GPSMainForm);
	  handled = true;
	  break;

	case SkyviewMenu:
	  FrmGotoForm(SkyviewForm);
	  handled = true;
	  break;

	case NavigationMenu:
	  FrmGotoForm(NavigationForm);
	  handled = true;
	  break;

	case MapMenu:
	  FrmGotoForm(MapForm);
	  handled = true;
	  break;

	case MapOptsMenu:
	  FrmGotoForm(MapOptsForm);
	  handled = true;
	  break;

	case PortsMenu:
	  if (gNewSerialManager) {
	    FrmGotoForm(GPSPortForm);
	  }
	  handled = true;
	  break;

	case WaypointMenu:
	  FrmGotoForm(WaypointForm);
	  handled = true;
	  break;

	case RouteMenu:
	  FrmGotoForm(RouteForm);
	  handled = true;
	  break;

	case TrackMenu:
	  FrmGotoForm(TrackForm);
	  handled = true;
	  break;

	case AboutMenu:
	  FrmGotoForm(AboutForm);
	  handled = true;
	  break;
      }
      break;

    case popSelectEvent:
      /* Popup List */

      /* Track Log Mode */
      if (eventP->data.popSelect.controlID == TrkModePopup) {
	gPrefs.trk_log.mode = eventP->data.popSelect.selection;

	/* Change Track Log Interval Popup */
	switch (gPrefs.trk_log.mode) {
	  case 0:
	    /* Distance */
	    FrmHideObject(frmP, FrmGetObjectIndex(frmP, TrkIntTimePopup));
	    FrmShowObject(frmP, FrmGetObjectIndex(frmP, TrkIntDstPopup));

            lstP = GetObjectFromActiveForm(TrkIntDstList);
	    n = LstGetNumberOfItems(lstP);

	    if (gPrefs.trk_log.ival > n-1) {
	      /* set to last index */
	      gPrefs.trk_log.ival = n-1;
	    }

	    /* Show Track Log Interval */
	    LstSetSelection(lstP, gPrefs.trk_log.ival);
	    CtlSetLabel(GetObjectFromActiveForm(TrkIntDstPopup),
               LstGetSelectionText(lstP, gPrefs.trk_log.ival));
	    break;

	  case 1:
	    /* Time */
	    FrmHideObject(frmP, FrmGetObjectIndex(frmP, TrkIntDstPopup));
	    FrmShowObject(frmP, FrmGetObjectIndex(frmP, TrkIntTimePopup));

	    lstP = GetObjectFromActiveForm(TrkIntTimeList);
	    n = LstGetNumberOfItems(lstP);

	    if (gPrefs.trk_log.ival > n-1) {
	      /* set to last index */
	      gPrefs.trk_log.ival = n-1;
	    }

            /* Show Track Log Interval */
	    LstSetSelection(lstP, gPrefs.trk_log.ival);
	    CtlSetLabel(GetObjectFromActiveForm(TrkIntTimePopup),
               LstGetSelectionText(lstP, gPrefs.trk_log.ival));	      
	    break;
	}          
      }

      /* Track Log Interval */
      if (eventP->data.popSelect.controlID == TrkIntTimePopup ||
          eventP->data.popSelect.controlID == TrkIntDstPopup) {
	gPrefs.trk_log.ival = eventP->data.popSelect.selection;
      }

      /* Track Log Full Option */
      if (eventP->data.popSelect.controlID == TrkFullPopup) {
	gPrefs.trk_log.wrap = eventP->data.popSelect.selection;
      }

      /* Note: popup needs to handle the event, too */
      break;
	  
    case ctlSelectEvent:
      switch (eventP->data.ctlSelect.controlID) {
	/* Auto Power-Off Check Box */
	case AutoPwrOffCheckbox:
	  gPrefs.pwroff = GetControlValueFromId(AutoPwrOffCheckbox);
	  break;

	case AutoWaypointCheckbox:
	  gPrefs.auto_wpt = GetControlValueFromId(AutoWaypointCheckbox);
	  break;

	case ApproachAlertCheckbox:
	  gPrefs.approach_alert = GetControlValueFromId(ApproachAlertCheckbox);
	  break;

	case ApproachSoundCheckbox:
	  gPrefs.approach_sound = GetControlValueFromId(ApproachSoundCheckbox);
	  break;

	/* Position Unit Push Buttons */
	case PosDPushb:
	  gPrefs.units.pos_unit = POS_D;
	  break;

	case PosDMPushb:
	  gPrefs.units.pos_unit = POS_DM;
	  break;

	case PosDMSPushb:
	  gPrefs.units.pos_unit = POS_DMS;
	  break;

	case PosUTMPushb:
	  gPrefs.units.pos_unit = POS_UTM;
	  break;
      }

      /* Distance Unit Push Buttons */
      switch (eventP->data.ctlSelect.controlID) {
	case DstKmPushb:
	  gPrefs.units.dst_unit = DST_KM;
	  break;

	case DstMiPushb:
	  gPrefs.units.dst_unit = DST_MI;
	  break;

	case DstNmPushb:
	  gPrefs.units.dst_unit = DST_NM;
	  break;
      }

      /* Altitude Unit Push Buttons */
      switch (eventP->data.ctlSelect.controlID) {
	case AltMPushb:
	  gPrefs.units.alt_unit = ALT_M;
	  break;

	case AltFtPushb:
	  gPrefs.units.alt_unit = ALT_FT;
	  break;

      }

      /* Speed Unit Push Buttons */
      switch (eventP->data.ctlSelect.controlID) {
	case SpdKmhPushb:
	  gPrefs.units.spd_unit = SPD_KMH;
	  break;

	case SpdMsPushb:
	  gPrefs.units.spd_unit = SPD_MS;
	  break;

	case SpdMihPushb:
	  gPrefs.units.spd_unit = SPD_MIH;
	  break;

	case SpdKtsPushb:
	  gPrefs.units.spd_unit = SPD_KTS;
	  break;
      }

      break;

    case frmCloseEvent:
      break;

  }
  return(handled);
} /* MiscOptsFormHandleEvent() */
