/*****************************************************************************
 *
 * $RCSfile: NavigationForm.c,v $
 *
 * GPS4Palm Navigation Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-06 13:38:34 $
 *
 * $Revision: 1.26 $
 *
 * $Log: NavigationForm.c,v $
 * Revision 1.26  2005-05-06 13:38:34  mp
 * removed assignment to gHdFtrSet
 *
 * Revision 1.25  2005/04/24 14:06:43  mp
 * added High-Density Display support
 *
 * Revision 1.24  2005/04/02 10:09:43  mp
 * modified Position2Clipboard() and Position2Geodb() to return 'handled'
 *
 * Revision 1.23  2005/04/02 07:26:59  mp
 * added PosgeodbMenu event handling
 *
 * Revision 1.22  2005/03/25 13:45:11  mp
 * added PosclipMenu event handling
 *
 * Revision 1.21  2005/01/23 18:40:10  mp
 * added PortsMenu event handling
 *
 * Revision 1.20  2004/12/10 20:03:04  mp
 * added waypoint selection with up/down keys,
 * replaced DmGet1Resource by DmGetResource
 *
 * Revision 1.19  2004/12/09 17:33:37  mp
 * replaced strings by string resources
 *
 * Revision 1.18  2004/12/08 20:49:22  mp
 * added track logging
 *
 * Revision 1.17  2004/11/30 20:44:06  mp
 * added TrackMenu event handling
 *
 * Revision 1.16  2004/11/27 10:20:31  mp
 * replaced error handling code by calling Die()
 *
 * Revision 1.15  2004/11/26 20:09:40  mp
 * added error handling for WinCreateOffscreenWindow(), fixed comments
 *
 * Revision 1.14  2004/11/25 20:06:20  mp
 * added NotifyApproach()
 *
 * Revision 1.13  2004/11/24 21:15:20  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.12  2004/11/23 17:48:33  mp
 * removed unused variables
 *
 * Revision 1.11  2004/11/20 15:51:25  mp
 * added display_wpt(), NavigationPrevWptButton/NavigationNextWptButton
 * handling, and automatic route waypoint selection
 *
 * Revision 1.10  2004/11/11 21:36:25  mp
 * implemented active Route Information
 *
 * Revision 1.9  2004/04/29 19:02:05  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.8  2004/03/16 20:00:42  mp
 * Fixed waypoint ID display. Added deletion of window handle upon
 * frmCloseEvent. Modified bearing pointer.
 *
 * Revision 1.7  2004/03/13 14:59:28  mp
 * added/modified data display
 *
 * Revision 1.6  2004/03/11 21:40:06  mp
 * added data display and DTK pointer
 *
 * Revision 1.5  2004/02/28 17:21:43  mp
 * added menuEvent WaypointMenu
 *
 * Revision 1.4  2003/11/20 20:45:01  mp
 * added status redraw on frmOpenEvent
 *
 * Revision 1.3  2003/11/18 20:55:40  mp
 * added MiscOptsMenu, added current Form to Application Preferences
 *
 * Revision 1.2  2003/10/20 17:25:50  mp
 * added field deallocation on frmCloseEvent
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "MathLib.h"
#include "ResourceDefines.h"
#include "Serial.h"
#include "Sinetab.h"
#include "NavigationForm.h"
#include "GPS.h"
#include "HandleMessage.h"
#include "Utils.h"
#include "Data.h"
#include "fp.h"
#include "geo.h"
#include "Clip.h"
#include "common.h"

/* Global Variables */

/** Navigation window handle for save-behind (Compass). */
WinHandle		gNavigationH;

extern PrefsType	gPrefs;			/* Preferences data structure */
extern DmOpenRef	gWaypointDB;		/* Data Base reference */
extern GPSType		gGPSData;		/* GPS Data */
extern Boolean		gHdFtrSet;		/* High-Density Display */

/** Constants for distance unit conversions and display */
extern const struct {
  char		*unit;
  double	conv;
} dst_c[];

/** Constants for speed unit conversions and display */
extern const struct {
  char		*unit;
  double	conv;
} spd_c[];

#ifndef WinDrawPixel
#define WinDrawPixel(x,y) WinDrawLine((x),(y),(x),(y))
#endif

/* Static Functions */
static void display_sog(void)					NAV_SECTION;
static void display_cmg(void)					NAV_SECTION;
static void display_dst(double dist)				NAV_SECTION;
static void display_trn(double turn)				NAV_SECTION;
static void display_vmg(double vmg)				NAV_SECTION;
static void display_wpt(double *lat, double *lon)		NAV_SECTION;
static void hours_to_hms(double hours, UInt16 *hh, UInt16 *mm,
  UInt16 *ss)							NAV_SECTION;
static void NavigationFormInit(void)				NAV_SECTION;
static void DrawNavigation(void)				NAV_SECTION;


// abmh
int abmh_debug(char *pos, unsigned long l, char *stri);



/*****************************************************************************
 * FUNCTION:	WayptEditFormInit
 *
 * DESCRIPTION: Navigation Form Init Function
 *
 * GLOBALS:	gFormOpened	-- indicates if a form (using GPS data)
 *				   has been opened
 ****************************************************************************/
static 
void NavigationFormInit(void)
{  
  gFormOpened = true;
}


/*****************************************************************************
 * FUNCTION:	hours_to_hms
 *
 * DESCRIPTION:	Convert fractional hours to integer hours, minutes and seconds 
 *
 * PARAMETERS:	hours		-- decimal hours
 *		hh		-- integer hours   (returned by ref)
 *		mm		-- integer minutes (returned by ref)
 *		ss		-- integer seconds (returned by ref)
 *
 * RETURNED:	%
 ****************************************************************************/
static
void hours_to_hms(double hours, UInt16 *hh, UInt16 *mm, UInt16 *ss)
{
  double	sec;		/* seconds */
  double	frac;		/* fractional part */
  double	ipart;		/* integer part */
  
  /* hours */
  frac = modf(hours, &ipart) * 60.0;
  *hh = (UInt16)ipart;
  
  /* minutes */
  sec = modf(frac, &ipart) * 60.0;
  *mm = (UInt16)ipart;
  
  /* seconds */
  *ss = (UInt16)sec;
}


/*****************************************************************************
 * FUNCTION:	DrawNavigation
 *
 * DESCRIPTION: Draw Navigation Form ("Compass") and save it to off-screen
 *		window.
 *
 * GLOBALS:	gNavigationH	-- Off-screen window handle
 *
 * PARAMETERS:	-
 *
 * RETURNED:	-
 ****************************************************************************/
static
void DrawNavigation(void)
{
  UInt16	x;		/* x-coordinate */
  UInt16	y;		/* y-coordinate */
  PointType	p1;		/* point 1 */
  PointType	p2;		/* point 2 */
  UInt16	r;		/* radius */
  UInt16	phi;		/* angle */
  UInt16	err;		/* error code */
  RectangleType	rect;		/* rectangle structure */

  if (gHdFtrSet) {
    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesNative);
  } 

  /* Outer circle */
  r    = (gHdFtrSet) ? WinScaleCoord(NAV_R2, false) : NAV_R2;
  p1.x = (gHdFtrSet) ? WinScaleCoord(NAV_X, false) : NAV_X;
  p1.y = (gHdFtrSet) ? WinScaleCoord(NAV_Y, false) : NAV_Y;
  for (phi=0; phi<360; phi++) {
    /* Sinus/Cosinus functions using table lookup */
    x = p1.x + r * icos(phi) / 32768;
    y = p1.y - r * isin(phi) / 32768;
   
    /*
     * Note:
     * For High-Density, a step size of one degree does not provide
     * enough points for a continuous circle. Connecting the points
     * with a line seems to work good enough. (and is still fast) 
     */
    if (phi == 0) {
      WinDrawPixel(x, y);
    } else {
      WinDrawLine(x, y, p2.x, p2.y);
    }
    p2.x = x;
    p2.y = y;
  }

  /* Inner circle */
  r = (gHdFtrSet) ? WinScaleCoord(NAV_R1, false) : NAV_R1;
  for (phi=0; phi<360; phi++) {
    /* Sinus/Cosinus functions using table lookup */
    x = p1.x + r * icos(phi) / 32768;
    y = p1.y - r * isin(phi) / 32768;
    
    WinDrawPixel(x, y);
  }

  /* Center */
  p1.x = NAV_X-2;
  p1.y = NAV_Y;
  p2.x = NAV_X+2;
  p2.y = NAV_Y;
  if (gHdFtrSet) {
    WinScalePoint(&p1, false);
    WinScalePoint(&p2, false);
  }
  WinDrawLine(p1.x, p1.y, p2.x, p2.y);

  p1.x = NAV_X;
  p1.y = NAV_Y-2;
  p2.x = NAV_X;
  p2.y = NAV_Y+2;
  if (gHdFtrSet) {
    WinScalePoint(&p1, false);
    WinScalePoint(&p2, false);
  }
  WinDrawLine(p1.x, p1.y, p2.x, p2.y);

  p1.x = 2 * NAV_R2;
  p1.y = 2 * NAV_R2;
  if (gHdFtrSet) WinScalePoint(&p1, false);
  gNavigationH = WinCreateOffscreenWindow(p1.x, p1.y, nativeFormat, &err);
  
  if (!gNavigationH) {
    /* error handling */
    if (gHdFtrSet) {
      WinPopDrawState();
    } 
    Die("Cannot create OffscreenWindow in NavigationForm!");
    return;
  }
  
  p1.x = NAV_X-NAV_R2;
  p1.y = NAV_Y-NAV_R2;
  if (gHdFtrSet) WinScalePoint(&p1, false);
  x = (gHdFtrSet) ? WinScaleCoord(2 * NAV_R2, false) : 2 * NAV_R2;
  y = (gHdFtrSet) ? WinScaleCoord(2 * NAV_R2, false) : 2 * NAV_R2;
  RctSetRectangle(&rect, p1.x, p1.y, x /* w */, y /* h */);

  WinCopyRectangle(	NULL, /* WinHandle srcWin, draw window */
	  		gNavigationH, /* WinHandle dstWin, draw window */
			&rect, /* RectangleType *srcRect */
			0, /* Coord destX */
			0, /* Coord destY */
			winPaint /* WinDrawOperation mode */);
  if (gHdFtrSet) {
    WinPopDrawState();
  } 
}


/*****************************************************************************
 * FUNCTION:	display_sog
 *
 * DESCRIPTION: Displays Speed Over Ground
 *
 * GLOBALS:	gPrefs		-- application preferences
 *		gGPSData	-- GPS data
 *
 * PARAMETERS:	%
 *
 * RETURNED:	%
 ****************************************************************************/
static
void display_sog(void)
{
  double	tmp;
  Char		sog_str[12];

  /* convert to preferred unit */
  tmp = gGPSData.sog * spd_c[gPrefs.units.spd_unit].conv;

  /* convert to string, adapt number of decimals to value */
  if (tmp < 100.0) {
    format_number(tmp, 2, sog_str);
  } else if (tmp < 1000.0) {
    format_number(tmp, 1, sog_str);
  } else {
    format_number(tmp, 0, sog_str);
  }
  
  /* append unit to string */
  StrCat(sog_str, " ");
  StrCat(sog_str, spd_c[gPrefs.units.spd_unit].unit);

  /* display string */
  SetFieldText(NavigationSogField, sog_str, false, true);
}


/*****************************************************************************
 * FUNCTION:	display_cmg
 *
 * DESCRIPTION: Displays Course Made Good (= Course Over Ground)
 *
 * GLOBALS:	gGPSData	-- GPS data
 *
 * PARAMETERS:	%
 *
 * RETURNED:	%
 ****************************************************************************/
static
void display_cmg(void)
{
  Char		cmg_str[12];
  
  /* convert to string */
  format_number(gGPSData.cmg, 1, cmg_str);

  /* append unit to string */
  StrCat(cmg_str, "");

  /* display string */
  SetFieldText(NavigationCmgField, cmg_str, false, true);
}


/*****************************************************************************
 * FUNCTION:	display_dst
 *
 * DESCRIPTION: Displays Distance To Waypoint
 *
 * GLOBALS:	gPrefs		-- application preferences
 *
 * PARAMETERS:	dist		-- distance in km
 *
 * RETURNED:	%
 ****************************************************************************/
static
void display_dst(double dist)
{
  Char	dist_str[8];

  /* convert to preferred unit */
  dist *= dst_c[gPrefs.units.dst_unit].conv;

  /* adapt number of decimals to value */
  if (dist < 2.0) {
    //    format_number(dist, 3, dist_str);
    //    ABMH
    format_number(dist, 2, dist_str);
  } else if (dist < 10.0) {
    format_number(dist, 2, dist_str);
  } else if (dist < 100.0) {
    format_number(dist, 1, dist_str);
  } else {
    format_number(dist, 0, dist_str);
  }
	
  /* append unit to string */
  StrCat(dist_str, " ");
  StrCat(dist_str, dst_c[gPrefs.units.dst_unit].unit);

  /* display DST string */
  SetFieldText(NavigationDstField, dist_str, false, true);
}


/*****************************************************************************
 * FUNCTION:	display_trn
 *
 * DESCRIPTION: Displays Turn Angle
 *
 * PARAMETERS:	turn		-- turn in degrees, (pos: right/neg: left)
 *
 * RETURNED:	%
 ****************************************************************************/
static
void display_trn(double turn)
{
  Char	trn_str[8];
  
  if (turn < 0.0) {
    format_number(-turn, 1, trn_str);
    StrCat(trn_str, "L");
  } else {
    format_number(turn, 1, trn_str);
    StrCat(trn_str, "R");
  }
  SetFieldText(NavigationTrnField, trn_str, false, true);
}


/*****************************************************************************
 * FUNCTION:	display_vmg
 *
 * DESCRIPTION: Displays Velocity Made Good
 *
 * GLOBALS:	gPrefs		-- application preferences
 *
 * PARAMETERS:	vmg		-- velocity made good (in kts)
 *
 * RETURNED:	%
 ****************************************************************************/
static
void display_vmg(double vmg)
{
  Char		vmg_str[10];
  double	tmp;

  /* convert to preferred unit */
  tmp = vmg * spd_c[gPrefs.units.spd_unit].conv;
  
  /* convert to string, adapt number of decimals to value */
  if (tmp < 100.0) {
    format_number(tmp, 2, vmg_str);
  } else if (tmp < 1000.0) {
    format_number(tmp, 1, vmg_str);
  } else {
    format_number(tmp, 0, vmg_str);
  }
  
  /* append unit to string */
  StrCat(vmg_str, " ");
  StrCat(vmg_str, spd_c[gPrefs.units.spd_unit].unit);

  /* display string */
  SetFieldText(NavigationVmgField, vmg_str, false, true);	  
}


/****************************************************************************/
/**
 * \brief	Display waypoint ID and update waypoint position
 *
 * \param	lat		latitude returned by reference
 * \param	lon		longitude returned by reference
 *
 ****************************************************************************/
static
void display_wpt(double *lat, double *lon)
{
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  UInt16		rec;			/* record index */
  Err			err;			/* error code */
  Char			tmp[7];			/* waypoint identifier */ 
  MemHandle		strH;			/* string handle */
  MemPtr		strP;			/* string ptr */
  
  /* default "Waypoint Identifier" */ 
  /* get string from resource */
  strH = DmGetResource(strRsc, NoWptStr);

  if (strH != NULL) {
    /* string handle valid, lock it */
    strP = MemHandleLock(strH);
  
    /* copy string */
    StrCopy(tmp, strP);

    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  }

  /* display active waypoint */
  err = DmFindRecordByID(gWaypointDB, gPrefs.act_wpt.waypointID,
    &rec);
	
  if (!err) {
    /* active waypoint found */
    waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
    packed_waypoint =
      (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    MemSet((Char *)tmp, 7, 0);
    StrNCopy(tmp, (Char *)&waypoint.ident, 6);
    *lat = semi2deg(waypoint.posn.lat);
    *lon = semi2deg(waypoint.posn.lon);
    MemHandleUnlock(waypointDBEntry);
  }

  CtlSetLabel((ControlPtr)GetObjectFromActiveForm(NavigationWptIDSel), tmp);
}


/****************************************************************************/
/**
 * \brief	Navigation Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean NavigationFormHandleEvent(EventPtr eventP)
{
  Boolean		handled;		/* flag: event handled */
  Boolean		updatedDisplay;		/* flag: updated display */
  Err			err;			/* error code */
  UInt16		rec;			/* record index */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  route_t		*route;			/* route */
  MemHandle		routeDBEntry;		/* route record handle */
  static Char		tmp[10];		/* temp. string buffer */
  static double		wpt_lat;		/* lat of waypoint */
  static double		wpt_lon;		/* lon of waypoint */
  double		dist;			/* distance to waypoint */
  double		brg;			/* bearing */
  double		turn;			/* turn */
  double		vmg;			/* velocity made good */
  double		ttg;			/* time to go in dec. hours */
  Int32			trk;			/* track as integer */
  static Int32		otrk;			/* old track */
  Int32			x, y;			/* screen coordinates */
  Int32			r;			/* radius for track pointer */
  PointType		p1;			/* point for track pointer */
  MemHandle		strH;			/* string handle */
  MemPtr		strP;			/* string ptr */

  handled = false;

  switch (eventP->eType)
  {
    case frmOpenEvent:
      gPrefs.form = NavigationForm;

      NavigationFormInit();
      FrmDrawForm(FrmGetActiveForm());
      DrawNavigation();
      UpdateStatus(STAT_REDRAW);

      if (gPrefs.act_rte.valid || gPrefs.act_wpt.valid) {
        /* display waypoint and get position */
        display_wpt(&wpt_lat, &wpt_lon);
      }

      /* default "Route Identifier" */ 
      /* get string from resource */
      strH = DmGetResource(strRsc, NoRteStr);

      if (strH != NULL) {
	/* string handle valid, lock it */
	strP = MemHandleLock(strH);

	/* copy string */
	StrCopy(tmp, strP);

	/* unlock string handle */
	MemHandleUnlock(strH);

	/* release string resource */
	DmReleaseResource(strH);
      }
      
      if (gPrefs.act_rte.valid) {
        /* display active route */
	err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID,
	  &rec);
	
	if (!err) {
	  /* active route found */
	  routeDBEntry = DmQueryRecord(gRouteDB, rec);
          route = (route_t *)MemHandleLock(routeDBEntry);
          MemSet((Char *)tmp, 9, 0);

	  /* get string from resource */
	  strH = DmGetResource(strRsc, RteNoStr);

	  if (strH != NULL) {
	    /* string handle valid, lock it */
	    strP = MemHandleLock(strH);
	    
	    StrPrintF(tmp, "%s%-3d", strP, route->nmbr);

	    /* unlock string handle */
	    MemHandleUnlock(strH);

	    /* release string resource */
	    DmReleaseResource(strH);
	  }	  	  
 	  MemHandleUnlock(routeDBEntry);
        }
      }
      CtlSetLabel((ControlPtr)GetObjectFromActiveForm(NavigationRteIDSel), tmp);
     
      handled = true;
      break;

    case menuEvent:
      switch (eventP->data.menu.itemID) {

        case PositionMenu:
	  FrmGotoForm(GPSMainForm);
	  handled = true;
	  break;

        case SkyviewMenu:
	  FrmGotoForm(SkyviewForm);
	  handled = true;
	  break;

        case MapMenu:
	  FrmGotoForm(MapForm);
	  handled = true;
	  break;

        case MiscOptsMenu:
	  FrmGotoForm(MiscOptsForm);
	  handled = true;
	  break;

        case MapOptsMenu:
	  FrmGotoForm(MapOptsForm);
	  handled = true;
	  break;

	case PortsMenu:
	  if (gNewSerialManager) {
	    FrmGotoForm(GPSPortForm);
	  }
	  handled = true;
	  break;
	    
        case WaypointMenu:
	  FrmGotoForm(WaypointForm);
	  handled = true;
	  break;

        case RouteMenu:
	  FrmGotoForm(RouteForm);
	  handled = true;
	  break;

        case TrackMenu:
	  FrmGotoForm(TrackForm);
	  handled = true;
	  break;
	       
	case PosclipMenu:
	  handled = Position2Clipboard(gGPSData, gPrefs.units);
	  break;

	case PosgeodbMenu:
	  handled = Position2Geodb(gGPSData);
	  break;

        case AboutMenu:
	  FrmGotoForm(AboutForm);
	  handled = true;
	  break;
	      
#ifdef DEBUG_FORM
        case DebugMenu:
	  FrmGotoForm(DebugForm);
	  handled = true;
	  break;
#endif
      }
      break;
    
    case ctlSelectEvent:
    case keyDownEvent:
      
      /* Previous Waypoint */
      if ((eventP->eType == ctlSelectEvent &&
           eventP->data.ctlSelect.controlID == NavigationPrevWptButton) ||
	  (eventP->eType == keyDownEvent &&
	   eventP->data.keyDown.chr == pageUpChr)) {

	if (gPrefs.act_rte.valid) {
	  if (gPrefs.act_rte.wpt_index == 0)
	    break;

	  /* select previous waypoint */
	  gPrefs.act_rte.wpt_index--;

	  /* get active route */
	  err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID,
	    &rec);

	  if (!err) {
	    /* active route found */
	    routeDBEntry = DmQueryRecord(gRouteDB, rec);

	    /* lock record */
	    route = (route_t *)MemHandleLock(routeDBEntry);

	    /* update active waypoint record ID */
	    gPrefs.act_wpt.waypointID = 
	      route->wpt_rec_id[gPrefs.act_rte.wpt_index];

	    /* unlock route record */
	    MemHandleUnlock(routeDBEntry);

	    /* display new active waypoint */
	    display_wpt(&wpt_lat, &wpt_lon);

	  } else {
	    gPrefs.act_wpt.valid = false;
	  } /* if (!err) */
	} /* if (gPrefs.act_rte.valid) */
      }
      
      /* Next Waypoint */
      if ((eventP->eType == ctlSelectEvent &&
           eventP->data.ctlSelect.controlID == NavigationNextWptButton) ||
	 (eventP->eType == keyDownEvent &&
	  eventP->data.keyDown.chr == pageDownChr)) {     

	if (gPrefs.act_rte.valid) {

	  /* get active route */
	  err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID,
	    &rec);

	  if (!err) {
	    /* active route found */
	    routeDBEntry = DmQueryRecord(gRouteDB, rec);

	    /* lock record */
	    route = (route_t *)MemHandleLock(routeDBEntry);

	    if (gPrefs.act_rte.wpt_index == route->items-1) {
	      /* last waypoint already active */
	      MemHandleUnlock(routeDBEntry);
	      break;
	    }

	    /* select next waypoint */
	    gPrefs.act_rte.wpt_index++;

	    /* update active waypoint record ID */
	    gPrefs.act_wpt.waypointID = 
	      route->wpt_rec_id[gPrefs.act_rte.wpt_index];

	    /* unlock route record */
	    MemHandleUnlock(routeDBEntry);

	    /* display new active waypoint */
	    display_wpt(&wpt_lat, &wpt_lon);

	  } else {
	    gPrefs.act_wpt.valid = false;
	  } /* if (!err) */
	} /* if (gPrefs.act_rte.valid) */
      }
	  
      if (eventP->eType == ctlSelectEvent) {
        if (eventP->data.ctlSelect.controlID == NavigationWptIDSel) {
	  if (gPrefs.act_wpt.valid) {
            /* display active waypoint */
	    err = DmFindRecordByID(gWaypointDB, gPrefs.act_wpt.waypointID,
	      &rec);
	
	    if (!err) {
	      Char	prox_str[10];
	      
	      /* active waypoint found */
	      waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
              packed_waypoint =
		(packed_waypoint_t *)MemHandleLock(waypointDBEntry);
              UnpackWaypoint(&waypoint, packed_waypoint);
              MemSet((Char *)tmp, 7, 0);
	      StrNCopy(tmp, (Char *)&waypoint.ident, 6);
	      format_number(waypoint.dst, 0, prox_str);
	      FrmCustomAlert(NavigationWptAlert,
	        tmp, (Char *)waypoint.cmnt, prox_str);
              
	      MemHandleUnlock(waypointDBEntry);
            } /* if (!err) */
	  } /* if (gPrefs.act_wpt.valid) */
	} /* if (eventP->data.ctlSelect.controlID == NavigationWptIDSel) */

        if (eventP->data.ctlSelect.controlID ==  NavigationRteIDSel) {
	  if (gPrefs.act_rte.valid) {
            /* display active route */
	    err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID,
	      &rec);
	
	    if (!err) {
	      Char	rte_str[4];	/* route items string */
	      Char	wpt_str[16];	/* start/dst. waypoint ID string */
	      UInt16	i;		/* loop index */
	      UInt16	n;		/* last waypoint array index */
	      
	      /* clear string buffer */
	      MemSet((Char *)wpt_str, 16, 0);
	      
	      /* active route found */
	      routeDBEntry = DmQueryRecord(gRouteDB, rec);

	      /* lock record */
	      route = (route_t *)MemHandleLock(routeDBEntry);

	      /* print route number to string */
	      StrPrintF(rte_str, "%-3d", route->items);
              
	      /* get destination waypoint's index*/
	      n = route->items-1;
	      
	      /* do first index, last index */
	      for (i=0; i<=n; i+=n) {
	      
		/* get waypoint record ID from route record */
        	err = DmFindRecordByID(gWaypointDB,
	          route->wpt_rec_id[i], &rec);

		/* get waypoint record */
		waypointDBEntry = DmQueryNextInCategory(gWaypointDB, &rec, 
		  dmAllCategories);

		/* lock waypoint record */
		packed_waypoint =
		  (packed_waypoint_t *)MemHandleLock(waypointDBEntry);

		/* unpack waypoint record */
		UnpackWaypoint(&waypoint, packed_waypoint);

		/* Copy Identifier to first/second half of buffer */
		StrNCopy(&wpt_str[(i==0) ? 0 : 8], waypoint.ident, 6);

		/* unlock waypoint record */
		MemHandleUnlock(waypointDBEntry);
	      } /* for (i=0; i<=n; i+=n) */
	      FrmCustomAlert(NavigationRteAlert,
	        rte_str, wpt_str, &wpt_str[8]);
              
	      /* unlock route record */
	      MemHandleUnlock(routeDBEntry);
            } /*if (!err) */
	  } /* if (gPrefs.act_rte.valid) */
	} /* if (eventP->data.ctlSelect.controlID ==  NavigationRteIDSel) */
      } /* if (eventP->eType == ctlSelectEvent) */
      handled = true;
      break;


    case nilEvent:
      handled = true;
      //abmh_debug("NavigationFormHandleEvent.00", 0, "");
      /* throw away anything in the buffer-- we want fresh data */
      DoReceiveFlush(gPortID, 1);

      //abmh_debug("NavigationFormHandleEvent.01", 0, "");
      /*
       * Get GPS Data and update compass (COG)
       */
      updatedDisplay = ReadFromGPS();

      if (gHdFtrSet) {
      //abmh_debug("NavigationFormHandleEvent.02", 0, "");
	WinPushDrawState();
	WinSetCoordinateSystem(kCoordinatesNative);
      } 

      //abmh_debug("NavigationFormHandleEvent.03", 0, "");
      /*
       * Remove old pointer
       */
      r = (gHdFtrSet) ? WinScaleCoord(NAV_R1 - 5, false) : NAV_R1 - 5;
      p1.x = (gHdFtrSet) ? WinScaleCoord(NAV_X, false) : NAV_X;
      p1.y = (gHdFtrSet) ? WinScaleCoord(NAV_Y, false) : NAV_Y;
      x = p1.x + r * icos(otrk) / 32768;
      y = p1.y + r * isin(otrk) / 32768;

      // abmh
      WinEraseLine(p1.x, p1.y, x, y);
      WinEraseLine(p1.x+1, p1.y, x+1, y);
      WinEraseLine(p1.x, p1.y+1, x, y+1);

      if (gHdFtrSet) {
	WinPopDrawState();
      } 
      

      if (gGPSData.valid) {
	      //abmh_debug("NavigationFormHandleEvent.04", 0, "");
	/* Speed Over Ground */
	display_sog();
	      //abmh_debug("NavigationFormHandleEvent.05", 0, "");
	/* Course Made Good (= Course Over Ground = Track) */
        display_cmg();
	      //abmh_debug("NavigationFormHandleEvent.06", 0, "");
      } else {
      //abmh_debug("NavigationFormHandleEvent.07", 0, "");
        SetFieldText(NavigationSogField, NULL, false, true);
	SetFieldText(NavigationCmgField, NULL, false, true);
      //abmh_debug("NavigationFormHandleEvent.08", 0, "");
      } /* if (gGPSData.valid) */
      
      if (gGPSData.valid && gPrefs.act_rte.valid && gPrefs.auto_wpt) {
	if ( UpdateActWpt(gGPSData.lat, gGPSData.lon, false /* init */) ) {
	  /* update and display new active waypoint */
	  display_wpt(&wpt_lat, &wpt_lon);
	}
      }
      //abmh_debug("NavigationFormHandleEvent.09", 0, "");
      /* Active Waypoint approach detection and display */
      NotifyApproach();
      //abmh_debug("NavigationFormHandleEvent.10", 0, "");
      /* Track Logging */
      if (gPrefs.act_trk.log_state) {
	Boolean new_trk;

	if (TrackIntervalCheck(false /* init */, &new_trk)) {
	  TrackWriteLog(new_trk);
      //abmh_debug("NavigationFormHandleEvent.11", 0, "");
	}
      }		

      if (gGPSData.valid && gPrefs.act_wpt.valid) {
	Char	brg_str[8];

	/*
	 * Distance to Waypoint
	 */
	dist = gc_dist_sphere(gGPSData.lat, gGPSData.lon, wpt_lat, wpt_lon);
	      //abmh_debug("NavigationFormHandleEvent.12", 0, "");
	/* calculate distance in km */
	dist = dist * calcR(gGPSData.lat) / 1000.0;
	display_dst(dist);	
      //abmh_debug("NavigationFormHandleEvent.13", 0, "");
	/*
	 * Bearing (= Direction to Waypoint)
	 */
	brg = gc_course_sphere(gGPSData.lat, gGPSData.lon, wpt_lat, wpt_lon);
	brg = rad2deg(brg);
      //abmh_debug("NavigationFormHandleEvent.14", 0, "");
	/*
	 * Turn = bearing - track (pos: right/neg: left)
	 */
	turn = brg - gGPSData.cmg;
	display_trn(turn);
      //abmh_debug("NavigationFormHandleEvent.15", 0, "");
	trk = (Int32)turn - 90;

	format_number(brg, 1, brg_str);
	StrCat(brg_str, "");
	SetFieldText(NavigationBrgField, brg_str, false, true);
      //abmh_debug("NavigationFormHandleEvent.16", 0, "");
	if (gHdFtrSet) {
	  WinPushDrawState();
	  WinSetCoordinateSystem(kCoordinatesNative);
      //abmh_debug("NavigationFormHandleEvent.17", 0, "");
	} 


	/*
	 * Draw new  pointer
	 */
	r = (gHdFtrSet) ? WinScaleCoord(NAV_R1 - 5, false) : NAV_R1 - 5;
	p1.x = (gHdFtrSet) ? WinScaleCoord(NAV_X, false) : NAV_X;
	p1.y = (gHdFtrSet) ? WinScaleCoord(NAV_Y, false) : NAV_Y;
	x = p1.x + r * icos(trk) / 32768;
	y = p1.y + r * isin(trk) / 32768;

	//abmh
	WinDrawLine(p1.x, p1.y, x, y);
	WinDrawLine(p1.x, p1.y+1, x, y+1);
	WinDrawLine(p1.x+1, p1.y, x+1, y);
      //abmh_debug("NavigationFormHandleEvent.18", 0, "");
	if (gHdFtrSet) {
	  WinPopDrawState();
	} 
	
	/* save old track */
	otrk = trk;
        	
	/* velocity made good: velocity in direction to waypoint (in kts) */
	vmg = gGPSData.sog * cos( deg2rad(turn) );
	      //abmh_debug("NavigationFormHandleEvent.19", 0, "");
	/*
	 * VMG is negative if leaving waypoint. Additionally,
	 * we set a threshold (0.5 kts) to determine if we are
	 * really approaching the waypoint.
	 */
	      if ((vmg > 0.5) &&
		  (vmg * spd_c[SPD_KMH].conv) > 0.5) {
	  UInt16	h;
	  UInt16	m;
	  UInt16	s;
	  Char		ttg_str[10];	/* 999:59:59 */
	  
	  display_vmg(vmg);
	        //abmh_debug("NavigationFormHandleEvent.20", 0, "");
	  /*
	   * time to go: dist / vmg
	   *   Averaging VMG would yield more meaningful results...
	   */
		// abmhERR 3x, last executed statement
	  ttg = dist / (vmg * spd_c[SPD_KMH].conv);
	  
	  hours_to_hms(ttg, &h, &m, &s);
	        //abmh_debug("NavigationFormHandleEvent.21", 0, "");
	  if (h < 999) {
	    StrPrintF(ttg_str, "%u:%02u:%02u", h, m, s);
	    SetFieldText(NavigationTtgField, ttg_str, false, true);
      //abmh_debug("NavigationFormHandleEvent.22", 0, "");
	  } else {
	    /*
	     * More than 999 hours are considered to be an overflow,
	     * due to limited screen space. And most likely the result
	     * will be inaccurate, anyway...
	     */
	    SetFieldText(NavigationTtgField, "Overflow", false, true);
      //abmh_debug("NavigationFormHandleEvent.23", 0, "");
	  }
	} else {
	  /* leaving waypoint or not moving */
	  SetFieldText(NavigationVmgField, "--", false, true);
	  SetFieldText(NavigationTtgField, "--:--:--", false, true);
      //abmh_debug("NavigationFormHandleEvent.24", 0, "");
	} /* if (vmg > 1) */
      } else {
        /* no active WPT or GPS invalid */
	SetFieldText(NavigationDstField, NULL, false, true);
        SetFieldText(NavigationBrgField, NULL, false, true);
        SetFieldText(NavigationTrnField, NULL, false, true);
        SetFieldText(NavigationVmgField, NULL, false, true);
	SetFieldText(NavigationTtgField, NULL, false, true);
      //abmh_debug("NavigationFormHandleEvent.25", 0, "");
      } /* if (gGPSData.valid && gPrefs.act_wpt.valid) */
      break;


    case frmCloseEvent:
      /* Deallocate Fields' Text Memory */
      SetFieldText(NavigationStatusField, NULL, false, true);
      SetFieldText(TimeField, NULL, false, true);
      SetFieldText(NavigationDstField, NULL, false, true);
      SetFieldText(NavigationSogField, NULL, false, true);
      SetFieldText(NavigationVmgField, NULL, false, true);
      SetFieldText(NavigationTtgField, NULL, false, true);
      SetFieldText(NavigationCmgField, NULL, false, true);
      SetFieldText(NavigationBrgField, NULL, false, true);
      SetFieldText(NavigationTrnField, NULL, false, true);
      WinDeleteWindow(gNavigationH, true /* eraseIt */);
      break;

  }
  return(handled);
} /* NavigationFormHandleEvent() */
