/*****************************************************************************
 *
 * $RCSfile: RouteEditForm.c,v $
 *
 * GPS4Palm Route Edit Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004-12-10 19:58:25 $
 *
 * $Revision: 1.9 $
 *
 * $Log: RouteEditForm.c,v $
 * Revision 1.9  2004-12-10 19:58:25  mp
 * implemented clipboard, SysKeyboardDialog, and SysGraffitiReferenceDialog
 *
 * Revision 1.8  2004/11/27 10:16:57  mp
 * replaced ErrFatalDisplay() by Die() for memory allocation failure
 *
 * Revision 1.7  2004/11/24 21:21:42  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.6  2004/11/23 17:48:08  mp
 * removed unused variables
 *
 * Revision 1.5  2004/11/16 20:47:30  mp
 * added check after creating new record, modified PackWaypoint() call
 *
 * Revision 1.4  2004/11/16 18:46:28  mp
 * fixed route editing, changed waypoint add function from append to insert
 *
 * Revision 1.3  2004/11/15 19:56:51  mp
 * create_route()/store_route() rewritten,
 * buttons "ok" and "cancel" replaced by "done" (better PalmOS UI compliancy),
 * implemented route waypoint editing: add, delete, move up, move down
 *
 * Revision 1.2  2004/11/14 22:45:45  mp
 * added RouteEditAllTable and RouteEditSelTable functions
 *
 * Revision 1.1  2004/11/13 15:55:17  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "ResourceDefines.h"
#include <Unix/sys_types.h>
#include "stringil.h"
#include "Data.h"
#include "RouteEditForm.h"
#include "GPS.h"		/*  ReadFromGPS() */
#include "Serial.h"
#include "Utils.h"
#include "common.h"


/* Global Variables */

extern PrefsType	gPrefs;		/* Preferences data structure */
extern DmOpenRef	gWaypointDB;	/* Waypoint Data Base reference */
extern DmOpenRef	gRouteDB;	/* Route Data Base reference */
extern GPSType		gGPSData;	/* GPS Data */
extern ActRouteType	gSelRoute;	/* Selected route */


/* Static Functions */
static void RouteEditFormInit(void)			ROUTEEDIT_SECTION;
static Err create_route(route_t *route)			ROUTEEDIT_SECTION;
static void store_route(route_t *route)			ROUTEEDIT_SECTION;


/*****************************************************************************
 * FUNCTION:	RouteEditFormInit
 *
 * DESCRIPTION: Route Edit Form Init Function (not used)
 *
 ****************************************************************************/
static 
void RouteEditFormInit(void)
{  

}


/*****************************************************************************
 * FUNCTION:	TableDrawWptCell
 *
 * DESCRIPTION:	Custom table cell drawing function for Waypoint ID
 *              (max. 6 characters). The DB record number is
 *		retrieved via TblGetRowData().
 *
 * GLOBALS:	gWayptDB	-- waypoint DB reference
 *		gRouteDB	-- route DB reference
 *
 * PARAMETERS:	Note: The function's prototype is defined by PalmOS
 *		      (TableDrawItemFuncType).
 *
 * RETURNED:	-
 ****************************************************************************/
static void TableDrawWaypointIdCell(void *tblP, Int16 row, Int16 column,
                   RectangleType *bounds)
{
  FormPtr		frmP;			/* form ptr */
  UInt16		tbl_id;			/* table object ID */
  Char			str[7];			/* string buffer */
  route_t		*route;			/* route */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
  MemHandle		routeDBEntry;		/* record handle */
  MemHandle		waypointDBEntry;	/* record handle */
  UInt16		rec;			/* record index */
  UInt16		index;			/* waypoint index */
  Err			err;			/* error code */
  
  frmP = FrmGetActiveForm();
  FntSetFont(stdFont);
  MemSet((char *)str, 7, 0);
  
  /* get object ID of current table */
  tbl_id = FrmGetObjectId(frmP, MyFrmGetObjectIndexFromPtr(frmP, tblP));
    
  /* erase table cell */
  WinEraseRectangle(bounds, 0);
  
  /* check if row is usable */
  if (TblRowUsable(tblP, row)) {
  
    if (tbl_id == RouteEditAllTable) {
      /* All available waypoints */

      /* get record index from row data */
      rec = (UInt16)TblGetRowData(tblP, row);
      
      waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
      packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
      UnpackWaypoint(&waypoint, packed_waypoint);
      StrNCopy(str, (Char *)&waypoint.ident, 6);
      MemHandleUnlock(waypointDBEntry);
 
    } else {
      /* Waypoints of selected route */
   
      /* get waypoint index from row data */
      index = (UInt16)TblGetRowData(tblP, row);

      err = DmFindRecordByID(gRouteDB,
	gSelRoute.routeID, &rec);
      
      routeDBEntry = DmQueryRecord(gRouteDB, rec);
            
      /* lock route record */
      route = (route_t *)MemHandleLock(routeDBEntry);

      /* get waypoint record ID from route record */
      err = DmFindRecordByID(gWaypointDB,
	route->wpt_rec_id[index], &rec);

      /* unlock route record */
      MemHandleUnlock(routeDBEntry);

      if (!err) {

	/* get waypoint record */
	waypointDBEntry = DmQueryNextInCategory(gWaypointDB, &rec, 
	  dmAllCategories);

	/* lock waypoint record */
	packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);

	/* unpack waypoint record */
	UnpackWaypoint(&waypoint, packed_waypoint);

	/* Copy Identifier */
	StrNCopy(str, waypoint.ident, 6);

	/* unlock waypoint record */
	MemHandleUnlock(waypointDBEntry);
      } else {
        /* debugging only, we should never go here */
        StrCopy(str, "???");
      }
    }
    
    /* Write to table cell */
    WinDrawChars(str, StrLen(str), bounds->topLeft.x + 1, bounds->topLeft.y);
  } /* if (TblRowUsable(tblP, row)) */
}


/****************************************************************************/
/**
 * \brief	All waypoints/selected waypoints table drawing function
 *
 * \param	tbl_id		table id
 * \param	top		top table row
 * \param	sel		ptr to selected table row, may be NULL if
 *				  selection/scrollbar is handled elsewhere
 *
 *****************************************************************************/
static
void routeEditDrawTable(UInt16 tbl_id, Int32 top, UInt16 *sel)
{
  TablePtr 		tblP;			/* table ptr */
  FormPtr 		frmP;			/* form ptr */
  ScrollBarPtr		sclP;			/* scrollbar ptr */
  UInt16		row;			/* table row counter */
  UInt16		rec;			/* record nr */
  UInt16		max_rows = 0;		/* total nr. of routes */
  UInt16		numRows;		/* number of visible rows */
  MemHandle		routeDBEntry;		/* record handle */
  route_t		*route;			/* route */
  MemHandle		waypointDBEntry;	/* record handle */
  Err			err;			/* error code */
  
  frmP = FrmGetActiveForm();
  tblP = GetObjectFromForm(frmP, tbl_id);
  
  if (tbl_id == RouteEditAllTable) {
    sclP = GetObjectFromForm(frmP, RouteEditAllScrl);
  
    /* total number of waypoints */
    max_rows = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);
  
  } else {
    sclP = GetObjectFromForm(frmP, RouteEditSelScrl);
  
    /* get number of waypoints on this route */
    err = DmFindRecordByID(gRouteDB, gSelRoute.routeID,
      &rec);

    if (!err) {
      /* selected route found */
      routeDBEntry = DmQueryRecord(gRouteDB, rec);    
      
      /* lock record */
      route = (route_t *)MemHandleLock(routeDBEntry);

      max_rows = route->items;
      
      /* unlock record */
      MemHandleUnlock(routeDBEntry);
      
    } /* if (!err) */
  } /* if (tbl_id == RouteEditAllTable) */
  
  /* number of visible rows */
  numRows = TblGetNumberOfRows(tblP);

  /* set columns usable */
  TblSetColumnUsable(tblP, 0, true);
      
  
  if (tbl_id == RouteEditAllTable) {
    /* set row data to Waypoint DB entries (record index) */
    for (row = 0; row < numRows; row++) {

      /* default: mark row as not usable (empty or no valid record) */
      TblSetRowUsable(tblP, row, false);

      if (row + top < max_rows) {
	/* seek from beginning of database */
	rec = 0;

	/* get next record */
	err = DmSeekRecordInCategory(gWaypointDB, &rec /* index */,
          top + row /* offset */, dmSeekForward, dmAllCategories);

	if (err == errNone) {
          /* query record */
          waypointDBEntry = DmQueryRecord(gWaypointDB, rec);

	  if (waypointDBEntry) {
	    /*
	     * Record is valid, set row data to index and
	     * mark table row as usable.
	     */
	    TblSetRowData(tblP, row, rec);
            TblSetRowUsable(tblP, row, true);
	  }
	}
      } /*  if (row + top < max_rows) */
    } /* for (row = 0; row < numRows; row++) */  
  
  } else {
    /* (tbl_id == RouteEditSelTable) */
    
    /* set row data to Waypoint DB entries (record index) */
    for (row = 0; row < numRows; row++) {
      /* default: mark row as not usable (empty or no valid record) */
      TblSetRowUsable(tblP, row, false);

      if (row + top < max_rows) {
        TblSetRowData(tblP, row, row + top);
        TblSetRowUsable(tblP, row, true);
      }
    }
  }
  
  /* initialize table */
  for (row = 0; row < numRows; row++) {
    
    if (row + top < max_rows) {
      /* set all rows selectable */
      TblSetRowSelectable(tblP, row, true);
      
      /* set columns to custom style, set drawing functions */
      TblSetItemStyle(tblP, row, 0, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 0, TableDrawWaypointIdCell);
       
      /* mark all rows as invalid to trigger table update by FrmDrawForm() */
      TblMarkRowInvalid(tblP, row);
    }
  }

  /* set scrollbar: value, min, max, pagesize */
  if (max_rows <= numRows) {
    SclSetScrollBar(sclP, 0, 0, 0, numRows);
  } else {
    SclSetScrollBar(sclP, top, 0, max_rows - numRows, numRows);
  }
  
  /*
   * Selected route scolled below visible table rows,
   *   selection is switched to last visible entry.
   */
  if ( *sel > (top + numRows - 1) ) {
    *sel = top + numRows - 1;
  }

  /*
   * Selected route scolled above visible table rows,
   *   selection is switched to first visible entry.
   */  
  if (*sel < top) {
    *sel = top;
  }

  for (row = 0; row < numRows; row++, top++) {
    if ((top < max_rows) && (top == *sel)) {
      TblSelectItem(tblP, row, 0);
      break;
    }
  }
  
  FrmDrawForm(frmP);
}


/*****************************************************************************
 * FUNCTION:	create_route
 *
 * DESCRIPTION: Add new route to database
 *
 * GLOBALS:	gRouteDB	-- Route Database Reference
 *		gSelRoute	-- return new routeID
 *
 * PARAMETERS:	route		-- ptr to route record
 *		
 *
 * RETURNED:    false if no unused route number exists
 ****************************************************************************/
static
Err create_route(route_t *route)
{
  UInt16	nmbr;			/* route number */
  UInt16	index;			/* record index */
  MemHandle	rteH;			/* route record handle */
  MemHandle	routeDBEntry;		/* route record handle */
  route_t	*rteP;			/* route record ptr */
  Boolean	rte_exists = false;	/* flag: route exists */

  /* init route comment */
  MemSet((Char *)route->cmnt, 21, 0);
    
  /* clear number of items (waypoints) */
  route->items = 0;
  
  for (nmbr=0; nmbr <= 255; nmbr++) {
    route->nmbr = (G_byte)nmbr;
    
    /* Find sort position of new record in route DB */
    index = DmFindSortPosition(gRouteDB, route, 0 /* newRecordInfo */,
      (DmComparF *)CompareRteNmbr /* DmComparF *compar */, 0 /* other */);

    if (index > 0) {
      /* check if route with same number already exists */
      
      rteH = DmQueryRecord(gRouteDB, index - 1);
      rteP = MemHandleLock(rteH);
      if (route->nmbr == rteP->nmbr) {
        MemHandleUnlock(rteH);
        rte_exists = true;
      } else {
        MemHandleUnlock(rteH);
        rte_exists = false;
	break;
      }
    } else {
      rte_exists = false;
      break;
    } /* if (index > 0) */
  } /* for (nmbr=0; nmbr <= 255; nmbr++) */
	   
  if (rte_exists) {
    /* failure */
    gSelRoute.valid = false;
    return 1;
  } else {    
    /* success */

    /* create comment */
    StrPrintF(route->cmnt, "<Route #%d>", route->nmbr);	 
    
    /* create new route record */
    routeDBEntry = DmNewRecord(gRouteDB, &index, 
      1 /* size, will be set by PackRoute() */);

    if (routeDBEntry) {
      DmReleaseRecord(gRouteDB, index, false /* dirty */);
    } else {
      Die("Cannot create route record!");
    }

    /* Store Route Header in Database */
    PackRouteHdr(route, index);    
    
    /* get unique route ID */
    DmRecordInfo(gRouteDB, index, NULL /* *attrP */,
      &gSelRoute.routeID /* *uniqueIDP */, NULL /* *chunkIDP */);
    
    gSelRoute.valid = true;
    return 0;
  }    
}


/*****************************************************************************
 * FUNCTION:	store_route
 *
 * DESCRIPTION: Add new route to database
 *
 * GLOBALS:	gRouteDB	-- Route Database Reference
 *		gSelRoute	-- Selected Route (unique ID)
 *
 * PARAMETERS:	route		-- Route Ptr
 *		
 *
 * RETURNED:    %
 ****************************************************************************/
static
void store_route(route_t *route)
{
  FormPtr		frmP;
  Char			*strP;
  UInt16		len;
  UInt16		rec;
  Err			err;

  frmP = FrmGetActiveForm();
  
  /* Copy Route Comment from Field */
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
    FrmGetObjectIndex(frmP, RouteCmntField)));
  len = StrLen(strP);
  MemSet((Char *)route->cmnt, 21, 0);
  StrNCopy(route->cmnt, strP, (len > 20) ? 20 : len);
  
  err = DmFindRecordByID(gRouteDB, gSelRoute.routeID,
    &rec);

  if (!err) {
    /* selected route found */

    /* Store Route Header in Database */
    PackRouteHdr(route, rec);
  }  
}


/*****************************************************************************
 * FUNCTION:	RouteEditFormHandleEvent
 *
 * DESCRIPTION: Route Edit Form Event Handler
 *
 * GLOBALS:	gPrefs		-- Application Preferences
 *		gRouteDB	-- Route Data Base reference
 *		gSelRoute	-- Selected or new route
 *
 * PARAMETERS:	eventP		-- pointer to event structure
 *
 * RETURNED:    Status flag: event handled
 ****************************************************************************/
Boolean RouteEditFormHandleEvent(EventPtr eventP)
{
  Boolean		handled;		/* flag: event handled */
  FormPtr		frmP;			/* form ptr */
  TablePtr		tblP;			/* table ptr */
  ScrollBarPtr		sclP;			/* scrollbar ptr */
  FieldPtr		fldP;			/* field ptr */
  FieldAttrType		fld_attr;		/* field attributes */
  route_t		*routeP;		/* route ptr */
  MemHandle		routeDBEntry;		/* route record handle */
  UInt16		rec;			/* record index */
  Char			tmp[7];			/* temp. string buffer */
  UInt16		nr_waypoints;		/* total number of waypoints */
  UInt16		index;			/* waypoint index */
  UInt32		wpt_rec_id;		/* waypoint record ID */
  Err			err;			/* error code */
  static route_t	route;			/* route */
  static Int32		waypt_top = 0;		/* all waypoints table top */
  static Int32		route_top = 0;		/* route waypoints table top */
  static UInt16		waypt_sel = 0;		/* all waypoints selected row */
  static UInt16		route_sel = 0;		/* route waypoints sel. row */

  frmP = FrmGetActiveForm();
  handled = false;

  /* get no. of all waypoints */
  nr_waypoints = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);

  
  switch (eventP->eType) {
    case frmOpenEvent:
      RouteEditFormInit();
      FrmDrawForm(frmP);

      /* initialize table view and selection */
      waypt_top = 0;
      route_top = 0;
      waypt_sel = 0;
      route_sel = 0;

      if (!gSelRoute.valid) {
        /* RouteEditForm opened with button "new" */
	
	/* create new route */
	err = create_route(&route);
	
	if (err) {
	  /* no route number available */
	  FrmAlert(RouteCreateAlert);
	  FrmGotoForm(RouteForm);
	  handled = true;
	  break;
	}
	
      }

      /* write route data to fields */
      
      /* find route record */
      err = DmFindRecordByID(gRouteDB, gSelRoute.routeID,
	&rec);
	
      if (!err) {	
	/* selected or new route found */
	routeDBEntry = DmQueryRecord(gRouteDB, rec);
	
	/* lock record */
        routeP = (route_t *)MemHandleLock(routeDBEntry);
          
	/* copy route number from record */
	MemSet((Char *)tmp, 7, 0);
	StrPrintF(tmp, "%d", routeP->nmbr);
	SetFieldText(RouteNmbrField, tmp, false, true);
	route.nmbr = routeP->nmbr;
	  
	/* copy comment from record */
	SetFieldText(RouteCmntField, routeP->cmnt, false, true);
	StrNCopy(route.cmnt, routeP->cmnt, 20);
	  
	/* copy number of items from record */
	route.items = routeP->items;
	  
	/* unlock record */
	MemHandleUnlock(routeDBEntry);
      } /* if (!err) */
      
      /* draw tables */
      routeEditDrawTable(RouteEditAllTable, waypt_top, &waypt_sel);
      routeEditDrawTable(RouteEditSelTable, route_top, &route_sel);
      
      handled = true;
      break;

    case menuEvent:    
      /* check if RouteCmntField has focus */
      fldP = (FieldPtr)GetObjectFromActiveForm(RouteCmntField);
      FldGetAttributes(fldP, &fld_attr);
      
      switch (eventP->data.menu.itemID) {

	case UndoMenu:
	  if (fld_attr.hasFocus == 1)
	    FldUndo(fldP);
	  handled = true;
	  break;
	  
	case CutMenu:
	  if (fld_attr.hasFocus == 1)
	    FldCut(fldP);
	  handled = true;
	  break;
	  
	case CopyMenu:
	  if (fld_attr.hasFocus == 1)
	    FldCopy(fldP);
	  handled = true;
	  break;
	  
	case PasteMenu:
	  if (fld_attr.hasFocus == 1)
	    FldPaste(fldP);
	  handled = true;
	  break;
	  
	case KeyboardMenu:
	  if (fld_attr.hasFocus == 1)
	    SysKeyboardDialog(kbdAlpha);
	  handled = true;
	  break;
	  
	case GraffitiMenu:
	  SysGraffitiReferenceDialog(referenceDefault);
          handled = true;
	  break;
	  
      } /* switch (eventP->data.menu.itemID) */
      break;
  
  return (fld_attr.hasFocus == 1);  
    case tblSelectEvent:
      /* update table row selection */
      {
        UInt16 id = eventP->data.tblSelect.tableID;
	
	if (id == RouteEditAllTable) {
	  waypt_sel = eventP->data.tblSelect.row + waypt_top;
	} else {
	  route_sel = eventP->data.tblSelect.row + route_top;
        }
      }
      handled = true;
      break;

      
    case sclRepeatEvent:
      /* handle scroll enevt */
      {
        UInt16 id = eventP->data.sclRepeat.scrollBarID;
	
	if (id == RouteEditAllScrl) {
          Doscroll(RouteEditAllTable, id, nr_waypoints, &waypt_top);
          routeEditDrawTable(RouteEditAllTable, waypt_top, &waypt_sel);
        } else {
	  DmFindRecordByID(gRouteDB,
	    gSelRoute.routeID, &rec);
      
          routeDBEntry = DmQueryRecord(gRouteDB, rec);
            
	  /* lock route record */
	  routeP = (route_t *)MemHandleLock(routeDBEntry);

	  Doscroll(RouteEditSelTable, id, routeP->items, &route_top);
	  
	  /* unlock route record */
	  MemHandleUnlock(routeDBEntry);
	
	  routeEditDrawTable(RouteEditSelTable, route_top, &route_sel);
	}
      }
      
      /* Note: scrollbar needs to handle the event, too */
      handled = false;
      break;
      

    case keyDownEvent:
      /* "virtual" up/down keys trigger paging of "All Waypoints" table */
      switch (eventP->data.keyDown.chr) {
	case pageUpChr:
	case pageDownChr:
          Dopage(RouteEditAllTable,
	    (eventP->data.keyDown.chr == pageDownChr),
	    nr_waypoints, &waypt_top);
          routeEditDrawTable(RouteEditAllTable, waypt_top, &waypt_sel);
      
          /* Note: scrollbar needs to handle the event, too */
          handled = false;
          break;
      }
      break;

    
    case ctlSelectEvent:
      switch (eventP->data.ctlSelect.controlID) {
        case RouteEditAddButton:
	  /* add waypoint to route (after selection) */
	  
	  /* get ptr to waypoint table */
	  tblP = (TablePtr)GetObjectFromActiveForm(RouteEditAllTable);
	  
	  /* get waypoint record index from row data */
          rec = (UInt16)TblGetRowData(tblP, waypt_sel-waypt_top);
	  
	  /* get unique waypoint ID */
          DmRecordInfo(gWaypointDB, rec, NULL /* *attrP */,
            &wpt_rec_id /* *uniqueIDP */, NULL /* *chunkIDP */);
          
	  /* get route record index from unique ID */ 
	  DmFindRecordByID(gRouteDB, gSelRoute.routeID, &rec);
          
	  /* get ptr to route table */
	  tblP = (TablePtr)GetObjectFromActiveForm(RouteEditSelTable);
	  
	  /* get waypoint ID index from row data */
          index = (UInt16)TblGetRowData(tblP, route_sel-route_top);
	  	  	  
	  /* add selected waypoint to current route */
	  InsertRouteWpt(rec, wpt_rec_id, index);
	  
	  /* increment waypoint count */
	  route.items++;

	  /* 
	   * Redraw route table, select new Waypoint 
	   * and make sure it is visible.
	   */
          /* select new waypoint */
	  if (route.items > 1)
	    route_sel++;
	  
          /* set top table row to ensure selected item is visible */
	  if (route_sel-route_top > TblGetNumberOfRows(tblP)-1) {
            route_top++;	    
	  }
	  
	  /* update scroll bar */
	  sclP = (ScrollBarPtr)GetObjectFromActiveForm(RouteEditSelScrl);
	  SclSetScrollBar(sclP, route_sel /* value */,
	    0 /* min */, route.items-1 /* max */,
	    TblGetNumberOfRows(tblP) /* page size */);  
	  
	  /* redraw table */
	  routeEditDrawTable(RouteEditSelTable, route_top, &route_sel);
	  
	  handled = true;
	  break;

	  
	case RouteEditUpButton:
	case RouteEditDnButton:
	  /* move waypoint up/down in route */
	  if (route.items == 0)
	    break;
	  	  
	  /* get ptr to route table */
	  tblP = (TablePtr)GetObjectFromActiveForm(RouteEditSelTable);
	  
	  /* get waypoint ID index from row data */
          index = (UInt16)TblGetRowData(tblP, route_sel-route_top);
	  
          /* get route record index from unique ID */ 
	  DmFindRecordByID(gRouteDB, gSelRoute.routeID, &rec);
	  
	  if (eventP->data.ctlSelect.controlID == RouteEditUpButton) {
	    /* move up */
	    if (route_sel == 0)
	      break;
	    MoveRouteWpt(rec, index, true /* up */);
	    route_sel--;
	    
	  } else {
	    /* move down */
	    if (route_sel == route.items-1)
	      break;
	    MoveRouteWpt(rec, index, false /* up */);
	    route_sel++;
	  }

          /* set top table row to ensure selected item is visible */
	  if (route_sel-route_top > TblGetNumberOfRows(tblP)-1) {
            route_top++;	    
	  }
	  
	  if (route_sel-route_top < 0) {
	    route_top--;
	  }
	  
	  /* update scroll bar */
	  sclP = (ScrollBarPtr)GetObjectFromActiveForm(RouteEditSelScrl);
	  SclSetScrollBar(sclP, route_sel /* value */,
	    0 /* min */, route.items-1 /* max */,
	    TblGetNumberOfRows(tblP) /* page size */);  
	  
	  /* redraw table */
	  routeEditDrawTable(RouteEditSelTable, route_top, &route_sel);
	  
	  handled = true;
	  break;
	
	
	case RouteEditDelButton:
	  /* delete waypoint from route */
	  if (route.items == 0)
	    break;
	  
	  /* get ptr to route table */
	  tblP = (TablePtr)GetObjectFromActiveForm(RouteEditSelTable);
	  
	  /* get waypoint ID index from row data */
          index = (UInt16)TblGetRowData(tblP, route_sel-route_top);

          /* get route record index from unique ID */ 
	  DmFindRecordByID(gRouteDB, gSelRoute.routeID, &rec);
	  
          /* delete selected waypoint */
	  DelRouteWpt(rec, index);
	  
	  route.items--;
	  
	  /* 
	   * Redraw route table, select previous Waypoint
	   * of the route and make sure it is visible.
	   */
	  if (route_sel > 0) {
	    route_sel--;
	  }
	  
	  Doscroll(RouteEditSelTable, RouteEditSelScrl, route.items, &route_top);
	  
	  /* redraw table */
	  routeEditDrawTable(RouteEditSelTable, route_top, &route_sel);

	  /* update selection */
	  TblSelectItem(tblP, route_sel-route_top, 0);
	  handled = true;
	  break;	  
	
	
	case RouteEditDoneButton:
	  /* close form */
	  FrmGotoForm(RouteForm);
	  handled = true;
	  break;
	  
      }
      break;


    case frmCloseEvent:
      /* store route header (Comment Field!) before closing form */
      if (gSelRoute.valid)
        store_route(&route);
    
      /* Deallocate Fields' Text Memory */
      SetFieldText(RouteNmbrField, NULL, false, true);
      SetFieldText(RouteCmntField, NULL, false, true);
      break;
  }
  
  return(handled);
} /* RouteEditFormHandleEvent() */
