/*****************************************************************************
 *
 * $RCSfile: RouteForm.c,v $
 *
 * GPS4Palm Routes Form
 *
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-06 13:40:07 $
 *
 * $Revision: 1.19 $
 *
 * $Log: RouteForm.c,v $
 * Revision 1.19  2005-05-06 13:40:07  mp
 * added workaround for Bluetooth-Open Bug
 *
 * Revision 1.18  2005/01/30 21:22:13  mp
 * added PortsMenu event handling
 *
 * Revision 1.17  2005/01/23 10:01:32  mp
 * changed re-opening of GPS serial port after data transfer
 *
 * Revision 1.16  2004/12/18 16:11:53  mp
 * fixed serial port re-opening for PDAs with New Serial Manager
 *
 * Revision 1.15  2004/12/07 17:13:52  mp
 * Modified selection of Active Route: route must contain at least two
 * waypoints.
 *
 * Revision 1.14  2004/12/06 21:10:00  mp
 * changed display of Active Route to bold font
 *
 * Revision 1.13  2004/11/30 20:43:08  mp
 * added TransferTrkReqEvent and TrackMenu event handling, added GarminGetTrks()
 *
 * Revision 1.12  2004/11/26 20:07:02  mp
 * added string truncation to fit table cell in TableDrawCommentCell()
 *
 * Revision 1.11  2004/11/24 21:22:35  mp
 * moved static function declarations from header to implementation file,
 * added AboutForm menu event handling
 *
 * Revision 1.10  2004/11/23 17:47:52  mp
 * removed unused variables
 *
 * Revision 1.9  2004/11/20 15:45:11  mp
 * added UpdateActWpt() function call
 *
 * Revision 1.8  2004/11/19 18:36:16  mp
 * added setting of active waypoint from start of active route
 *
 * Revision 1.7  2004/11/15 19:52:38  mp
 * code cleanup
 *
 * Revision 1.6  2004/11/13 18:53:47  mp
 * fixed table selection, fixed status update after route download
 *
 * Revision 1.5  2004/11/13 15:59:24  mp
 * added route new/edit functions, fixed bugs
 *
 * Revision 1.4  2004/11/11 21:43:09  mp
 * Modified table handling (custom drawing function for comments),
 * added Active Route selection, added route deletion.
 *
 * Revision 1.3  2004/11/11 18:41:03  mp
 * modified table handling, still problems with memory manager
 *
 * Revision 1.2  2004/06/23 18:36:09  mp
 * added debug output
 *
 * Revision 1.1  2004/04/29 19:25:58  mp
 * initial version
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <SerialMgrOld.h>
#include "ResourceDefines.h"
#include "gpslib.h"
#include "RouteForm.h"
#include "Data.h"
#include "DataExchange.h"
#include "Utils.h"
#include "Serial.h"
#include "utm.h"
#include "geo.h"
#include "fp.h"
#include "common.h"

/* Global Variables */
extern PrefsType	gPrefs;			/* application preferences */
extern UInt32		gSerialPort;		/* serial port name/no. */

ActRouteType		gSelRoute;		/**< selected route */

/* Local Variables */
static Int32		gRouteTopRow;		/* top table row */
static UInt16		gRouteSel;		/* selected route (table) */
static UInt16		nr_waypoints;		/* number of waypoints */
static UInt16		nr_routes;		/* number of routes */


/* Static Functions */
static void RouteFormInit(FormPtr frmP)			ROUTE_SECTION;
static void RouteFormDeinit(FormPtr frmP)		ROUTE_SECTION;
static void ChangePorts(void)				ROUTE_SECTION;


/*****************************************************************************
 * FUNCTION:	RouteFormInit
 *
 * DESCRIPTION: Route Form Init Function
 *
 ****************************************************************************/
static void RouteFormInit(FormPtr frmP)
{
#pragma unused(frmP)
  // warning-- don't do any drawing in this routine.
  // Also, don't call FrmSetFocus from here (it must be called *after*
  // FrmDrawForm)
  
}


/*****************************************************************************
 * FUNCTION:	RouteFormDeinit
 *
 * DESCRIPTION: Route Form De-Init Function
 *
 ****************************************************************************/
static void RouteFormDeinit(FormPtr frmP)
{
#pragma unused(frmP)

}


/*****************************************************************************
 * FUNCTION:	TableDrawCommentCell
 *
 * DESCRIPTION:	Custom table cell drawing function for Route Number
 *              (max. 3 digits) and Route Comment (string, max. 20
 *		characters). The DB record number is retrieved via
 *              TblGetRowData().
 *
 * GLOBALS:	gRouteDB	-- route DB reference
 *
 * PARAMETERS:	Note: The function's prototype is defined by PalmOS
 *		      (TableDrawItemFuncType).
 *
 * RETURNED:	-
 ****************************************************************************/
static void TableDrawCommentCell(void *tblP, Int16 row, Int16 column,
                   RectangleType *bounds)
{
  Char			str[21];		/* string buffer */
  MemHandle		routeDBEntry;		/* record handle */
  UInt16		rec;			/* record index */
  route_t		*route;			/* route */
  UInt32		routeID;		/* route ID (unique) */
  Int16			cell_width;		/* table cell width in pixels */
  Boolean		fits;			/* string fits in table cell */
  Int16			length;			/* string length */
  
  /* get record index from row data */
  rec = (UInt16)TblGetRowData(tblP, row);
  
  if (column == 0) {
    WinEraseRectangle(bounds, 0);
    
    /*
     * Set font to bold, because route items column (which is numeric, too)
     * uses it due to ItemStyle numericTableItem.
     */
    FntSetFont(boldFont);

    if (TblRowUsable(tblP, row)) {
      routeDBEntry = DmQueryRecord(gRouteDB, rec);
      route = (route_t *)MemHandleLock(routeDBEntry);
      MemSet((Char *)str, 21, 0);
      StrPrintF(str, "%3d", route->nmbr);
      MemHandleUnlock(routeDBEntry);

      WinDrawChars(str, StrLen(str), bounds->topLeft.x + 1, bounds->topLeft.y);
    }
    
  } else if (column == 1) {
    WinEraseRectangle(bounds, 0);
    
    if (TblRowUsable(tblP, row)) {
      routeDBEntry = DmQueryRecord(gRouteDB, rec);

      /* get unique Route ID */
      DmRecordInfo(gRouteDB, rec, NULL /* *attrP */,
	&routeID /* *uniqueIDP */, NULL /* *chunkIDP */);

      if (gPrefs.act_rte.valid && (routeID == gPrefs.act_rte.routeID)) {
        FntSetFont(boldFont);
      } else {
        FntSetFont(stdFont);
      }

      route = (route_t *)MemHandleLock(routeDBEntry);
      MemSet((Char *)str, 21, 0);
      StrNCopy(str, (Char *)route->cmnt, 20);
      MemHandleUnlock(routeDBEntry);
      
      /* truncate string to fit in table cell */
      length = StrLen(str);
      cell_width = bounds->extent.x;
      FntCharsInWidth(str, &cell_width, &length, &fits);
     
      WinDrawChars(str, length, bounds->topLeft.x + 1, bounds->topLeft.y);
    }
  }
}


/****************************************************************************/
/**
 * \brief	Routes table drawing function
 *
 *****************************************************************************/
void routeDrawTable(void)
{
  TablePtr 		tblP;			/* table ptr */
  FormPtr 		frmP;			/* form ptr */
  ScrollBarPtr		sclP;			/* scrollbar ptr */
  UInt16		top;			/* table row counter */
  UInt16		row;			/* table row counter */
  UInt16		rec;			/* record nr */
  UInt16		max_rows;		/* total nr. of routes */
  UInt16		numRows;		/* number of visible rows */
  MemHandle		routeDBEntry;		/* record handle */
  route_t		*route;			/* route */
  Err			err;			/* error code */
  
  frmP = FrmGetActiveForm();
  tblP = (TablePtr)GetObjectFromForm(frmP, RouteTable);
  top = gRouteTopRow;
  
  /* total number of routes */
  max_rows = DmNumRecordsInCategory(gRouteDB, dmAllCategories);
  
  /* number of visible rows */
  numRows = TblGetNumberOfRows(tblP);

  /* set columns usable */
  TblSetColumnUsable(tblP, 0, true);
  TblSetColumnUsable(tblP, 1, true);
  TblSetColumnUsable(tblP, 2, true);
      
  /* set row data to Route DB entries (record index) */
  for (row = 0; row < numRows; row++) {

    /* default: mark row as not usable (empty or no valid record) */
    TblSetRowUsable(tblP, row, true);

    if (row + top < max_rows) {
      /* seek from beginning of database */
      rec = 0;
      
      /* get next record */
      err = DmSeekRecordInCategory(gRouteDB, &rec /* index */,
        top + row /* offset */, dmSeekForward, dmAllCategories);
      
      if (err == errNone) {
        /* query record */
        routeDBEntry = DmQueryRecord(gRouteDB, rec);

	if (routeDBEntry) {
	  /*
	   * Record is valid, set row data to index and
	   * mark table row as usable.
	   */
	  TblSetRowData(tblP, row, rec);
          TblSetRowUsable(tblP, row, true);
	}
      }
    } /*  if (row + top < max_rows) */
  } /* for (row = 0; row < numRows; row++) */  
  
  /* initialize table */
  for (row = 0; row < numRows; row++) {
    
    if (row + top < max_rows) {
      /* set all rows selectable */
      TblSetRowSelectable(tblP, row, true);

      /* get record index from row data */
      rec = TblGetRowData(tblP, row);

      /* get record */
      routeDBEntry = DmQueryRecord(gRouteDB, rec);
      
      /* lock record */
      route = (route_t *)MemHandleLock(routeDBEntry);
      
      /* set columns to custom style, set drawing functions */
      /*
       * route number
       *
       * Note:	ItemStyle customTableItem is used instead of
       *	numericTableItem, because otherwise
       *	TblSelectItem() does not work properly.
       */
      TblSetItemStyle(tblP, row, 0, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 0, TableDrawCommentCell);
      
      /* route comment */
      TblSetItemStyle(tblP, row, 1, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 1, TableDrawCommentCell);

      /* number of waypoints */
      TblSetItemStyle(tblP, row, 2, numericTableItem);    
      TblSetItemInt(tblP, row, 2, route->items);
      
      /* unlock record */
      MemHandleUnlock(routeDBEntry);
      
      /* mark row as usable */
      TblSetRowUsable(tblP, row, true);
      
      /* mark all rows as invalid to trigger table update by FrmDrawForm() */
      TblMarkRowInvalid(tblP, row);
    } else {
      /* rows empty, mark as not usable */
      TblSetRowUsable(tblP, row, false);
    }
  }

  sclP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, RouteScrl));

  /* set scrollbar: value, min, max, pagesize */
  if (max_rows <= numRows) {
    SclSetScrollBar(sclP, 0, 0, 0, numRows);
  } else {
    SclSetScrollBar(sclP, gRouteTopRow, 0, max_rows - numRows, numRows);
  }
  
  /*
   * Selected route scolled below visible table rows,
   *   selection is switched to last visible entry.
   */
  if ( gRouteSel > (gRouteTopRow + numRows - 1) ) {
    gRouteSel = gRouteTopRow + numRows - 1;
  }

  /*
   * Selected route scolled above visible table rows,
   *   selection is switched to first visible entry.
   */  
  if (gRouteSel < gRouteTopRow) {
    gRouteSel = gRouteTopRow;
  }
  
  for (row = 0; row < numRows; row++, top++) {
    if ((top < max_rows) && (top == gRouteSel)) {
      TblSelectItem(tblP, row, 0);
      break;
    }
  }
  
  FrmDrawForm(frmP);
}


/****************************************************************************/
/**
 * \brief	Terminate Garmin mode, close host port and open GPS port.
 *
 * \note	This function could actually be called
 *              from RouteFormHandleEvent() upon reception of a
 *              frmCloseEvent. Unfortunately, when a Bluetooth GPS device
 *		is used and SrmOpen/SrmExtOpen opens the connection progress
 *		window, the frmOpenEvent queued by calling FrmGotoForm()
 *		previously seems to have been lost.
 *		As a (rather awkward) workaround, we have to call
 *		ChangePorts() _before_ FrmGotoForm() (scattered throughout
 *		RouteFormHandleEvent()).
 *
 ****************************************************************************/
static void ChangePorts(void)
{
  Err		err;		/* error code */

  /* Terminate Garmin mode and close serial port */
  GPSEndOfOperation();

  /*
   * open GPS port
   *
   * Note:	At this stage we have to cut corners. Even if the
   *		NewSerialManager is installed, we do not open the
   *		PortForm again to let the user select the connection
   *		profile. The error checking is also slightly simplified
   *          (because we have already checked that the serial port
   *          is available).
   */
  if (gNewSerialManager) {

    if (gSerialPort == sysFileCVirtRfComm) {
      /*
       * Bluetooth Virtual Serial Port --
       * Connect to known device (address)
       */
      err = BluetoothOpen(gSerialPort, gPrefs.serprefs.gps_baud,
	gPrefs.serprefs.gps_btaddr, &gPortID);

    } else {
      /* open non-BT serial port */
      err = SrmOpen(gSerialPort, gPrefs.serprefs.gps_baud, &gPortID);
    }
  } else {
    err = SerOpen(gPortID, 0, kBaudRate);
  }

  if (err) {
    EventType		event;			/* event to queue */

    FrmAlert(CantOpenSerialAlert);
    event.eType = appStopEvent;
    EvtAddEventToQueue(&event);
  }
}


/****************************************************************************/
/**
 * \brief	Route Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean RouteFormHandleEvent(EventPtr eventP)
{
  Boolean		handled = false;	/* event handled flag */
  Err			err;			/* error */
  FormPtr		frmP;			/* form ptr */
  FieldPtr		fldP;			/* field ptr */
  TablePtr		tblP;			/* table ptr */
  UInt16		rec;			/* record index */
  MemHandle		routeDBEntry;		/* route record handle */
  route_t		*route;

  frmP = FrmGetActiveForm();
  tblP = (TablePtr)GetObjectFromForm(frmP, RouteTable);
	
  /* Number of Waypoints/Routes (could have been modified by download) */
  nr_waypoints = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);
  nr_routes = DmNumRecordsInCategory(gRouteDB, dmAllCategories);
  
  switch (eventP->eType) {
  case frmOpenEvent:
    fldP = (FieldPtr)GetObjectFromActiveForm(RouteGpslibField);
    
    /* close serial port - Garmin uses 9600 baud */
    DoClose(gPortID);
    
    /* configure GPS4Palm to emulate Garmin III */
    GarminSetHostMode(fldP, GIII /* EmulatedGPS */);
    
    /* receive waypoints from host */
    GarminGetWpts(store_wpt, fldP);
    
    /* receive routes from host */
    GarminGetRtes(store_rte, fldP);

    /* receive tracks from host */
    GarminGetTrks(store_trk, fldP);
    
    frmP = FrmGetActiveForm();
    RouteFormInit(frmP);
    FrmDrawForm(frmP);
        
    /* draw the table */
    routeDrawTable();

    handled = true;
    break;

  case nilEvent:
    GpslibTicks();
    handled = true;
    break;

  case TransferWptReqEvent:
    /* gpslib - request to transfer waypoints to host*/
    fldP = (FieldPtr)GetObjectFromActiveForm(RouteGpslibField);
    GarminPutWpts(get_next_wpt, fldP, nr_waypoints);
    handled = true;
    break;
  
  case TransferRteReqEvent:
    /* gpslib - request to transfer routes to host */
    fldP = (FieldPtr)GetObjectFromActiveForm(RouteGpslibField);
    GarminPutRtes(get_next_rte, fldP, NumRtePackets());
    handled = true;
    break;
    
  case TransferTrkReqEvent:
    /* gpslib - request to transfer tracks */
    fldP = (FieldPtr)GetObjectFromActiveForm(WaypointGpslibField);
    GarminPutTrks(get_next_trk, fldP, NumTrkPackets());
    handled = true;
    break;

  case menuEvent:
    switch (eventP->data.menu.itemID) {

      case PositionMenu:
	ChangePorts();
	FrmGotoForm(GPSMainForm);
	handled = true;
	break;

      case SkyviewMenu:
	ChangePorts();
	FrmGotoForm(SkyviewForm);
	handled = true;
	break;

      case NavigationMenu:
	ChangePorts();
	FrmGotoForm(NavigationForm);
	handled = true;
	break;

      case MapMenu:
	ChangePorts();
	FrmGotoForm(MapForm);
	handled = true;
	break;

      case MiscOptsMenu:
	ChangePorts();
	FrmGotoForm(MiscOptsForm);
	handled = true;
	break;

      case MapOptsMenu:
	ChangePorts();
	FrmGotoForm(MapOptsForm);
	handled = true;
	break;

      case PortsMenu:
	if (gNewSerialManager) {
	  ChangePorts();
	  FrmGotoForm(GPSPortForm);
	}
	handled = true;
	break;

      case WaypointMenu:
	ChangePorts();
	FrmGotoForm(WaypointForm);
	handled = true;
	break;

      case TrackMenu:
	ChangePorts();
	FrmGotoForm(TrackForm);
	handled = true;
	break;

      case AboutMenu:
	ChangePorts();
	FrmGotoForm(AboutForm);
	handled = true;
	break;

    }
    break;  

  case tblSelectEvent:
    /* update table selection */
    gRouteSel = eventP->data.tblSelect.row + gRouteTopRow;
    
    /* always select column 0 */
    if (eventP->data.tblSelect.column != 0) {
      TblSelectItem(tblP, eventP->data.tblSelect.row, 0);
    }           
    handled = true;
    break;
  
  case sclRepeatEvent:
    /* handle scrollbar */
    Doscroll(RouteTable, RouteScrl, nr_routes, &gRouteTopRow);
    routeDrawTable();
    /* Note: scrollbar needs to handle the event, too */
    handled = false;
    break;
  
  case keyDownEvent:
    switch (eventP->data.keyDown.chr) {
    
    case pageUpChr:
    case pageDownChr:
      Dopage(RouteTable, (eventP->data.keyDown.chr == pageDownChr), nr_routes,
        &gRouteTopRow);
      routeDrawTable();
      /* Note: scrollbar needs to handle the event, too */
      handled = false;
      break;
    }
    break;
  
  case ctlSelectEvent:
    switch (eventP->data.ctlSelect.controlID) {

      case RouteDeleteButton:
        if (nr_routes == 0)
	  break;

	if ( FrmAlert(RouteDelAlert) == 0 ) {	  
	  /* get record index from row data */
	  rec = (UInt16)TblGetRowData(tblP, gRouteSel - gRouteTopRow);
          
          /*
           * Only data base backup, no synchronizing available.
	   * This means a record can simply be deleted.
           */
	  DmRemoveRecord(gRouteDB, rec);

          /* update number of records */
	  nr_routes = DmNumRecordsInCategory(gRouteDB, dmAllCategories);
	  
	  /* redraw the table */
	  gRouteTopRow = 0;
	  gRouteSel = 0;
	  routeDrawTable();
	}
	handled = true;
	break;
    
      case RouteNewButton:
        /* create new route */
	ChangePorts();
	FrmGotoForm(RouteEditForm);
	gSelRoute.valid = false;
	handled = true;
	break;

      case RouteEditButton:
        /* modify route */
        if (nr_routes == 0)
	  break;

	/* get record index from row data */
	rec = (UInt16)TblGetRowData(tblP, gRouteSel - gRouteTopRow);
	
	/* get unique route ID */
        err = DmRecordInfo(gRouteDB, rec, NULL /* *attrP */,
	  &gSelRoute.routeID /* *uniqueIDP */, NULL /* *chunkIDP */);
	
	/* mark selected waypoint as valid/invalid */
	gSelRoute.valid = (err == errNone) ? true : false;

	ChangePorts();
	FrmGotoForm(RouteEditForm);
	handled = true;
	break;
      
      case RouteSelectButton:
        if (nr_routes == 0)
	  break;
	

	/* get record index from row data */
	rec = (UInt16)TblGetRowData(tblP, gRouteSel - gRouteTopRow);
	
	/* get unique Route ID */
        err = DmRecordInfo(gRouteDB, rec, NULL /* *attrP */,
	  &gPrefs.act_rte.routeID /* *uniqueIDP */, NULL /* *chunkIDP */);
	
	/* mark active route as valid/invalid */
	gPrefs.act_rte.valid = (err == errNone) ? true : false;
	
	/* start with first waypoint on route */
	gPrefs.act_rte.wpt_index = 0;
		    /* get active route */
	
	err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID,
	  &rec);

	if (!err) {
	  /* active route found */
	  routeDBEntry = DmQueryRecord(gRouteDB, rec);

	  /* lock record */
	  route = (route_t *)MemHandleLock(routeDBEntry);

          /* an active route must contain at least two waypoints */
          if (route->items > 1) {
	    gPrefs.act_wpt.waypointID = 
	      route->wpt_rec_id[gPrefs.act_rte.wpt_index];
	    gPrefs.act_wpt.valid = true;
          
	  } else {
	    gPrefs.act_rte.valid = false;
	    gPrefs.act_wpt.valid = false;
	  }
	  
	  /* unlock route record */
	  MemHandleUnlock(routeDBEntry);
	  
	  /* initialize auto waypoint selection */
	  UpdateActWpt(0.0 /* dummy */, 0.0 /* dummy */, true /* init */);
		  
	  /* mark row as invalid to trigger table update by FrmDrawForm() */
          TblMarkRowInvalid(tblP, gRouteSel - gRouteTopRow );
          FrmDrawForm(frmP);
	
	} else {
	  gPrefs.act_wpt.valid = false;
	}

	handled = true;
	break;
    }
    break;

  case frmCloseEvent:
    if (gPrefs.act_rte.valid) {
      /* check if active route still exists, mark as invalid if not */
      err = DmFindRecordByID(gRouteDB, gPrefs.act_rte.routeID, &rec);
      gPrefs.act_rte.valid = (err) ? false : true;
    }

    SetFieldText(RouteGpslibField, NULL, false, true);

    RouteFormDeinit(FrmGetActiveForm());
    handled = false;
    break;

  default:
    break;
  }
  
  return handled;
}
