/*****************************************************************************
 *
 * $RCSfile: SkyviewForm.c,v $
 *
 * GPS4Palm Sky View Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-06 13:37:47 $
 *
 * $Revision: 1.19 $
 *
 * $Log: SkyviewForm.c,v $
 * Revision 1.19  2005-05-06 13:37:47  mp
 * added High-Density Display support
 *
 * Revision 1.18  2005/04/02 10:09:43  mp
 * modified Position2Clipboard() and Position2Geodb() to return 'handled'
 *
 * Revision 1.17  2005/04/02 07:26:59  mp
 * added PosgeodbMenu event handling
 *
 * Revision 1.16  2005/03/25 13:45:11  mp
 * added PosclipMenu event handling
 *
 * Revision 1.15  2005/01/23 18:40:29  mp
 * added PortsMenu event handling
 *
 * Revision 1.14  2004/12/08 20:48:23  mp
 * added track logging
 *
 * Revision 1.13  2004/11/30 20:44:19  mp
 * added TrackMenu event handling
 *
 * Revision 1.12  2004/11/27 10:20:31  mp
 * replaced error handling code by calling Die()
 *
 * Revision 1.11  2004/11/26 20:08:36  mp
 * added error handling for WinCreateOffscreenWindow()
 *
 * Revision 1.10  2004/11/25 20:08:42  mp
 * added NotifyApproach()
 *
 * Revision 1.9  2004/11/25 16:57:46  mp
 * added function call UpdateActWpt()
 *
 * Revision 1.8  2004/11/24 21:27:49  mp
 * moved static function declarations from header to implementation file,
 * added AboutForm menu event handling
 *
 * Revision 1.7  2004/11/23 17:47:31  mp
 * removed unused variable
 *
 * Revision 1.6  2004/04/29 18:18:15  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.5  2004/02/28 17:21:43  mp
 * added menuEvent WaypointMenu
 *
 * Revision 1.4  2003/11/20 20:46:31  mp
 * added status redraw on frmOpenEvent, added Menu Event MapOptsMenu
 *
 * Revision 1.3  2003/11/18 20:54:49  mp
 * Added MiscOptsMenu, added current Form to Application Preferences.
 *
 * Revision 1.2  2003/10/20 17:24:48  mp
 * added field deallocation and gSkyViewH window handle deletion on frmCloseEvent
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "ResourceDefines.h"
#include "Serial.h"
#include "Sinetab.h"
#include "SkyviewForm.h"
#include "GPS.h"
#include "HandleMessage.h"
#include "Utils.h"
#include "Data.h"		/* UpdateActWpt() */
#include "Clip.h"
#include "common.h"

/** SkyView window handle for save-behind (Sky View Plot). */
WinHandle	gSkyViewH;

/* Preferences data structure */
extern PrefsType gPrefs;

/* GPS Data */
extern GPSType		gGPSData;

/* High-Density Display */
extern Boolean		gHdFtrSet;

#ifndef WinDrawPixel
#define WinDrawPixel(x,y) WinDrawLine((x),(y),(x),(y))
#endif


/* Static Functions */
static void SkyviewFormInit(void)			SKYVIEW_SECTION;
static void DrawSkyView(void)				SKYVIEW_SECTION;


static 
void SkyviewFormInit(void)
{ 
  gFormOpened = true;
}

static
void DrawSkyView(void)
{
  Coord		maxx;		/* screen width */
  Coord		maxy;		/* screen height */
  UInt16	x;		/* x-coordinate */
  UInt16	y;		/* y-coordinate */
  PointType	p1;		/* a point */
  PointType	p2;		/* another point */
  UInt16	r;		/* radius */
  UInt16	phi;		/* angle */
  UInt16	err;		/* error code */
  RectangleType	rect;		/* rectangle structure */

  if (gHdFtrSet) {
    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesNative);
  }
  
  /* get screen dimensions */
  WinGetDisplayExtent(&maxx, &maxy);

  /* Base line for S/N bar graph */
  y = (gHdFtrSet) ? WinScaleCoord(SKY_L2, false) : SKY_L2;
  WinDrawLine(0, y, maxx-1, y);
  y = (gHdFtrSet) ? WinScaleCoord(SKY_L1, false) : SKY_L1;
  WinDrawLine(0, y, maxx-1, y);  

  /* Tick marks for S/N bar graph */
  for (y=SKY_L1+5; y<SKY_L2; y += 5) {
    /* left */
    p1.x = 0;
    p1.y = y;
    p2.x = 2;
    p2.y = y;
    
    if (gHdFtrSet) {
      WinScalePoint(&p1, false);
      WinScalePoint(&p2, false);
    }
   
    WinDrawLine(p1.x, p1.y, p2.x, p2.y);
    
    /* right */
    p1.x = 157;
    p1.y = y;
    p2.x = 159;
    p2.y = y;
    
    if (gHdFtrSet) {
      WinScalePoint(&p1, false);
      WinScalePoint(&p2, false);
    }
    
    WinDrawLine(p1.x, p1.y, maxx-1, p2.y);

  } /* for (y=SKY_L1+5; y<SKY_L2; y += 5) */
  
  /* Outer circle */
  r    = (gHdFtrSet) ? WinScaleCoord(SKY_R2, false) : SKY_R2;  
  p1.x = (gHdFtrSet) ? WinScaleCoord(SKY_X, false) : SKY_X;
  p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_Y, false) : SKY_Y;
  for (phi=0; phi<360; phi++) {
    /* Sinus/Cosinus functions using table lookup */
    x = p1.x + r * icos(phi) / 32768;
    y = p1.y - r * isin(phi) / 32768;
  
    /*
     * Note:
     * For High-Density, a step size of one degree does not provide
     * enough points for a continuous circle. Connecting the points
     * with a line seems to work good enough. (and is still fast) 
     */
    if (phi == 0) {
      WinDrawPixel(x, y);
    } else {
      WinDrawLine(x, y, p2.x, p2.y);
    }
    p2.x = x;
    p2.y = y;
  } /* for (phi=0; phi<360; phi++) */

  /* Inner circle */
  r = (gHdFtrSet) ? WinScaleCoord(SKY_R1, false) : SKY_R1;
  for (phi=0; phi<360; phi++) {
    /* Sinus/Cosinus functions using table lookup */
    x = p1.x + r * icos(phi) / 32768;
    y = p1.y - r * isin(phi) / 32768;
  
    WinDrawPixel(x, y);
  } /* for (phi=0; phi<360; phi++) */
  
  /* Center */
  p1.x = SKY_X-2;
  p1.y = SKY_Y;
  p2.x = SKY_X+2;
  p2.y = SKY_Y;
  if (gHdFtrSet) {
    WinScalePoint(&p1, false);
    WinScalePoint(&p2, false);
  }
  WinDrawLine(p1.x, p1.y, p2.x, p2.y);

  p1.x = SKY_X;
  p1.y = SKY_Y-2;
  p2.x = SKY_X;
  p2.y = SKY_Y+2;
  if (gHdFtrSet) {
    WinScalePoint(&p1, false);
    WinScalePoint(&p2, false);
  }
  WinDrawLine(p1.x, p1.y, p2.x, p2.y);
  
  p1.x = 2 * SKY_R2;
  p1.y = 2 * SKY_R2;
  if (gHdFtrSet) WinScalePoint(&p1, false);
  gSkyViewH = WinCreateOffscreenWindow(p1.x, p1.y, nativeFormat, &err);
  
  if (!gSkyViewH) {
    /* error handling */
    Die("Cannot create OffscreenWindow in SkyviewForm!");
    return;
  }
  
  p1.x = SKY_X-SKY_R2;
  p1.y = SKY_Y-SKY_R2;
  if (gHdFtrSet) WinScalePoint(&p1, false);
  x = y = (gHdFtrSet) ? WinScaleCoord(2 * SKY_R2, false) : 2 * SKY_R2;
  RctSetRectangle(&rect, p1.x, p1.y, x /* w */, y /* h */);

  WinCopyRectangle(	NULL, /* WinHandle srcWin, draw window */
	  		gSkyViewH, /* WinHandle dstWin, draw window */
			&rect, /* RectangleType *srcRect */
			0, /* Coord destX */
			0, /* Coord destY */
			winPaint /* WinDrawOperation mode */);
  
  if (gHdFtrSet) {
    WinPopDrawState();
  } 
}


/****************************************************************************/
/**
 * \brief	Sky View Form event handler
 *
 * \param	event		pointer to event structure
 *
 * \return	event handled flag
 ****************************************************************************/
Boolean SkyviewFormHandleEvent(EventPtr event)
{
	Boolean		handled;
	Boolean		updatedDisplay;

	handled = false;
	switch (event->eType)
	{
        case menuEvent:
          switch (event->data.menu.itemID) {
      
            case PositionMenu:
	      FrmGotoForm(GPSMainForm);
	      handled = true;
	      break;

	    case NavigationMenu:
	      FrmGotoForm(NavigationForm);
	      handled = true;
	      break;

	    case MapMenu:
	      FrmGotoForm(MapForm);
	      handled = true;
	      break;

            case MiscOptsMenu:
	      FrmGotoForm(MiscOptsForm);
	      handled = true;
	      break;
	       
            case MapOptsMenu:
	      FrmGotoForm(MapOptsForm);
	      handled = true;
	      break;

	    case PortsMenu:
	      if (gNewSerialManager) {
	        FrmGotoForm(GPSPortForm);
	      }
	      handled = true;
	      break;
	    
            case WaypointMenu:
	      FrmGotoForm(WaypointForm);
	      handled = true;
	      break;
	    	       
            case RouteMenu:
	      FrmGotoForm(RouteForm);
	      handled = true;
	      break;

            case TrackMenu:
	      FrmGotoForm(TrackForm);
	      handled = true;
	      break;
	       
	    case PosclipMenu:
	      handled = Position2Clipboard(gGPSData, gPrefs.units);
	      break;

	    case PosgeodbMenu:
	      handled = Position2Geodb(gGPSData);
	      break;

            case AboutMenu:
	      FrmGotoForm(AboutForm);
	      handled = true;
	      break;
	      	    	       
#ifdef DEBUG_FORM
            case DebugMenu:
	      FrmGotoForm(DebugForm);
	      handled = true;
	      break;
#endif
	  }
          break;
	  
	case nilEvent:
		handled = true;
		/* throw away anything in the buffer-- we want fresh data */
		DoReceiveFlush(gPortID, 1);
		
		/* read data from GPS */
		updatedDisplay = ReadFromGPS();
		
		if (gGPSData.valid && gPrefs.act_rte.valid && gPrefs.auto_wpt) {
	          /* update active waypoint */
		  UpdateActWpt(gGPSData.lat, gGPSData.lon, false /* init */);
		}
		
		/* Active Waypoint approach detection and display */
      		NotifyApproach();

		/* Track Logging */
		if (gPrefs.act_trk.log_state) {
		  Boolean new_trk;

		  if (TrackIntervalCheck(false /* init */, &new_trk)) {
		    TrackWriteLog(new_trk);
		  }
		}		

		break;
	
	case frmOpenEvent:
		gPrefs.form = SkyviewForm;
		
		SkyviewFormInit();
		FrmDrawForm(FrmGetActiveForm());
		DrawSkyView();
		UpdateStatus(STAT_REDRAW);
		
		handled = true;
		break;
	
	case frmCloseEvent:
        	/* Deallocate Fields' Text Memory */
	  	SetFieldText(SkyViewStatusField, NULL, false, true);
	  	SetFieldText(TimeField, NULL, false, true);
	  	SetFieldText(PDOPField, NULL, false, true);
	  	SetFieldText(HDOPField, NULL, false, true);
	  	SetFieldText(VDOPField, NULL, false, true);
	  	SetFieldText(FixField, NULL, false, true);
	  	SetFieldText(SatsField, NULL, false, true);
	  	SetFieldText(BatField, NULL, false, true);
		if (gSkyViewH)
		  WinDeleteWindow(gSkyViewH, true /* eraseIt */);
		handled = false;
		break;
	}
	return(handled);
} /* SkyviewFormHandleEvent() */
