/*****************************************************************************
 *
 * $RCSfile: TrackForm.c,v $
 *
 * GPS4Palm Tracks Form
 *
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-06 13:40:29 $
 *
 * $Revision: 1.13 $
 *
 * $Log: TrackForm.c,v $
 * Revision 1.13  2005-05-06 13:40:29  mp
 * added workaround for Bluetooth-Open Bug
 *
 * Revision 1.12  2005/01/30 21:22:13  mp
 * added PortsMenu event handling
 *
 * Revision 1.11  2005/01/23 10:01:32  mp
 * changed re-opening of GPS serial port after data transfer
 *
 * Revision 1.10  2004/12/18 16:11:58  mp
 * fixed serial port re-opening for PDAs with New Serial Manager
 *
 * Revision 1.9  2004/12/14 20:10:34  mp
 * fixed alert
 *
 * Revision 1.8  2004/12/10 19:54:45  mp
 * replaced DmGet1Resource by DmGetResource, implemented clipboard,
 * SysKeyboardDialog, and SysGraffitiReferenceDialog
 *
 * Revision 1.7  2004/12/09 17:29:25  mp
 * replaced default track comment and Start/Stop button labels by
 * string resources
 *
 * Revision 1.6  2004/12/08 20:47:45  mp
 * added track logging, fixed table display
 *
 * Revision 1.5  2004/12/07 17:20:50  mp
 * removed TrackSelectButton event handling,
 * modified TrackLogButton event handling (gTrackDB open/close,
 * Active Track selection)
 *
 * Revision 1.4  2004/12/06 21:12:57  mp
 * implemented display and editing of TrackIdentField
 *
 * Revision 1.3  2004/12/04 11:48:01  mp
 * modified TrackTable display and track deleting, added track transfer
 *
 * Revision 1.2  2004/12/02 23:22:48  mp
 * added track header adding/deletion and table handling
 *
 * Revision 1.1  2004/11/30 21:10:41  mp
 * initial version
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <SerialMgrOld.h>
#include "ResourceDefines.h"
#include "gpslib.h"
#include "TrackForm.h"
#include "Data.h"
#include "DataExchange.h"
#include "Utils.h"
#include "Serial.h"
#include "geo.h"
#include "fp.h"
#include "common.h"

/* Global Variables */
extern PrefsType	gPrefs;			/* application preferences */
extern UInt32		gSerialPort;		/* serial port name/no. */
ActRouteType		gSelTrack;		/**< selected track */

/* Local Variables */
static UInt16		nr_waypoints;		/* number of waypoints */
static UInt16		nr_routes;		/* number of routes */
static UInt16		nr_tracks;		/* number of tracks */


/* Static Functions */
static void TrackFormInit(FormPtr frmP)			TRACK_SECTION;
static void TrackFormDeinit(FormPtr frmP)		TRACK_SECTION;
static Err create_track(void)				TRACK_SECTION;
static void ChangePorts(void)				TRACK_SECTION;


/*****************************************************************************
 * FUNCTION:	TrackFormInit
 *
 * DESCRIPTION: Track Form Init Function
 *
 ****************************************************************************/
static void TrackFormInit(FormPtr frmP)
{
#pragma unused(frmP)
  // warning-- don't do any drawing in this routine.
  // Also, don't call FrmSetFocus from here (it must be called *after*
  // FrmDrawForm)
  
}


/*****************************************************************************
 * FUNCTION:	TrackFormDeinit
 *
 * DESCRIPTION:	Track Form De-Init Function
 *
 ****************************************************************************/
static void TrackFormDeinit(FormPtr frmP)
{
#pragma unused(frmP)

}


/*****************************************************************************
 * FUNCTION:	TableDrawCommentCell
 *
 * DESCRIPTION:	Custom table cell drawing function for Route Number
 *              (max. 3 digits) and Route Comment (string, max. 20
 *		characters). The DB record number is retrieved via
 *              TblGetRowData().
 *
 * GLOBALS:	gRouteDB	-- route DB reference
 *
 * PARAMETERS:	Note: The function's prototype is defined by PalmOS
 *		      (TableDrawItemFuncType).
 *
 * RETURNED:	-
 ****************************************************************************/
static void TableDrawCell(void *tblP, Int16 row, Int16 column,
                   RectangleType *bounds)
{
  Char			str[51];		/* string buffer */
  MemHandle		trkhDBEntry;		/* record handle */
  UInt32		trackID;		/* track ID */ 
  UInt16		rec;			/* record index */
  trk_header_t		*trkh;			/* track header */
  Int16			cell_width;		/* table cell width in pixels */
  Boolean		fits;			/* string fits in table cell */
  Int16			offset;			/* offset within table cell */
  Int16			length;			/* string length */
  DmOpenRef		db_ref;			/* data base reference */
  UInt16		nr_trackpoints;		/* number of track points */
  
  /* get record index from row data */
  rec = (UInt16)TblGetRowData(tblP, row);
  
  /* Display track identifier (truncated to fit) */
  if (column == 0) {
    WinEraseRectangle(bounds, 0);
    
    if (TblRowUsable(tblP, row)) {
      trkhDBEntry = DmQueryRecord(gTrackHdrDB, rec);

      /* get unique Track ID */
      DmRecordInfo(gTrackHdrDB, rec, NULL /* *attrP */,
	&trackID /* *uniqueIDP */, NULL /* *chunkIDP */);

      if (gPrefs.act_trk.valid && (trackID == gPrefs.act_trk.trackID)) {
        FntSetFont(boldFont);
      } else {
        FntSetFont(stdFont);
      }

      trkh = (trk_header_t *)MemHandleLock(trkhDBEntry);
      MemSet((Char *)str, sizeof(str), 0);
      StrCopy(str, (Char *)trkh->trk_ident);
      MemHandleUnlock(trkhDBEntry);
      
      /* truncate string to fit in table cell */
      length = StrLen(str);
      cell_width = bounds->extent.x;
      FntCharsInWidth(str, &cell_width, &length, &fits);
     
      /* display track identifier */
      WinDrawChars(str, length, bounds->topLeft.x + 1, bounds->topLeft.y);
      
    } /* (TblRowUsable(tblP, row)) */
  } /* if (column == 0) */
  
  /* Display number of track points (right aligned, bold) */
  if (column == 1) {
    WinEraseRectangle(bounds, 0);
    
    /*
     * Set font to bold, because route items column (which is numeric, too)
     * uses it due to ItemStyle numericTableItem.
     */
    FntSetFont(boldFont);

    if (TblRowUsable(tblP, row)) {
      trkhDBEntry = DmQueryRecord(gTrackHdrDB, rec);

      /* get unique Track ID */
      DmRecordInfo(gTrackHdrDB, rec, NULL /* *attrP */,
	&trackID /* *uniqueIDP */, NULL /* *chunkIDP */);      

      /* create track data base name */
      TrackDBName(str, trackID);

      /* open track point data base */
      db_ref = DmOpenDatabase(0 /* cardNo */,
        DmFindDatabase(0 /* cardNo */, str), dmModeReadOnly);
      
      if (db_ref) {
        /* open succeeded, get number of track points */
        nr_trackpoints = DmNumRecordsInCategory(db_ref, dmAllCategories);
	
	/* close DB */
        DmCloseDatabase(db_ref);
        
        /* print number of track points to string */
	StrPrintF(str, "%u", nr_trackpoints);
	
	/* calculate offset to right-align number */
	length = StrLen(str);
        cell_width = bounds->extent.x;
        offset = cell_width - FntCharsWidth(str, length);
	
	/* display no. of track points, aligned right */
	WinDrawChars(str, length,
	  bounds->topLeft.x + offset, bounds->topLeft.y);
      } /* if (db_ref) */
      
    } /* if (TblRowUsable(tblP, row)) */
  } /* if (column == 1) */
}


/****************************************************************************/
/**
 * \brief	Tracks table drawing function
 *		TBD: adapt template from RouteForm for tracks
 *
 * \param	top		top table row
 * \param	sel		pointer to selected table row
 *
 *****************************************************************************/
void trackDrawTable(Int32 top, UInt16 *sel)
{
  TablePtr 		tblP;			/* table ptr */
  FormPtr 		frmP;			/* form ptr */
  ScrollBarPtr		sclP;			/* scrollbar ptr */
  UInt16		row;			/* table row counter */
  UInt16		rec;			/* record nr */
  UInt16		max_rows;		/* total nr. of routes */
  UInt16		numRows;		/* number of visible rows */
  MemHandle		trkhDBEntry;		/* record handle */
  trk_header_t		*trkh;			/* track header */
  Err			err;			/* error code */
  
  frmP = FrmGetActiveForm();
  tblP = (TablePtr)GetObjectFromForm(frmP, TrackTable);
  sclP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, TrackScrl));
  
  /* total number of routes */
  max_rows = DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories);
  
  /* number of visible rows */
  numRows = TblGetNumberOfRows(tblP);

  /* set columns usable */
  TblSetColumnUsable(tblP, 0, true);
  TblSetColumnUsable(tblP, 1, true);
      
  /* set row data to Track Header DB entries (record index) */
  for (row = 0; row < numRows; row++) {

    /* default: mark row as not usable (empty or no valid record) */
    TblSetRowUsable(tblP, row, true);

    if (row + top < max_rows) {
      /* seek from beginning of database */
      rec = 0;
      
      /* get next record */
      err = DmSeekRecordInCategory(gTrackHdrDB, &rec /* index */,
        top + row /* offset */, dmSeekForward, dmAllCategories);
      
      if (err == errNone) {
        /* query record */
        trkhDBEntry = DmQueryRecord(gTrackHdrDB, rec);

	if (trkhDBEntry) {
	  /*
	   * Record is valid, set row data to index and
	   * mark table row as usable.
	   */
	  TblSetRowData(tblP, row, rec);
          TblSetRowUsable(tblP, row, true);
	}
      }
    } /*  if (row + top < max_rows) */
  } /* for (row = 0; row < numRows; row++) */  
  
  /* initialize table */
  for (row = 0; row < numRows; row++) {
    
    if (row + top < max_rows) {
      /* set all rows selectable */
      TblSetRowSelectable(tblP, row, true);

      /* get record index from row data */
      rec = TblGetRowData(tblP, row);
      
      /* track identifier */
      TblSetItemStyle(tblP, row, 0, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 0, TableDrawCell);

      /* number of track points */
      /*
       * Note:	ItemStyle customTableItem is used instead of
       *	numericTableItem, because otherwise
       *	TblSelectItem() does not work properly.
       */      
      TblSetItemStyle(tblP, row, 1, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 1, TableDrawCell);
      
      /* mark row as usable */
      TblSetRowUsable(tblP, row, true);
      
      /* mark all rows as invalid to trigger table update by FrmDrawForm() */
      TblMarkRowInvalid(tblP, row);
    } else {
      /* rows empty, mark as not usable */
      TblSetRowUsable(tblP, row, false);
    }
  }
  
  /* set scrollbar: value, min, max, pagesize */
  if (max_rows <= numRows) {
    SclSetScrollBar(sclP, 0, 0, 0, numRows);
  } else {
    SclSetScrollBar(sclP, top, 0, max_rows - numRows, numRows);
  }

  /*
   * Selected route scolled below visible table rows,
   *   selection is switched to last visible entry.
   */
  if ( *sel > (top + numRows - 1) ) {
    *sel = top + numRows - 1;
  }

  /*
   * Selected route scolled above visible table rows,
   *   selection is switched to first visible entry.
   */  
  if (*sel < top) {
    *sel = top;
  }

  for (row = 0; row < numRows; row++, top++) {
    if ((top < max_rows) && (top == *sel)) {
      TblSelectItem(tblP, row, 0);
      break;
    }
  }

  /* get record handle of selected track */
  trkhDBEntry = DmQueryRecord(gTrackHdrDB, *sel);

  if (trkhDBEntry) {
    /* show track identifier */
    trkh = (trk_header_t *)MemHandleLock(trkhDBEntry);
    SetFieldText(TrackIdentField, trkh->trk_ident, false, true);
    MemHandleUnlock(trkhDBEntry);
  }  

  FrmDrawForm(frmP);
}


/****************************************************************************/
/**
 * \brief	Add new track to database
 *
 * GLOBALS:	gTrackHdrDB	-- Route Database Reference
 *
 *
 * \returned	false if Track Database full/max. number of entries reached
 ****************************************************************************/
static
Err create_track(void)
{
  UInt16	index;			/* record index */
  MemHandle	trkhDBEntry;		/* track header record handle */
  trk_header_t	trkhdr;			/* track header record ptr */
  UInt32	trackID;		/* track ID */
  UInt32	seconds;		/* current time */
  DateTimeType	dt;			/* date/time structure */
  Char		dbname[dmDBNameLength];	/* data base name */
  Err		err;			/* error code */
  MemHandle	strH;			/* string handle */ 
  MemPtr	strP;			/* string ptr */
  
  /* Limit number of Tracks */
  if ( DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories) == 0x100 )
    return 1;
  
  /* No particular order - add at the end of the database */
  index = dmMaxRecordIndex;
  
  trkhDBEntry = DmNewRecord(gTrackHdrDB, &index, sizeof(trk_header_t));
  
  if (!trkhDBEntry) {
    err = DmGetLastErr();
    
    if (err == memErrNotEnoughSpace) {
      /* Data Base is full */
      return err;
    } else {
      /* Some other error occurred - fatal! */
      Die("Cannot create track header record!");
      return err;
    }
  }
  
  /* release new record */
  DmReleaseRecord(gTrackHdrDB, index, false /* dirty */);
  
  /* set default display flag */
  trkhdr.dspl = true;
  
  /* set default color */
  trkhdr.color = Default_Color;
  
  /*
   * create comment: <Track yy-mm-dd hh:mm>
   */
  MemSet((Char *)trkhdr.trk_ident, sizeof(trkhdr.trk_ident), 0);

  /* get string from resource */
  strH = DmGetResource(strRsc, TrackStr);
  if (strH != NULL) {
    /* string handle valid, lock it */
    strP = MemHandleLock(strH);
  
    /* copy string */
    StrCopy(trkhdr.trk_ident, strP);

    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  }

  
  /* get current time and date */
  seconds = TimGetSeconds();
  TimSecondsToDateTime(seconds, &dt);
  
  /* write date to string */
  DateToAscii(dt.month, dt.day, dt.year,
    dfYMDWithDashes,
    &trkhdr.trk_ident[StrLen(trkhdr.trk_ident)]);

  /* append space */
  StrCat(trkhdr.trk_ident, " ");
  
  /* write time to string */
  TimeToAscii(dt.hour, dt.minute,
    tfColon24h,
    &trkhdr.trk_ident[StrLen(trkhdr.trk_ident)]);
  
  /* append bracket */
  StrCat(trkhdr.trk_ident, ">");
  
  /* Store Route Header in Database */
  StoreTrkHdr(&trkhdr, index);    

  DmRecordInfo(gTrackHdrDB, index, NULL /* *attrP */,
    &trackID /* *uniqueIDP */, NULL /* *chunkIDP */);

  TrackDBName(dbname, trackID);
  
  err = DmCreateDatabase(0 /* cardNo */, dbname,
    kCreatorId /* creator */, kTrkDBType /* type */, false /* resDB */);
  
  if (err != errNone) {
    DmRemoveRecord(gTrackHdrDB, index);
    return err;
  }
  
  /* success */
  return 0;
}


/****************************************************************************/
/**
 * \brief	Terminate Garmin mode, close host port and open GPS port.
 *
 * \note	This function could actually be called
 *              from TrackFormHandleEvent() upon reception of a
 *              frmCloseEvent. Unfortunately, when a Bluetooth GPS device
 *		is used and SrmOpen/SrmExtOpen opens the connection progress
 *		window, the frmOpenEvent queued by calling FrmGotoForm()
 *		previously seems to have been lost.
 *		As a (rather awkward) workaround, we have to call
 *		ChangePorts() _before_ FrmGotoForm() (scattered throughout
 *		TrackFormHandleEvent()).
 *
 ****************************************************************************/
static void ChangePorts(void)
{
  Err		err;		/* error code */

  /* Terminate Garmin mode and close serial port */
  GPSEndOfOperation();

  /*
   * open GPS port
   *
   * Note:	At this stage we have to cut corners. Even if the
   *		NewSerialManager is installed, we do not open the
   *		PortForm again to let the user select the connection
   *		profile. The error checking is also slightly simplified
   *          (because we have already checked that the serial port
   *          is available).
   */
  if (gNewSerialManager) {

    if (gSerialPort == sysFileCVirtRfComm) {
      /*
       * Bluetooth Virtual Serial Port --
       * Connect to known device (address)
       */
      err = BluetoothOpen(gSerialPort, gPrefs.serprefs.gps_baud,
	gPrefs.serprefs.gps_btaddr, &gPortID);

    } else {
      /* open non-BT serial port */
      err = SrmOpen(gSerialPort, gPrefs.serprefs.gps_baud, &gPortID);
    }
  } else {
    err = SerOpen(gPortID, 0, kBaudRate);
  }

  if (err) {
    EventType		event;			/* event to queue */

    FrmAlert(CantOpenSerialAlert);
    event.eType = appStopEvent;
    EvtAddEventToQueue(&event);
  }
}


/****************************************************************************/
/**
 * \brief	Track Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean TrackFormHandleEvent(EventPtr eventP)
{
  Boolean		handled = false;	/* event handled flag */
  Err			err;			/* error */
  FormPtr		frmP;			/* form ptr */
  FieldPtr		fldP;			/* field ptr */
  TablePtr		tblP;			/* table ptr */
  ScrollBarPtr		sclP;			/* scrollbar ptr */
  Char			*strP;			/* string ptr */
  MemHandle		strH;			/* string handle */ 
  FieldAttrType		fld_attr;		/* field attributes */
  MemHandle		trkhDBEntry;		/* record handle */
  trk_header_t		*trkh;			/* track header ptr */
  trk_header_t		trkheader;		/* track header */
  Char			dbname[dmDBNameLength];	/* data base name */
  UInt32		trackID;		/* track header record ID */
  UInt16		rec;			/* record index */
  UInt16		len;			/* string length */
  static Int32		track_top = 0;		/* track table top */
  static UInt16		track_sel = 0;		/* track selected row */

  frmP = FrmGetActiveForm();
  tblP = (TablePtr)GetObjectFromForm(frmP, TrackTable);
  fldP = (FieldPtr)GetObjectFromActiveForm(TrackIdentField);

  /* copy Track Identifier to Track Header record if modified */
  if ( FldDirty(fldP) ) {
    strP = FldGetTextPtr(fldP);
    len = StrLen(strP);
    
    /* get record handle of selected track */
    trkhDBEntry = DmQueryRecord(gTrackHdrDB, track_sel);

    if (trkhDBEntry) {
      /* copy other struct members */
      trkh = (trk_header_t *)MemHandleLock(trkhDBEntry);
      trkheader.dspl  = trkh->dspl;
      trkheader.color = trkh->color;
      MemSet((Char *)trkheader.trk_ident, 50, 0);
      MemHandleUnlock(trkhDBEntry);
    
      /* copy modified string */
      StrNCopy(trkheader.trk_ident, strP, (len > 49) ? 49 : len);

      /* update track header record in data base */
      StoreTrkHdr(&trkheader, track_sel);

      /* update table row */
      TblMarkRowInvalid(tblP, track_sel - track_top);
      FrmDrawForm(frmP);

      /* clear dirty flag */
      FldSetDirty(fldP, false);   
    } /* if (trkhDBEntry) */
  } /* if ( FldDirty(fldP) ) */


  /* Number of Waypoints/Routes/Tracks (could have been modified by download) */
  nr_waypoints = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);
  nr_routes    = DmNumRecordsInCategory(gRouteDB, dmAllCategories);
  nr_tracks    = DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories);
  
  switch (eventP->eType) {
  case frmOpenEvent:
    fldP = (FieldPtr)GetObjectFromActiveForm(TrackGpslibField);
    
    /* close serial port - Garmin uses 9600 baud */
    DoClose(gPortID);
    
    /* configure GPS4Palm to emulate Garmin III */
    GarminSetHostMode(fldP, GIII /* EmulatedGPS */);
    
    /* receive waypoints from host */
    GarminGetWpts(store_wpt, fldP);
    
    /* receive routes from host */
    GarminGetRtes(store_rte, fldP);

    /* receive tracks from host */
    GarminGetTrks(store_trk, fldP);
    
    frmP = FrmGetActiveForm();
    TrackFormInit(frmP);
    
    /* get string from resource */
    strH = DmGetResource(strRsc,
      (gPrefs.act_trk.log_state) ? StopStr : StartStr);
    
    if (strH != NULL) {
      /* string handle valid, lock it */
      strP = (Char *)MemHandleLock(strH);

      /* Set Start/Stop button label from preferences */
      CtlSetLabel((ControlPtr)GetObjectFromActiveForm(TrackLogButton), strP);

      /* unlock string handle */
      MemHandleUnlock(strH);

      /* release string resource */
      DmReleaseResource(strH);
    }

    FrmDrawForm(frmP);
        
    /* draw the table */
    trackDrawTable(track_top, &track_sel);
        
    handled = true;
    break;

  case nilEvent:
    GpslibTicks();
    handled = true;
    break;

  case TransferWptReqEvent:
    /* gpslib - request to transfer waypoints to host*/
    fldP = (FieldPtr)GetObjectFromActiveForm(TrackGpslibField);
    GarminPutWpts(get_next_wpt, fldP, nr_waypoints);
    handled = true;
    break;
  
  case TransferRteReqEvent:
    /* gpslib - request to transfer routes to host */
    fldP = (FieldPtr)GetObjectFromActiveForm(TrackGpslibField);
    GarminPutRtes(get_next_rte, fldP, NumRtePackets());
    handled = true;
    break;

  case TransferTrkReqEvent:
    /* gpslib - request to transfer tracks */
    fldP = (FieldPtr)GetObjectFromActiveForm(TrackGpslibField);
    GarminPutTrks(get_next_trk, fldP, NumTrkPackets());
    handled = true;
    break;
    
  case menuEvent:
    /* check if TrackIdentField has focus */
    fldP = (FieldPtr)GetObjectFromActiveForm(TrackIdentField);
    FldGetAttributes(fldP, &fld_attr);
    
    switch (eventP->data.menu.itemID) {

      case PositionMenu:
        ChangePorts();
	FrmGotoForm(GPSMainForm);
	handled = true;
	break;

      case SkyviewMenu:
        ChangePorts();
	FrmGotoForm(SkyviewForm);
	handled = true;
	break;

      case NavigationMenu:
        ChangePorts();
	FrmGotoForm(NavigationForm);
	handled = true;
	break;

      case MapMenu:
        ChangePorts();
	FrmGotoForm(MapForm);
	handled = true;
	break;

      case MiscOptsMenu:
        ChangePorts();
	FrmGotoForm(MiscOptsForm);
	handled = true;
	break;

      case MapOptsMenu:
        ChangePorts();
	FrmGotoForm(MapOptsForm);
	handled = true;
	break;

      case PortsMenu:
	if (gNewSerialManager) {
          ChangePorts();
	  FrmGotoForm(GPSPortForm);
	}
	handled = true;
	break;

      case WaypointMenu:
        ChangePorts();
	FrmGotoForm(WaypointForm);
	handled = true;
	break;

      case RouteMenu:
        ChangePorts();
	FrmGotoForm(RouteForm);
	handled = true;
	break;

      case AboutMenu:
        ChangePorts();
	FrmGotoForm(AboutForm);
	handled = true;
	break;
	
      case UndoMenu:
	if (fld_attr.hasFocus == 1)
	  FldUndo(fldP);
	handled = true;
	break;

      case CutMenu:
	if (fld_attr.hasFocus == 1)
	  FldCut(fldP);
	handled = true;
	break;

      case CopyMenu:
	if (fld_attr.hasFocus == 1)
	  FldCopy(fldP);
	handled = true;
	break;

      case PasteMenu:
	if (fld_attr.hasFocus == 1)
	  FldPaste(fldP);
	handled = true;
	break;

      case KeyboardMenu:
	if (fld_attr.hasFocus == 1)
	  SysKeyboardDialog(kbdAlpha);
	handled = true;
	break;

      case GraffitiMenu:
	SysGraffitiReferenceDialog(referenceDefault);
	handled = true;
	break;
    }
    break;  

  case tblSelectEvent:
    /* update table selection */
    track_sel = eventP->data.tblSelect.row + track_top;
    
    /* always select column 0 */
    if (eventP->data.tblSelect.column != 0) {
      TblSelectItem(tblP, eventP->data.tblSelect.row, 0);
    }
    
    /* get record index from row data */
    rec = (UInt16)TblGetRowData(tblP, track_sel - track_top);
    
    /* get record handle */
    trkhDBEntry = DmQueryRecord(gTrackHdrDB, rec);
    
    /* show track identifier */
    trkh = (trk_header_t *)MemHandleLock(trkhDBEntry);
    SetFieldText(TrackIdentField, trkh->trk_ident, false, true);
    MemHandleUnlock(trkhDBEntry);

    handled = true;
    break;
  
  case sclRepeatEvent:
    /* handle scrollbar */
    Doscroll(TrackTable, TrackScrl, nr_tracks, &track_top);
    trackDrawTable(track_top, &track_sel);
    /* Note: scrollbar needs to handle the event, too */
    handled = false;
    break;
  
  case keyDownEvent:
    switch (eventP->data.keyDown.chr) {
    
    case pageUpChr:
    case pageDownChr:
      Dopage(TrackTable, (eventP->data.keyDown.chr == pageDownChr), nr_tracks,
        &track_top);
      trackDrawTable(track_top, &track_sel);
      /* Note: scrollbar needs to handle the event, too */
      handled = false;
      break;
    }
    break;
  
  case ctlSelectEvent:
    switch (eventP->data.ctlSelect.controlID) {
      case TrackLogButton:
        if (nr_tracks == 0)
	  break;
       
        /* start/stop track log (toggle) */
        if (gPrefs.act_trk.log_state) {
	  /* old state: running, new state: stopped */

	  /* get string from resource */
	  strH = DmGetResource(strRsc, StartStr);

	  if (strH != NULL) {
	    /* string handle valid, lock it */
	    strP = (Char *)MemHandleLock(strH);

	    /* Set Start/Stop button label from preferences */
	    CtlSetLabel((ControlPtr)GetObjectFromActiveForm(TrackLogButton),
	      strP);

	    /* unlock string handle */
	    MemHandleUnlock(strH);

	    /* release string resource */
	    DmReleaseResource(strH);
	  }
	  
	  gPrefs.act_trk.log_state = false;
	  
	  /* mark active track as invalid */
	  gPrefs.act_trk.valid = false;

          /* close track point DB */
          DmCloseDatabase(gTrackDB);
	  gTrackDB = 0;
	  
	  /* mark row as invalid to trigger table update by FrmDrawForm() */
          TblMarkRowInvalid(tblP, track_sel - track_top );
          FrmDrawForm(frmP);

	} else {
	  /* old state: stopped, new state: running */

	  /* get string from resource */
	  strH = DmGetResource(strRsc, StopStr);

	  if (strH != NULL) {
	    /* string handle valid, lock it */
	    strP = (Char *)MemHandleLock(strH);

	    /* Set Start/Stop button label from preferences */
	    CtlSetLabel((ControlPtr)GetObjectFromActiveForm(TrackLogButton),
	      strP);

	    /* unlock string handle */
	    MemHandleUnlock(strH);

	    /* release string resource */
	    DmReleaseResource(strH);
	  }
	  
	  gPrefs.act_trk.log_state = true;
	  
	  /* get record index from row data */
	  rec = (UInt16)TblGetRowData(tblP, track_sel - track_top);

	  /* get unique Track ID */
          err = DmRecordInfo(gTrackHdrDB, rec, NULL /* *attrP */,
	    &gPrefs.act_trk.trackID /* *uniqueIDP */, NULL /* *chunkIDP */);

	  /* mark active track as valid/invalid */
	  gPrefs.act_trk.valid = (err == errNone) ? true : false;

          if (!err) {
	    Boolean new_trk = true;
	    
	    /* generate data base name */
	    TrackDBName(dbname, gPrefs.act_trk.trackID);
	    
	    /* open track DB */
	    gTrackDB = DmOpenDatabase(0 /* cardNo */,
	      DmFindDatabase(0 /* cardNo */, dbname), dmModeReadWrite);
	  
	    /* init track logging */
	    TrackIntervalCheck(true /* init */, &new_trk);
	  }

	  /* mark row as invalid to trigger table update by FrmDrawForm() */
          TblMarkRowInvalid(tblP, track_sel - track_top );
          FrmDrawForm(frmP);
	  
	}
        handled = true;
	break;
        

      case TrackDeleteButton:
        if (nr_tracks == 0)
	  break;
        
	/* get record index from row data */
	rec = (UInt16)TblGetRowData(tblP, track_sel - track_top);

	/* get unique recored ID */
	DmRecordInfo(gTrackHdrDB, rec, NULL /* *attrP */,
	  &trackID /* *uniqueIDP */, NULL /* *chunkIDP */);

	if (gPrefs.act_trk.valid && gPrefs.act_trk.trackID == trackID) {
	  /* selected track is active track, mark as invalid */
	  gPrefs.act_trk.valid = false;
	  
	  if (gPrefs.act_trk.log_state) {
	    /* selected track in use */
	    FrmAlert(TrackDelErrorAlert);
	    break;
	  }
	}
	
	if ( FrmAlert(TrackDelConfirmAlert) == 0 ) {          
	  /*
           * Only data base backup, no synchronizing available.
	   * This means a record can simply be deleted.
           */
	  DmRemoveRecord(gTrackHdrDB, rec);

          /* update number of records */
	  nr_tracks = DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories);
	  
	  /* 
	   * Redraw route table, select previous Waypoint
	   * of the route and make sure it is visible.
	   */
	  if (track_sel > 0) {
	    track_sel--;
	  }
	  
	  /* update scroll bar */
	  Doscroll(TrackTable, TrackScrl, nr_tracks, &track_top);	  

          /* redraw table */
	  trackDrawTable(track_top, &track_sel);
	  
	  /* generate data base name */
	  TrackDBName(dbname, trackID);
	  
	  /* delete track point data base */
	  DmDeleteDatabase(0 /* cardNo */,
	    DmFindDatabase(0 /* cardNo */, dbname));
	}
	handled = true;
	break;
    
      case TrackNewButton:
        /* create new track */
	if ( create_track() ) {
	  /* Track Header DB full */ 
	  FrmAlert(TrackCreateAlert);
	  handled = true;
	  break;
	}

        /* update number of records */
	nr_tracks = DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories);
	
	/* select last (e.g. new) track */
        track_sel = nr_tracks-1;

        /* set top table row to ensure selected item is visible */
	if (track_sel-track_top > TblGetNumberOfRows(tblP)-1) {
          track_top = track_sel-TblGetNumberOfRows(tblP)+1;	    
	}	  

	/* update scroll bar */
	sclP = (ScrollBarPtr)GetObjectFromActiveForm(TrackScrl);
	SclSetScrollBar(sclP, track_sel /* value */,
	  0 /* min */, nr_tracks /* max */,
	  TblGetNumberOfRows(tblP) /* page size */);  
		
	/* redraw table */
	trackDrawTable(track_top, &track_sel);
	
	handled = true;
	break;

    }
    break;

  case frmCloseEvent:
    /* Deallocate Fields' Text Memory */
    SetFieldText(TrackIdentField, NULL, false, true);
    SetFieldText(TrackGpslibField, NULL, false, true);

    TrackFormDeinit(FrmGetActiveForm());
    handled = false;
    break;

  default:
    break;
  }
  
  return handled;
}
