/*****************************************************************************
 *
 * $RCSfile: common.h,v $
 *
 * GPS4Palm
 * Common Data Structure and Function Declarations
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2007-10-08 21:35:10 $
 *
 * $Revision: 1.27.2.1 $
 *
 * $Log: common.h,v $
 * Revision 1.27.2.1  2007-10-08 21:35:10  mp
 * changed kVersion to 0x09
 *
 * Revision 1.27  2005-05-13 18:30:10  mp
 * added AUTOPWROFF
 *
 * Revision 1.26  2005/04/20 21:16:23  mp
 * added gps_btaddr and host_btaddr, changed kVersion to 0x08
 *
 * Revision 1.25  2005/02/25 18:35:47  mp
 * added gPrefs.filecard, changed kVersion to 0x07
 *
 * Revision 1.24  2005/02/19 20:06:36  mp
 * changed kVersion to 0x06
 *
 * Revision 1.23  2005/02/19 20:04:55  mp
 * added FILENAMEMAXLEN, changed filename[32] to filename[FILENAMEMAXLEN]
 *
 * Revision 1.22  2005/02/06 14:56:47  mp
 * changed kVersion to 0x05, added serprefs to PrefsType
 *
 * Revision 1.21  2004/12/19 10:02:23  mp
 * changed kVersion from 0x03 to 0x04
 *
 * Revision 1.20  2004/12/07 17:12:29  mp
 * added field 'time' to GPSType
 *
 * Revision 1.19  2004/12/04 09:34:37  mp
 * modified ActTrackType
 *
 * Revision 1.18  2004/12/02 23:01:36  mp
 * added ActTrackType and TrkPrefsType
 *
 * Revision 1.17  2004/11/30 21:02:29  mp
 * added kTrkHdrDBType, kTrkHdrDBName, kTrkDBType, and kTrkDBName
 *
 * Revision 1.16  2004/11/27 10:33:15  mp
 * updated kVersion
 *
 * Revision 1.15  2004/11/25 17:00:13  mp
 * added approach_alert and approach_sound to PrefsType
 *
 * Revision 1.14  2004/11/20 15:54:39  mp
 * updated kVersion, added auto_wpt to PrefsType
 *
 * Revision 1.13  2004/11/11 21:37:05  mp
 * added Active Route
 *
 * Revision 1.12  2004/04/12 10:28:24  mp
 * modified kVersion
 *
 * Revision 1.11  2004/03/11 21:25:16  mp
 * added GPSType
 *
 * Revision 1.10  2004/03/10 17:20:06  mp
 * removed progressbar() prototype, added ActWayptType and gPrefs.act_wpt
 *
 * Revision 1.9  2004/02/27 16:21:32  mp
 * added database constants
 *
 * Revision 1.8  2004/01/18 11:44:22  mp
 * added POS_UTM
 *
 * Revision 1.7  2003/12/30 14:46:39  mp
 * modified defines to facilitate array indexing
 *
 * Revision 1.6  2003/12/27 17:37:59  mp
 * added map selection and map fetch states
 *
 * Revision 1.5  2003/11/20 20:51:07  mp
 * added units and mapprefs structures and encoding
 *
 * Revision 1.4  2003/11/16 17:46:14  mp
 * added MapParamType Struct
 *
 * Revision 1.3  2003/10/18 16:11:21  mp
 * renamed ImgInfoType to PrefsType
 *
 * Revision 1.2  2003/07/14 20:45:37  mp
 * Added kCreatorId, kPrefsId, and kVersion
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#ifndef COMMON_H
#define COMMON_H
#include <BtCommVdrv.h>

#define kCreatorId	'Gps4'
#define kPrefsId	0x00
#define kVersion	0x09

#define kWaypointDBType	'WAYP'
#define kWaypointDBName	"Waypoints-Gps4"
#define kRouteDBType	'ROUT'
#define kRouteDBName	"Routes-Gps4"
#define kTrkHdrDBType	'TRKH'			/* Track Headers */
#define kTrkHdrDBName	"Trkheaders-Gps4"
#define kTrkDBType	'TRKP'			/* Track Points */
#define kTrkDBName	"Trkpoints-Gps4-"	/* +number (6 hex digits) */

#define APPROACH_DST	100		/**< default approach distance (m) */
#define AUTOPWROFF	1		/**< default Auto Power-Off state */
#define FILENAMEMAXLEN	64		/**< maximum filename length */

/*-------*/
/* Units */
/*-------*/

/* Position */
#define POS_D		0	/* Degrees */ 
#define	POS_DM		1	/* Degrees and Minutes of Arc */
#define POS_DMS		2	/* Degrees, Minutes and Seconds of Arc */
#define POS_UTM		3	/* Universal Transverse Mercator */

/* Distance */
#define DST_KM		0	/* km */
#define DST_MI		1	/* mile (1.609344 km) */
#define DST_NM		2	/* nautic mile (1.852 km) */

/* Altitude */
#define ALT_M		0	/* m */
#define ALT_FT		1	/* feet (0.3048 m) */

/* Speed */
#define SPD_KMH		0	/* km/h */
#define SPD_MS		1	/* m/s (3.6 km/h) */
#define SPD_MIH		2	/* mile/h (1.609344 km/h) */
#define SPD_KTS		3	/* knots (1.852 km/h) */



/*-----------------*/
/* Map Preferences */
/*-----------------*/

/* Download Policy */
#define DLD_A		1	/* download automatically */
#define DLD_Q		2	/* download after query */
#define	DLD_O		3	/* download off */

#define SEL_USER	1	/* user selected map */
#define SEL_AUTO	2	/* auto selected map */

#define FETCH_READY	1	/* FetchMap can be started again */
#define FETCH_START	2	/* FetchMap has been started */
#define FETCH_STOP	3	/* FetchMap disabled */

/* Units Data Structure */ 
typedef struct {
  UInt8		pos_unit;
  UInt8		dst_unit;
  UInt8		alt_unit;
  UInt8		spd_unit;
} UnitsType;

/* Map Preferences Data Type */
typedef struct {
  UInt32	scale;
  UInt8		download;
  UInt8		select;
  UInt8		fetch;
} MapPrefsType;

/* Active Waypoint Type */
typedef struct {
  UInt32	waypointID;	/* waypoint ID (unique) */
  Boolean	valid;		/* waypoint ID valid */
} ActWayptType;
 
/* Active Route Type */
typedef struct {
  UInt32	routeID;	/* route ID (unique) */
  Boolean	valid;		/* route ID valid */
  UInt16	wpt_index;	/* index of active waypoint in route */
} ActRouteType;

/* Active Track Type */
typedef struct {
  UInt32	trackID;	/* track ID (unique) */
  Boolean	valid;		/* track ID valid */
  Boolean	log_state;	/* false: log disabled / true: log enabled */
} ActTrackType;

/* Track Log Preferences Type */
typedef struct {
  UInt16	mode;		/* log mode (0: time / 1: distance) */
  UInt16	ival;		/* log interval (index) */
  UInt16	wrap;		/* when log DB full (0: stop / 1: wrap) */
} TrkPrefsType;

/* Serial Port Preferences Type */
typedef struct {
  UInt32	gps_port;	/* GPS serial port or CncMgr Profile ID */ 
  UInt32	gps_baud;	/* GPS baud rate */
  UInt8		gps_btaddr[btLibDeviceAddressSize];	/* GPS Bluetooth device address */
  UInt32	host_port;	/* Host serial port or CncMgr Profile ID */
  UInt32	host_baud;	/* Host baud rate */
  UInt8		host_btaddr[btLibDeviceAddressSize];	/* GPS Bluetooth device address */
  Boolean	rts_ctrl;	/* RTS controls multiplexer */
} SerPrefsType;

/* Preferences Data Structure */
typedef struct {
  Char		filename[FILENAMEMAXLEN];
  UInt32	filecard;		/* file stream card/VFS volume */  
  Char		imgtype[32];
  UInt16	depth;
  UInt16	mag;
  UInt16	xofs;
  UInt16	yofs;
  UInt16	dx;
  UInt16	dy;
  UInt16	dw;
  UInt16	dh;
  UInt16	form;			/* current form (mode) */
  UInt16	pwroff;			/* auto power-off */
  Boolean	auto_wpt;		/* auto waypoint selection (route) */
  Boolean	approach_alert;		/* approach alert */
  Boolean	approach_sound;		/* approach sound */
  UnitsType	units;			/* units preferences */
  SerPrefsType	serprefs;		/* serial port preferences */
  TrkPrefsType	trk_log;		/* track log preferences */
  MapPrefsType	mapprefs;		/* map preferences */
  ActWayptType	act_wpt;		/* active waypoint */
  ActRouteType	act_rte;		/* active route */
  ActTrackType  act_trk;		/* active track */
  // abmh
  Boolean       mapdispwayptname;
  Boolean       mapdispwaypt;
} PrefsType;

/* MapParamType Struct */
typedef struct {
  double	lat;		/* latitude */
  double	lon;		/* longitude */
  
  UInt32	scale;		/* map scale */

  UInt16	width;		/* map width (px) */
  UInt16	height;		/* map height (px) */

} MapParamType;


/* GPSType Struct */
typedef struct {
  Boolean	valid;		/* GPS Status */
  Boolean	scroll;		/* GPS Status */
  double	lat;		/* Latitude */
  double	lon;		/* Longitude */
  double	sog;		/* Speed Over Ground (in kts !!!) */
  double	cmg;		/* Course Made Good */
  UInt32	time;		/* Palm epoch: seconds since 1904-01-01 12:00 */
} GPSType;

#endif
