/*****************************************************************************
 *
 * $RCSfile: dobitmap.c,v $
 *
 * Display image stored in memory area pointed to by **lines
 *   using WinDrawBitmap
 *
 * created: 2002-12-15
 *
 * History:
 * 2007-07-11:	16 bit support for the Palm TX (fs)
 *              uses little endian coding (intel specific: lowByte / highByte)
 * 2002-12-15:	first version, taken from the program PiNGer which is
 *		  part of the ZBoxZ application suite (Rev. 0.30)
 *		  by Tom Zerucha
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * Changes for Palm TX are Copyright (C) 07/2007 Frank Saurbier
 * <frank.saurbier@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2007-10-05 22:51:06 $
 *
 * $Revision: 1.4.2.2 $
 *
 * $Log: dobitmap.c,v $
 * Revision 1.4.2.2  2007-10-05 22:51:06  mp
 * fixed doxygen comment
 *
 * Revision 1.4.2.1  2007-10-05 22:37:21  mp
 * added Frank's changes (fixes reset bug on TX)
 *
 * Revision 1.5  2007/07/11 16:00:00  fs
 * added 16bit display support, color map from extern gifread
 *
 * Revision 1.4  2005/04/08 14:46:20  mp
 * added High-Density Display support
 *
 * Revision 1.3  2005/03/24 09:16:04  mp
 * updated from SDK-4 to SDK-5r3
 * Note: BitmapType is now called BitmapTypeV2!!! (thanks, PalmSoft ...)
 *
 * Revision 1.2  2003/10/18 16:11:51  mp
 * renamed ImgInfo to gPrefs
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
//#define VERBOSE_DOBIT
#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Bitmap.h>
#include "BitmapRsrc.h"
#include "stringil.h"
#include "dobitmap.h"
#include "common.h"

extern PrefsType gPrefs;
extern char **lines;
extern unsigned short int w, h, bandh;
extern Boolean			gHdFtrSet;

/**********************************************************
 get palette for 16bit TX from gifread.c
 fs 07/2007
 need the palette to build the RGB656 pixel in 16 bit mode
***********************************************************/
extern RGBColorType map[256];


/**********************************************************************/
/**
 * \brief	Display image stored in memory area pointed to by **lines
 *   		using WinDrawBitmap
 *
 * \param	xofs		x-offset within image
 * \param	yofs		y-offset within image
 * \param	mag		magnification (1..8)
 * \param	dx		x-coordinate of bitmap (display coords)
 * \param	dy		y-coordinate of bitmap (display coords)
 * \param	dw		width of bitmap (display coords)
 * \param	dh		height of bitmap (display coords)
 * \param	deep		display depth
 *
 * \note	Depends on w, h, bandh globals
 *
 * \note	The function relies on the internals of the Palm OS
 *		bitmap structures defined in Bitmap.h. Since internals
 * 		of Palm OS structures are subject to change,
 *		this may not work with future OS versions.
 *
 * \note	From SDK-4 to SDK-5r3, BitmapType changed to
 *		BitmapTypeV2 (same members, different name...)
 *
 ***********************************************************************/
void dobitmap(	short unsigned int xofs, short unsigned int yofs,
		short unsigned int mag,
		short unsigned int dx, short unsigned int dy,
		short unsigned int dw, short unsigned int dh,
		short unsigned int deep)
{
  BmpRsrcType	*bmprsrc = NULL;

  #define bp bmprsrc
  
  unsigned char *loc, *loc2;
  unsigned short int imask, bw, i, j, j0, k, bit;
// (fs) 20070710 pointer and helper for 16bit pixel generation
  unsigned short int *locTX, locTmp, locCalc;
  unsigned char loc8;
  PixelFormatType locPF;

  gPrefs.xofs = xofs;
  gPrefs.yofs = yofs;
  gPrefs.dx = dx;
  gPrefs.dy = dy;
  gPrefs.dw = dw;
  gPrefs.dh = dh;
  gPrefs.mag = mag;
  gPrefs.depth = deep;
  
#ifdef VERBOSE_DOBIT
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "dobitmap");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);

  StrPrintF(tmp, "xofs = %d", xofs);
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
  StrPrintF(tmp, "yofs = %d", yofs);
  WinDrawChars(tmp, StrLen(tmp), 0, 10);
  SysTaskDelay(50);
  StrPrintF(tmp, "mag = %d", mag);
  WinDrawChars(tmp, StrLen(tmp), 0, 20);
  SysTaskDelay(50);
  StrPrintF(tmp, "deep = %d", deep);
  WinDrawChars(tmp, StrLen(tmp), 0, 30);
  SysTaskDelay(50);
  StrPrintF(tmp, "dx = %d", dx);
  WinDrawChars(tmp, StrLen(tmp), 0, 40);
  SysTaskDelay(50);
  StrPrintF(tmp, "dy = %d", dy);
  WinDrawChars(tmp, StrLen(tmp), 0, 50);
  SysTaskDelay(50);
  StrPrintF(tmp, "dw = %d", dw);
  WinDrawChars(tmp, StrLen(tmp), 0, 60);
  SysTaskDelay(50);
  StrPrintF(tmp, "dh = %d", dh);
  WinDrawChars(tmp, StrLen(tmp), 0, 70);
  SysTaskDelay(50);

}
#endif


#ifdef OLDSTUFF_PLEASE
/***************
 * old stuff: could be deleted 20070713
 */
  if (gHdFtrSet) {

/* *************************************************************************************************
 * Palm TX:  for 16bit set "pixelFormat565" with "BitmapTypeV3"
 * fs 07/2007
 * nothing changed from the old code, should still work
*/
    if (deep > 8)                 /* Palm TX */
    bmprsrc = BmpRsrcV3Create(dw, 1, 
                              deep, /* deep, works with 8 too, but very bad colors */
                              NULL, /* NULL, no colortable */
                              NULL, kDensityDouble,
                              pixelFormat565 /* 16bit:pixelFormat565, 8bit:pixelFormatIndexed */
                              );
/* Palm TX end ***********************************************************************************/

// old coding
    else
   bmprsrc = BmpRsrcCreate(
       dw, /* width */
      1, /* height */
      deep, /* pixelSize - 1, 2, 4, 8, 16 - (fs: on TX use 16bit) */
      NULL, /* * ct -  null for no color table (preferred) */
      kDensityDouble, /* density - only kDensityLow, kDensityDouble supported */
      -1 /* pixelFormat  pixelFormat565LE , default = -1*/
      );
#ifdef VERBOSE_DOBIT
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "BmpCreate deep = %d", deep);
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif // VERBOSE_DOBIT
  } 
  else 
  {
    /* Note: BmpRsrcV2Create() requires OS Version > 3.5 */
    bmprsrc = (BmpRsrcType *)MemPtrNew(sizeof(BitmapTypeV2) + 320 + 1024);
    bmprsrc->v2.width = dw;
    bmprsrc->v2.height = 1;              // dh
    bmprsrc->v2.pixelSize = deep;
    bmprsrc->v2.rowBytes = deep * 20;    // (deep * (dw+15) / 8) & ~1, i.e. word boundary
    bmprsrc->v2.version = 2;
  }

#endif // OLDSTUFF_PLEASE

#ifndef OLDSTUFF_PLEASE

/**********************************************************************************
 * (fs): thats all, we have to do. The coding above is handles in "BitmapRsrc"
 *       this is a chunk for one line of the bitmap.
 **********************************************************************************/
   bmprsrc = BmpRsrcCreate(
      dw, /* width */
      1, /* height */
      deep, /* pixelSize - 1, 2, 4, 8, 16 - (fs: on TX 16bits are used) */
      NULL, /* * ct -  null for no color table (preferred) */
      kDensityDouble, /* density - only kDensityLow, kDensityDouble supported */
      -1 /* pixelFormat  pixelFormat565LE , default = -1, 16bit:pixelFormat565, 8bit:pixelFormatIndexed */
      );

#endif

  loc = BmpRsrcGetBits(bp);
  /* (fs 2007) Pixelformat Type: defines the endian */
  locPF = BmpRsrcGetPixelFormat(bp);
  
#ifdef VERBOSE_DOBIT
{
  Char		tmp[60];	/* temporary string buffer for debugging */
  
  StrPrintF(tmp, "dobit GetBits ok");
  WinDrawChars(tmp, StrLen(tmp), 0, 0);
  SysTaskDelay(50);
}
#endif
    
  /* really should be bkgd, not 0 */// * dh
  memset(loc, 0, BmpRsrcGetRowBytes(bmprsrc));

//  bw = 8 / deep;
//  (fs) thinks bw meens black white
//  for deep > 8 I set this value to 1
  if (deep > 8)  bw = 1;               /* for 16bit and more bw=1 */
  else                                 /* old coding, for all non Palm TX users*/
  {
    bw = 8 / deep;
    imask = 0xff << (8 - deep);
    imask &= 0xff;
  }

  if (dw * mag + xofs >= w)
    j0 = (w / mag - xofs + bw - 1) / bw;
  else
    j0 = (dw + bw - 1) / bw;

  /* iterate over all lines, sub-sampling if mag > 1 */
//  for (i = 0; yofs < h && i < dh; i++, yofs += mag) {
  for (i = 0; yofs < h && i < dh; i++, yofs += 1) {

    /* destination BMP for drawing */
    loc = BmpRsrcGetBits(bp);

    /* fs needs this cast for Palm TX */
    locTX = (unsigned short int *) loc;

    /* source BMP from GIF/PNG/... decoding */
    loc2 = lines[yofs / bandh] + w * (yofs % bandh) + xofs;

    j = j0;
    switch (deep) {

// (fs) deep = 16, Palm TX support
    case 16:
        while (j--) 
	{
	    loc8 = *loc2;

	    // normal display RGB 565
	    //locTmp = (unsigned short int) (map[loc8].r & 0x1f);    /* R */
	    locCalc = (unsigned short int) map[loc8].r;    /* R */
	    locTmp = (locCalc + 4) * 31 / 255;
            locTmp = locTmp << 6;       
	    //locTmp |= (unsigned short int) (map[loc8].g & 0x3f);   /* G */
            locCalc = (unsigned short int) map[loc8].g;    /* G */
	    locTmp |= (locCalc + 2) * 63 / 255;
            locTmp = locTmp << 5;       
	    //locTmp |= (unsigned short int) (map[loc8].b & 0x1f);   /* B */
            locCalc = (unsigned short int) map[loc8].b;    /* B */
	    locTmp |= (locCalc + 4) * 31 / 255;
	    /* check for little endian */
	    //if (locPF==pixelFormatIndexedLE || locPF==pixelFormat565LE) then
            if (locPF==pixelFormat565LE) {
	      /* for ARM use little endian - checked in RsrcBitmap */
              // *locTX++ = ((locTmp >> 8) & 0x00ff) | ((locTmp << 8) & 0xff00);
	      // can do it without &0x00ff!
              *locTX++ = (locTmp >> 8) | (locTmp << 8);
            } else {
              /** this is big endian for 68K Processors */
              *locTX++ = locTmp;
            }
            loc2 += mag;
	    
	} // while j--
    
    break; /* case deep = 16 */
// end of 16bpp Palm TX support

    case 8:
      if (mag == 1)
        while (j--)
          *loc++ = *loc2++;
      else
        while (j--) {
          *loc++ = *loc2;
          loc2 += mag;
        }
      break;

    case 4:
      if (mag == 1)
        while (j--) {
          k = 0xf0 & *loc2++;
          k |= *loc2++ >> 4;
          *loc++ = ~k;
      } else
        while (j--) {
          k = 0xf0 & *loc2;
          loc2 += mag;
          k |= *loc2 >> 4;
          loc2 += mag;
          *loc++ = ~k;
        }
      break;

    default:                   /* 2 or 1 */
      while (j--) {
        *loc = 0;
        bit = 0;
        do {
          k = ~*loc2 & imask;
          loc2 += mag;
          k >>= deep * bit;
          *loc |= k;
        } while (++bit < bw);
        loc++;
      }
      break;
    }

    /* draw one line of the bitmap */
    WinDrawBitmap((void *)bp, dx, dy + i);

  }
  if (i < dh) {
    /* clear lines below image */
    RectangleType r = { {0, i}, {dh, dh - i}};
    WinEraseRectangle(&r, 0);   /* SHOULD BE BKGD! */
  }

  BmpRsrcDelete(bmprsrc);

}
