/*****************************************************************************
 *
 * $RCSfile: file_io.h,v $
 *
 * File Stream API / VFS API Wrapper Functions
 *
 *
 * This program is Copyright (C) 03/2005 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-15 10:42:25 $
 *
 * $Revision: 1.3 $
 *
 * $Log: file_io.h,v $
 * Revision 1.3  2005-05-15 10:42:25  mp
 * moved functions to linker section 'maps'
 *
 * Revision 1.2  2005/05/14 12:22:29  mp
 * added FileDescrType
 * added parameter fd to fread(), fclose(), feof(), rewind(), and fseek()
 * added return value to fopen
 *
 * Revision 1.1  2005/05/13 18:36:12  mp
 * moved functions from MapForm, added fseek()
 *
 *
 ****************************************************************************/ 

#ifndef FILE_IO_H
#define FILE_IO_H
#include <Extensions/ExpansionMgr/VFSMgr.h>

#define FILEIO_SECTION   __attribute__ ((section ("maps")))

#define SEEK_SET	0	/* Seek from beginning of file.  */
#define SEEK_CUR	1	/* Seek from current position.  */
#define SEEK_END	2	/* Seek from end of file.  */

typedef struct {
  FileHand	fh;	/**< File Stream API file handle */
  FileRef	fr;	/**< VFS API file reference */
} FileDescrType;

unsigned long fread(FileDescrType fd, void *buf,
  unsigned long size)						FILEIO_SECTION;
FileDescrType fopen(UInt32 filecard, const char *filename, 
  const char *dir, Err *err)					FILEIO_SECTION;
void fclose(FileDescrType fd)					FILEIO_SECTION;
int feof(FileDescrType fd)					FILEIO_SECTION;
void rewind(FileDescrType fd)					FILEIO_SECTION;
int fseek(FileDescrType fd, long offset, int whence)		FILEIO_SECTION;

#endif
