/*****************************************************************************
 *
 * $RCSfile: geo.h,v $
 *
 * GPS4Palm Geodesic/Geometric Utility Functions
 *
 * History:
 * 2003-07-06:	first version
 *
 * This program is Copyright (C) 07/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-02 07:22:16 $
 *
 * $Revision: 1.11 $
 *
 * $Log: geo.h,v $
 * Revision 1.11  2005-04-02 07:22:16  mp
 * added deg_to_geodb_str()
 *
 * Revision 1.10  2005/03/25 13:31:35  mp
 * added deg_to_str()
 *
 * Revision 1.9  2004/11/26 19:56:45  mp
 * added deg_to_d_str()
 *
 * Revision 1.8  2004/03/11 18:51:24  mp
 * modified gc_course_sphere()
 *
 * Revision 1.7  2004/03/10 21:57:36  mp
 * added calcR()
 *
 * Revision 1.6  2004/03/10 17:09:13  mp
 * Added rad2deg(), deg2rad(), gc_dist_sphere(), and gc_course_sphere().
 *
 * Revision 1.5  2004/02/28 13:20:51  mp
 * added semi2deg(), deg2semi(), deg_to_dm_str(), and deg_to_dms_str()
 *
 * Revision 1.4  2004/01/18 11:43:07  mp
 * added defines, added Lat_To_UTM_Belt()
 *
 * Revision 1.3  2004/01/17 15:46:16  mp
 * moved to code section geo
 *
 * Revision 1.2  2003/10/20 19:13:10  mp
 * replaced parameter zoom by scale
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#ifndef GEO_H
#define GEO_H

#define SEMIMAJOR_AXIS	6378137			// m, GRS80/WGS84 Ellipsoid
#define FLATTENING	(1 / 298.257223563)	// GRS80/WGS84 Ellipsoid
#define M_PI		3.14159265358979323e0	// pi
#define PIXELFACT	2834.65			// pixel/m

#define GEO_SECTION   __attribute__ ((section ("geo")))


void init_radius(void)						GEO_SECTION;
double calcR(double lat)					GEO_SECTION;
void calcxy(Int32 *posx, Int32 *posy, double lon, double lat,
		double zero_lon, double zero_lat, UInt32 scale)	GEO_SECTION;
char Lat_To_UTM_Belt(double lat) 				GEO_SECTION;
double semi2deg(Int32 semi)					GEO_SECTION;
Int32 deg2semi(double deg)					GEO_SECTION;
double rad2deg(double rad)					GEO_SECTION;
double deg2rad(double deg)					GEO_SECTION;
void deg_to_d_str(double lat, double lon,
		char *lat_d_str, char *lon_d_str)		GEO_SECTION;
void deg_to_dm_str(double lat, double lon,
		char *lat_dm_str, char *lon_dm_str)		GEO_SECTION;
void deg_to_dms_str(double lat, double lon,
		char *lat_dms_str, char *lon_dms_str)		GEO_SECTION;
void deg_to_geodb_str(double lat, double lon,
		char *lat_str, char *lon_str)			GEO_SECTION;
Boolean deg_to_str(double lat, double lon, UInt8 pos_unit,
                char *lat_strP, char *lon_strP, char *field_strP,
		   double *eastP, double *northP)		GEO_SECTION;
double gc_dist_sphere(double lat1, double lon1,
  double lat2, double lon2)					GEO_SECTION;
double gc_course_sphere(double lat1, double lon1,
  double lat2, double lon2)					GEO_SECTION;
#endif
