/*****************************************************************************
 *
 * $RCSfile: imginfo.c,v $
 *
 * Image File Information Functions
 *
 * Note: For GIF, the code has been taken from gifread.c.
 *       FOR PNG, the code has been taken from pngdec.c.
 *
 * This program is Copyright (C) 03/2005 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-14 12:31:29 $
 *
 * $Revision: 1.1 $
 *
 * $Log: imginfo.c,v $
 * Revision 1.1  2005-05-14 12:31:29  mp
 * initial version
 *
 *
 ****************************************************************************/ 
#include <PalmOS.h>
#include "file_io.h"
#include "imginfo.h"
#include "stringil.h"

#define USAT(a,b)	(((b)<<8)|(a))
#define BS		1440

/* Mark's macros to extract big-endian short and long ints: */
typedef unsigned char uch;
typedef unsigned short ush;
typedef unsigned long ulg;
#define SH(p) ((ush)(uch)((p)[1]) | ((ush)(uch)((p)[0]) << 8))
#define LG(p) ((ulg)(SH((p)+2)) | ((ulg)(SH(p)) << 16))

/* magic bytes to determine image type */
const char png_magic[]    = "\211PNG\r\n\032\n";
const char gif87a_magic[] = "GIF87a";
const char gif89a_magic[] = "GIF89a";
const char jpeg_magic[]   = "\xFF\xD8"; /* 0xFF 0xD8 (SOI - Start of Image) */


Boolean imginfo(FileDescrType fd, UInt8 *type, UInt16 *width, UInt16 *height)
{
  unsigned char		buf[24];
  unsigned int		mapmax;
  unsigned int		i;
  long int		sz;
  
  fread(fd, buf, 8);
  
  if (strncmp(buf, png_magic, strlen((char *)png_magic)) == 0) {
    /* file is a PNG */
    *type = IMGTYPE_PNG;

    /* search for IHDR */
    while (!feof(fd)) {

      if (8 != fread(fd, buf, 8))
	break;
      sz = LG(buf);
      (void) memcpy(buf, &buf[4], 4);

      buf[4] = 0;

      /*------* 
       | IHDR | 
       *------*/
      if (strcmp(buf, "IHDR") == 0) {

	fread(fd, buf, sz + 4);

	*width  = LG(buf);
	*height = LG(buf + 4);
	rewind(fd);
	return false;

      } else {
	/* BYPASS REST OF UNKNOWN RECORD */
	i = 0;
	while (sz > i) {
          sz -= i;
          i = (sz > BS) ? BS : sz;
          fread(fd, buf, i);
	}
	fread(fd, buf, 4);  /* crc */
      }
    } /* while (!feof()) */
    rewind(fd);
    return true;
    
  } else if (strncmp(buf, gif87a_magic, strlen((char *)gif87a_magic)) == 0 ||
             strncmp(buf, gif89a_magic, strlen((char *)gif89a_magic)) == 0) {
    rewind(fd);

    /* file is a GIF */
    *type = IMGTYPE_GIF;
  
    fread(fd, buf, 13);
    
    /* read past colormap */
    mapmax = 2 << (buf[10] & 0x07);
  
    if (buf[10] & 0x80) {
      /* Global CMap */
      fseek(fd, mapmax * 3, SEEK_CUR);
    }
  
    /* skip optional extension blocks (GIF89a) */
    while (true) {
      if (fread(fd, buf, 1) == 0) {
	/* error: end of file */
	rewind(fd);
	return true;
      }
      if (buf[0] == ',') {
	/* Image Separator found */
	break;
      } else if (buf[0] == '!') {
	/* Extension Introducer found */
	fread(fd, buf, 1);
	if (buf[0] == 0xFE) {
          /* Plain Text Extension */
	  /* read until Block Terminator found */
	  do {
	    fread(fd, buf, 1);
	  } while (buf[0] != '\0');
	} else {
          /* any other extension, next byte: Block Size */
	  fread(fd, buf, 1);
	  for (i = buf[0] + 1; i > 0; i--) {
	    /* read complete block and Block Terminator */
	    fread(fd, buf, 1);
	  }
	}
      } else {
	/* Invalid Data */
	rewind(fd);
	return true;
      }
    } /* while (true) */

    /* read remaining bytes of Image Descriptor */
    fread(fd, &buf[1], 9);
    
    *width  = USAT(buf[5], buf[6]);
    *height = USAT(buf[7], buf[8]);
    rewind(fd);
    return false;
    
  } else if (strncmp(buf, jpeg_magic, strlen((char *)jpeg_magic)) == 0) {
    /* file is a JPEG */
    *type = IMGTYPE_JPEG;
        
    /* currently not supported */
    *width  = 0;
    *height = 0;
    rewind(fd);
    return false;
    
  } else {
    /* file type is unknown */
    *type = IMGTYPE_UNKNOWN;
    rewind(fd);
    return true;
    
  }
}
