/*****************************************************************************
 *
 * $RCSfile: progressbar.c,v $
 *
 * Draws a progress bar
 *
 * created: 2002-12-30
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * Changes for Palm TX are Copyright (C) 07/2007 Frank Saurbier
 * <frank.saurbier@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2007-10-05 22:04:21 $
 *
 * $Revision: 1.2.2.1 $
 *
 * $Log: progressbar.c,v $
 * Revision 1.2.2.1  2007-10-05 22:04:21  mp
 * added Frank's changes (NOPROGRESSBAR)
 *
 * Revision 1.3  2007/07/11 16:00:00  fs
 * added NOPROGRESSBAR for testing
 *
 * Revision 1.2  2004/03/10 17:17:53  mp
 * added progressbar.h
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "progressbar.h"

/*
 * progressbar: 
 *   Draws a progress bar. The width is fixed to 100 pixels.
 *
 * Note: progressbar() has to be called with val = 0 initially
 *         for proper initialization!
 * 
 * Parameters:
 * -----------
 *	ypos	vertical position of progress bar
 *	val	current value
 *	total	final value (progress = 100%)
 */
void progressbar(UInt16 ypos, UInt32 val, UInt32 total)
{
#ifndef NOPROGRESSBAR
  static UInt32		inc;	/* increment for progress bar update */
  static UInt32		w;	/* bar width */
  static UInt32		mod;	/* modulo counter */
  RectangleType		rect;	/* rectangle structure */
  
  if (val == 0) {
    /* Initialization */
    inc = total / 100;
    mod = 0;
    w = 0;
    RctSetRectangle(&rect, 29 /* xpos */, ypos, 102 /* w */, 4 /* h */);
    WinEraseRectangle(&rect, 0 /* corner diameter */);
    WinDrawRectangleFrame(simpleFrame, &rect);
  } else {
    if (mod == inc) {
      mod = 0;
      w++;
      RctSetRectangle(&rect, 30 /* xpos */, ypos+1, w, 2 /* h */);
      WinDrawRectangle(&rect, 0 /* corner diameter */);
    } else {
      mod++;
    }
  }
#endif
}
