/*****************************************************************************
 *
 * $RCSfile: stringil.h,v $
 *
 * String Handling Inline Functions
 *
 * 2002-12-15:	first version, taken from the ZBoxZ application suite
 *		(Rev. 0.30) by Tom Zerucha
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-02-19 14:34:07 $
 *
 * $Revision: 1.3 $
 *
 * $Log: stringil.h,v $
 * Revision 1.3  2005-02-19 14:34:07  mp
 * added #include <Unix/sys_types.h>
 *
 * Revision 1.2  2004/03/08 19:06:13  mp
 * Changed declaration of toupper()/tolower() from extern to static
 * to solve linker error.
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/

/* linux (posix?) says val should be an int */
//#include <string.h>
#include <Unix/sys_types.h>
#define VALT int
#define SIZET size_t

//need to add international...
static
inline char toupper(register char c)
{
  if( c >= 'a' && c <= 'z' )
    return c - 32;
  return c;
}

static
inline char tolower(register char c)
{
  if( c >= 'Z' && c <= 'Z' )
    return c + 32;
  return c;
}

extern inline void *memset(void *dst, VALT val, SIZET num)
{
  register void *ds = dst;
  while (num--)
    *((char *) dst)++ = val;
  return ds;
}

extern inline void *memcpy(void *dst, const void *src, SIZET num)
{
  register void *ds = dst;
  while (num--)
    *((char *) dst)++ = *((char *) src)++;
  return ds;
}

extern inline int memcmp(void *dst, const void *src, SIZET num)
{

  while (num-- && (*((char *) dst)++ == *((char *) src)++));
  if (num >= 0)
    return *--((char *) dst) - *--((char *) src);
  else
    return 0;
}

extern inline void *memccpy(void *dst, const void *src, int c, SIZET num)
{
  register void *ds = dst;
  while (num-- && *((char *) src) != c)
    *((char *) dst)++ = *((char *) src)++;
  if (num >= 0)
    *((char *) dst)++ = *((char *) src)++;
  return ds;
}

extern inline void *memmove(void *dst, const void *src, SIZET num)
{
  register void *ds = dst;
  if (dst < src || dst >= src + num)
    while (num--)
      *((char *) dst)++ = *((char *) src)++;
  else {
    dst += num;
    src += num;
    while (num--)
      *--((char *) dst) = *--((char *) src);
  }
  return ds;
}

extern inline SIZET strlen(char *dst)
{
  register SIZET num = 0;
  while ((*dst++))
    num++;
  return num;
}

extern inline char *strcat(char *dst, const char *src)
{
  register char *ds = dst;
  while (*dst)
    dst++;
  while ((*dst++ = *src++));
  return ds;
}

extern inline char *strncat(char *dst, const char *src, SIZET num)
{
  register char *ds = dst;
  while (*dst)
    dst++;
  while (num-- && (*dst++ = *src++));
  return ds;
}

extern inline char *strcpy(char *dst, const char *src)
{
  register char *ds = dst;
  while ((*dst++ = *src++));
  return ds;
}

extern inline int strcmp(const char *dst, const char *src)
{
  while (*dst && *src && (*dst++ == *src++));
  if (*dst && *src)
    return *--src - *--dst;
  else
    return *src - *dst;
}

extern inline int strncmp(const char *dst, const char *src, SIZET num)
{
  while (num && *dst && *src && (*dst++ == *src++))
    num--;
  if( !num )
    return 0;
  if (*dst && *src)
    return *--src - *--dst;
  else
    return *src - *dst;
}

extern inline char *strncpy(char *dst, const char *src, SIZET num)
{
  register char *ds = dst;
  while (num-- && (*dst++ = *src++));
  if (num >= 0)
    while (num--)
      *dst++ = 0;
  return ds;
}

extern inline void *memchr(const void *dst, VALT val, SIZET num)
{
  char *dstl = (void *)dst;
  while (num-- && *dstl++ != val);
  if (num >= 0)
    return (void *) --dstl;
  else
    return NULL;
}

extern inline char *strchr(const char *dst, int val)
{
  char *dstl = (char *)dst;
  while (*dstl && *dstl != val)
    dstl++;
  if (*dstl != val)
    return NULL;
  else
    return dstl;
}
