/*
   gpslib.h : include file for gpslib
   Copyright (C) 1999,2000,2001 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* $Id: gpslib.h,v 1.1 2004-11-21 10:58:44 mp Exp $ */

/* Types Garmin : */

/* typedef unsigned short word; */
/* typedef unsigned long longword; */
/* typedef short Symbol_Type; */

#include "Garmin.h"

#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))

/* Debugging : */
#define GPSLDBG_INFOS 128				   // informational
#define GPSLDBG_PKTS  64				   // packets
#define GPSLDBG_CALLS 32				   // procedures calls/exits
#define GPSLDBG_UNEXP 8					   // unexpected things


#define ETX '\003'
#define DLE '\020'
// #define GPS_ONLY 1
#define Pid_Async_Event 0x1C
#define REPLYTIMEOUT 1000				   /* timeout de reponse du GPS en ms */
#define MAXREPEAT 3					   // nombre de reemissions max avant abandon
#define MAXCOMMERRORS 10				   /* nombre max d'erreurs de comm avant abandon */
// #define Rte_Hdr 1
#define Palm2GarminTimeDiff 2713910400UL		   // epoch(Garmin) - epoch (Palm)
// conversion "Garmin time" to "Palm secondes"
#define GarminTime2PalmSeconds(t) (t+Palm2GarminTimeDiff)
#define NmeaMaxFields 32				   /* max number of fields in a NMEA sentence */

typedef struct
{
    int entries_used;
    char *field[NmeaMaxFields];
}
NmeaFields;

#define MAXDEVS 16
typedef struct
{
  union{
    UInt32 serDevCreator;
    char serDevCreatorString[5];
  } u[MAXDEVS];
  Char *serDevPortInfoStr[MAXDEVS];
}
SerialDevInfo;

/// Type of GPS emulated
enum Personality
{
    GIIPlus, G12CX, GII, MAP175, GIII, eTrex, GPS76
};
typedef enum Personality Personality;

// Codes d'erreur renvoys par gpslib :
#define	gpslibErrParamErr			 (appErrorClass | 1)
#define	gpslibErrBadMode			 (appErrorClass | 2)
#define	gpslibErrBusy			         (appErrorClass | 3)
#define	gpslibErrCommError	   	         (appErrorClass | 4)
#define	gpslibErrNoData	   	             	(appErrorClass | 5)
#define	gpslibErrNothingToDo   	             	(appErrorClass | 6)


enum ProtocolFamily
{ ProdProto, ProtoProto, DevProto, WptProto, RteProto, TrkProto, ProxyProto, AlmanProto, TimeProto,
	PosProto, PVTProto };
typedef enum ProtocolFamily ProtocolFamily;

typedef struct			       /* enchainement des type de donnees suivant protocoles */
{
    LocalFormatID datatype;	       /*  type (format) de donnee */
    byte next;			       /* index prochain type de donnee */
}
ProtoDesc;

/// Controls accepted by GpslibControl
enum GpslibCtrl
  {GpslibCtrl_NoOp, GpslibCtrl_LooseProto };
typedef enum GpslibCtrl GpslibCtrl;


/*  Callbacks declarations passed to gpslib : */

/** callback proc called when a waypoint is received
 @param wpt pointer on a GarminObject received from gps
 @param total total count of items sent as indicated by gps
 @param current current item number
 @param format format of waypoint
 @param size size of waypoint
 @return error code
*/
typedef Err StoreWptCallbackProc (GarminObject * wpt, UInt16 total, UInt16 current,
				  LocalFormatID format, byte size);

/** Callback proc called when a trackpoint is received

 @param trk pointer on a raw track point or track header received from gps
 @param  total total count of items sent as indicated by gps
 @param  current current item number
 @param  format format of the object received
 @param  size size of object
 @return error code (0 if no error)
*/
typedef Err StoreTrkCallbackProc (GarminObject * trk, UInt16 total, UInt16 current,
				  LocalFormatID format, byte size);

/** Callback proc called when a route element is received

 @param  rte pointer on the raw object from the gps
 @param  total total count of items sent as indicated by gps
 @param  current current item number (first=1)
 @param  format format of the object received (route header or waypoint)
 @param  size size of object
 @return error code (0 if no error)
*/
typedef Err StoreRteCallbackProc (GarminObject * rte, UInt16 total, UInt16 current,
				  LocalFormatID format, byte size);

typedef StoreWptCallbackProc * StoreWptCallbackProcPtr;
typedef StoreTrkCallbackProc * StoreTrkCallbackProcPtr;
typedef StoreRteCallbackProc * StoreRteCallbackProcPtr;

/** Callback proc called to fetch a waypoint to send to the gps

 @param  current current item number (first=0)
 @param  format wanted format of waypoint
 @param size pointer on a byte to write returned object size
 @return pointer on waypoint ready to send
*/
typedef GarminObject * GetNextWptCallbackProc (UInt16 current, LocalFormatID format,
					      byte *size);

/** Callback proc called to fetch a track element to send to the gps

 @param  current index of current element (first=0)
 @param  format format of element requested
 @param size pointer to store size of element returned
 @param eot true if the current item is the last element of current track
 @return pointer on element returned (Garmin format)
*/
typedef GarminObject * GetNextTrkCallbackProc (UInt16 current, LocalFormatID format,
					      byte *size, Boolean * eot);
///


/** Callback proc called to fetch a route element to send to the gps

 @param current current item number (header or waypoint, first=0)
 @param format wanted format of element
 @param size pointer to store size of element returned
 @param h pointer on a boolean to set if item is a header
 @param eor pointer on a boolean to set if it's the last element
 @return pointer on element returned (Garmin format)
*/
typedef GarminObject * GetNextRteCallbackProc (UInt16 current, LocalFormatID format,
					      byte *size, Boolean * h, Boolean * eor);

/** Callback proc called to fetch a full route to send to a Furuno gps

 @param  current  current item number
 @return pointer on string to send to gps
*/
typedef char * GetNextFullRteCallbackProc (UInt16 current);

typedef GetNextWptCallbackProc * GetNextWptCallbackProcPtr;
typedef GetNextTrkCallbackProc * GetNextTrkCallbackProcPtr;
typedef GetNextRteCallbackProc * GetNextRteCallbackProcPtr;
typedef GetNextFullRteCallbackProc * GetNextFullRteCallbackProcPtr;


/** Callback proc called to fetch a waypoint to send to a Furuno gps

 @param  current item number to fetch (first=0)
 @return pointer on string in Furuno format
*/
typedef char * GetNextCharWptCallbackProc (UInt16 current);

typedef GetNextCharWptCallbackProc * GetNextCharWptCallbackProcPtr;


/** Callback proc called when a raw packet is received (in asynchronous events mode)

 @param  p pointer on raw packet
 @param  current packet number
 @return error code (0 if ok)
*/
typedef Err StorePacketCallbackProc (Garmin_Packet * p, UInt16 current);

typedef StorePacketCallbackProc * StorePacketCallbackProcPtr;


/** Callback proc called to fetch a raw packet

 @param current packet number
 @return pointer on the retrieved packet
*/
typedef Garmin_Packet * GetNextPacketCallbackProc (UInt16 current);
typedef GetNextPacketCallbackProc * GetNextPacketCallbackProcPtr;


/** Callback proc called to log strings (typicaly for debugging)

@param p string to log
@return nothing
*/
typedef void LogProc (char *p);
typedef LogProc *LogProcPtr;


/* prototypes gpslib.c */
extern void GpslibTicks (void);
extern Err GarminGetWpts (StoreWptCallbackProcPtr p, FieldPtr FPtr);
extern Err GarminGetTrks (StoreTrkCallbackProcPtr p, FieldPtr FPtr);
extern Err GarminGetRtes (StoreRteCallbackProcPtr p, FieldPtr FPtr);
extern Err GarminPutWpts (GetNextWptCallbackProcPtr p, FieldPtr FPtr, UInt16 count);
extern Err FurunoPutWpts (GetNextCharWptCallbackProcPtr, FieldPtr FPtr, UInt16 count);
extern Err GarminPutRtes (GetNextRteCallbackProcPtr p, FieldPtr FPtr, UInt16 count);
extern Err FurunoPutRtes (GetNextFullRteCallbackProcPtr p, FieldPtr FPtr, UInt16 count);
extern Err GarminPutTrks (GetNextTrkCallbackProcPtr p, FieldPtr FPtr, UInt16 count);
extern void GPSEndOfOperation (void);
extern Err GarminGetTime (FieldPtr FPtr);
extern UInt32 D600ToSeconds (Raw_D600_Date_Time_Type * d600p);
extern char *D600ToAscii (D600_Date_Time_Type * d600p);
extern Err GarminSetHostMode (FieldPtr FPtr, Personality Perso);
extern Err GarminGetPosition (FieldPtr FPtr);
extern Err GarminEnableAsyncEvents (StorePacketCallbackProcPtr p, FieldPtr FPtr, UInt16 mask);
extern Err GarminPutPacket (GetNextPacketCallbackProcPtr p, FieldPtr FPtr, UInt16 count);
extern Err GarminGetRemoteInfos (FieldPtr FPtr);
extern void GpslibSetDebugLevel (UInt16 debuglevel, LogProcPtr logprocp);
extern Err GarminStartPVT (void);
extern Err GarminStopPVT (void);
extern Err GarminGetPVT (D800_Pvt_Data_Type * pvtp);
extern void NMEAParse (byte *p, NmeaFields * fp);
extern UInt16 NMEAGetSentence (byte *p);
extern void NMEASendSentence (char *sentencep, int timeout);
extern GarminProtoID GarminGetProtocol (ProtocolFamily p);
extern UInt16 GetSerialDevsInfos (SerialDevInfo Infos[]);
extern void GpslibSetSerialPort (UInt32 port);
extern Err GpslibControl (GpslibCtrl op, void * datas);

// Evenements generes par gpslib :
/// Posted when a complete NMEA sentence is available
#define SentenceAvailableEvent firstUserEvent
/// Posted when the library is closed (*EndOfOperation called or end of transfer)
#define EndOfOperationEvent (firstUserEvent +2)
/// Posted when a transfer waypoints request (Cmnd_Transfer_Wpt) is received (host mode)
#define TransferWptReqEvent (firstUserEvent +3)
/// Posted when a transfer routes request (Cmnd_Transfer_Rte) is received (host mode)
#define TransferRteReqEvent (firstUserEvent +4)
/// Posted when a transfer tracks request (Cmnd_Transfer_Trk) is received (host mode)
#define TransferTrkReqEvent (firstUserEvent +5)
/// Posted when a GPS time packet (Pid_Date_Time_Data) is received (after calling GarminGetTime), data.sclRepeat.time contains PalmOS time in seconds
#define TimeReceivedEvent   (firstUserEvent +6)
/// Posted when a power off request (Cmnd_Turn_Off_Pwr) is received (host mode)
#define PowerOffEvent       (firstUserEvent +7)
/// Posted when a Pid_Position_Data packet is received (after calling GarminGetPosition)
#define PositionReceivedEvent (firstUserEvent +8)
/// Posted when a Pid_Async_Event request  is received
#define AsyncEventReqEvent  (firstUserEvent +9)
/// Posted when a PVT paquet is received
#define PvtReceivedEvent (firstUserEvent +10)
