/*****************************************************************************
 *
 * $RCSfile: box.c,v $
 *
 * linbox core functions
 *
 * linbox is part of
 *
 * ZBoxZ - http://palmboxer.sourceforge.net
 * 
 * by Tom Zerucha
 *
 *
 * 2002-12-30 Matthias Prinke <matthias.prinke@surfeu.de>
 *
 * Changes as compared to ZBoxZ Rev. 0.30:
 * - added check for file open (linbox.c, linunbox.c)
 * - modified linbox to use only base name (i.e. not complete path)
 *   for PDB header (linbox.c)
 * - fixed bug when boxing multiple files (linbox.c)
 * - fixed PDB header generation for long filenames (name field has at most 31
 *   characters + string terminator) (box.c)
 *
 * This is free software under the terms of GNU's GPL.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-19 21:44:53 $
 *
 * $Revision: 1.2 $
 *
 * $Log: box.c,v $
 * Revision 1.2  2005-04-19 21:44:53  mp
 * added parameter 'creator'
 *
 * Revision 1.1  2005/04/15 15:16:37  mp
 * imported from ZBoxZ Rev. 0.30
 *
 *
 ****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#define DEFAULT_TYPE	"DATA"
#define DEFAULT_CREATOR	"BOXR"

static unsigned short htons(unsigned short y ) {
  unsigned short x;
  unsigned char *c = (unsigned char *)&y;
  x = *c++ << 8;
  x |= *c;
  return x;
}

static unsigned long htonl(unsigned long y ) {
  unsigned long x;
  unsigned char *c = (unsigned char *)&y;
  x = *c++ << 24;
  x |= *c++ << 16;
  x |= *c++ << 8;
  x |= *c;
  return x;
}

int box(FILE * infd, const char *fname, const char *creator, FILE * outfd)
{

  unsigned long cofst;
  char bbuf[5000];
  unsigned short xshort;
  unsigned long xlong, xid;
  long bbuflen;
  int maxsect;
    
  memset(bbuf, 0, 32);
  /* 
   * The filename field of the Database Header is 31 bytes
   *   + 1 terminator byte long.
   */
  if (strlen(fname) < 32)
    strcpy(bbuf, fname);
  else
    strcpy(bbuf, &fname[strlen(fname) - 31]);

  fwrite(bbuf, 1, 32, outfd);

  memset(bbuf, 0, 4);
  bbuf[1] = 0x88;
  bbuf[3] = 1;
  fwrite(bbuf, 1, 4, outfd);

  xlong = time(NULL) + ((66 * 365 + 17) * 24 * 3600UL);
  xlong = htonl(xlong);

  fwrite(&xlong, 1, 4, outfd);
  fwrite(&xlong, 1, 4, outfd);
  fwrite(&xlong, 1, 4, outfd);

  memset(bbuf, 0, 12);
  fwrite(bbuf, 1, 12, outfd);

  fwrite(DEFAULT_TYPE,    1, 4, outfd);
  fwrite((*creator == 0) ? DEFAULT_CREATOR : creator, 1, 4, outfd);

  bbuf[0] = 0x28;
  fwrite(bbuf, 1, 8, outfd);

  fseek(infd, 0, SEEK_END);
  xlong = ftell(infd);
  fseek(infd, 0, SEEK_SET);

  maxsect = xlong / 4096 + 1;
  xshort = htons(maxsect);
  fwrite(&xshort, 1, 2, outfd);

  cofst = 80 + 8 * maxsect;

  xid = 0x40000000UL + (rand() & 0x7fffffUL);

  strcpy(bbuf, "DBLK");
  bbuf[5] = bbuf[7] = 0;
  bbuf[6] = 0x10;

  while (maxsect--) {
    xlong = htonl(cofst);
    fwrite(&xlong, 1, 4, outfd);
    xlong = htonl(xid++);
    fwrite(&xlong, 1, 4, outfd);
    cofst += 4104;
  }

  xshort = 0;
  fwrite(&xshort, 1, 2, outfd);

  for (;;) {

    bbuflen = fread(&bbuf[8], 1, 4096, infd);

    if (bbuflen != 4096)
      break;

    fwrite(bbuf, 1, bbuflen + 8, outfd);
  }
  if (bbuflen < 0)
    exit(-3);

  fwrite(bbuf, 1, 4, outfd);
  xlong = htonl(bbuflen);
  fwrite(&xlong, 1, 4, outfd);
  fwrite(&bbuf[8], 1, bbuflen, outfd);

  fclose(infd);
  fclose(outfd);
  return 0;

}
