/*****************************************************************************
 *
 * $RCSfile: linunbox.c,v $
 *
 * linunbox main functions
 *
 * linunbox is part of
 *
 * ZBoxZ - http://palmboxer.sourceforge.net
 * 
 * by Tom Zerucha
 *
 *
 * 2002-12-30 Matthias Prinke <matthias.prinke@surfeu.de>
 *
 * Changes as compared to ZBoxZ Rev. 0.30:
 * - added check for file open (linbox.c, linunbox.c)
 * - modified linbox/linunbox to use only base name (i.e. not complete path)
 *   for PDB header (linbox.c) and output file (linunbox.c)
 * - fixed bug when boxing multiple files (linbox.c)
 * - fixed PDB header generation for long filenames (name field has at most 31
 *   characters + string terminator) (box.c)
 *
 * This is free software under the terms of GNU's GPL.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-19 21:46:29 $
 *
 * $Revision: 1.2 $
 *
 * $Log: linunbox.c,v $
 * Revision 1.2  2005-04-19 21:46:29  mp
 * added command line options, modified output file-/pathname
 *
 * Revision 1.1  2005/04/15 15:16:37  mp
 * imported from ZBoxZ Rev. 0.30
 *
 *
 ****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>	/* exit() */
#include <getopt.h>
#ifdef HAVE_BASENAME
  #include <libgen.h>
#endif

#define VERSION "V1.0"

int unbox(FILE * infd, FILE * outfd);
static void usage(char *argv[]);

#ifndef HAVE_BASENAME
static char *basename(char *path)
{
  char *str;
  
  str = path + strlen(path) - 1;
  
  while (str != path) {
    if (*str == '\\' || *str == '/') {
	  str++;
	  break;
	}
	str--;
  }
  
  return str;
}
#endif

int main(int argc, char *argv[])
{
  char bbuf[256];
  char dir[256];
  FILE *infd;
  FILE *outfd;
  char opt;
  int argp;
  
  strcpy(dir, "");
  
  while ((opt = getopt(argc, argv, "hd:c:")) != -1) {
    switch (opt) {
      /* output directory */
      case ('d'):
	if (sscanf(optarg, "%s", dir) != 1) {
          usage(argv);
	  exit(EXIT_FAILURE);
	}
	strcat(dir, "/");
	break;
			
      /* Help */
      case ('h'):
	      usage(argv);
	      exit(EXIT_SUCCESS);
	      break;

      default:
	      fprintf(stderr, "Unknown option\n");
	      usage(argv);
	      exit(EXIT_FAILURE);
    }
  }

  if (argc <= optind) {
    fprintf(stderr, "Usage: %s inputfile(s)\n", argv[0]);
    exit(1);
  }

  for (argp = optind; argp < argc; argp++) {

    if (!(infd = fopen(argv[argp], "rb"))) {
      fprintf(stderr, "Cannot open %s for reading!\n", argv[argp]);
      exit(-1);
    }
    
    strcpy(bbuf, dir);
    strcat(bbuf, basename(argv[argp]));
    if (!strcmp(&bbuf[strlen(bbuf) - 4], ".pdb"))
      bbuf[strlen(bbuf) - 4] = 0;
    else
      strcat(bbuf, ".out");
    
    
    if (!(outfd = fopen(bbuf, "wb"))) {
      fprintf(stderr, "Cannot open %s for writing!\n", bbuf);
      exit(-1);
    }

    unbox(infd, outfd);

  }
  return 0;
}

static void usage(char *argv[])
{
  fprintf(stderr, "%s %s -- retrieve file from PalmOS database (PDB)\n\n",
    basename(argv[0]), VERSION);
  fprintf(stderr, "Usage: %s [options] <input_file(s)>\n\n",
    basename(argv[0]));
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "       -d <output_dir>: file(s) output directory\n");
  fprintf(stderr, "       -h             : Display help\n");
}
