/*****************************************************************************
 *
 * $RCSfile: unbox.c,v $
 *
 * linunbox core functions
 *
 * linunbox is part of
 *
 * ZBoxZ - http://palmboxer.sourceforge.net
 * 
 * by Tom Zerucha
 *
 *
 * 2002-12-30 Matthias Prinke <matthias.prinke@surfeu.de>
 *
 * Changes as compared to ZBoxZ Rev. 0.30:
 * - added check for file open (linbox.c, linunbox.c)
 * - modified linbox to use only base name (i.e. not complete path)
 *   for PDB header (linbox.c)
 * - fixed bug when boxing multiple files (linbox.c)
 * - fixed PDB header generation for long filenames (name field has at most 31
 *   characters + string terminator) (box.c)
 *
 * This is free software under the terms of GNU's GPL.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-15 15:16:37 $
 *
 * $Revision: 1.1 $
 *
 * $Log: unbox.c,v $
 * Revision 1.1  2005-04-15 15:16:37  mp
 * imported from ZBoxZ Rev. 0.30
 *
 *
 ****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>	/* exit() */

static unsigned short htons(unsigned short y ) {
  unsigned short x;
  unsigned char *c = (unsigned char *)&y;
  x = *c++ << 8;
  x |= *c;
  return x;
}

static unsigned long htonl(unsigned long y ) {
  unsigned long x;
  unsigned char *c = (unsigned char *)&y;
  x = *c++ << 24;
  x |= *c++ << 16;
  x |= *c++ << 8;
  x |= *c;
  return x;
}

int unbox(FILE * infd, FILE * outfd)
{
  unsigned char bbuf[5000];
  unsigned long xlong;
  unsigned short xshort;
  int i;

  fseek(infd, 76, SEEK_SET);
  fread(&xshort, 1, 2, infd);
  xshort = htons(xshort);
  fread(&xlong, 1, 4, infd);
  xlong = htonl(xlong);
  fseek(infd, xlong, SEEK_SET);

  for (i = 0; i < xshort; i++) {
    fread(bbuf, 1, 4, infd);
    if ( strncmp(bbuf, "DBLK", 4))
      exit(-2);
    fread(&xlong, 1, 4, infd);
    xlong = htonl(xlong);
    xlong = fread(bbuf, 1, xlong, infd);
    fwrite(bbuf, 1, xlong, outfd);
  }
  fclose(outfd);
  fclose(infd);

  return 0;
}
